/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */

#ifndef LPISUPPORT_PROGRAMCONFIG_H
#define LPISUPPORT_PROGRAMCONFIG_H
#include <LPISupport/LPISupportCommonDefs.hpp>
#include <iostream>
#include <string>
#include <memory>
#include <map>
#include <getopt.h>
#include <sstream> // string stream


namespace LPISupport {

/**
 * @brief Class with common variables and methods for program configuration (base for derived classes specific to testing programs)
 *
 */
class ProgramConfig {
public:
    /// Data type for the map with names of program arguments and their parameters
    typedef std::map<std::string, std::string> ExecArgsMap;

    /**
     * @brief Struct for basic program configuration
     */
    struct ConfigTraits {
        /// Name of the program
        std::string     programName = "";
        /// First line of program usage (summary)
        std::string     usageSummary = "[options]";
        /// Rest of line of program usage (with valid options and their parameters)
        std::string     usageOptions = "-h\tshow this help\n";
        /// Supported short arguments string (one letter for every supported option, in standard getopt function format)
        std::string     shortArgs = "h";
        /// Short required supported arguments string (not all the short arguments of previous field have to be mandatory)
        std::string     shortRequiredArgs = "";
        /// Map for storing short arguments read from program command line
        ProgramConfig::ExecArgsMap execArgsMap;

        ConfigTraits() {}
        /// Destructor for object
        virtual ~ConfigTraits() {}
        /**
         * @brief Method for setting usageOptions, shortArgs and shortRequiredArgs class fields
         * @param optionName letter for the short option of program
         * @param optionParameterName name of the option parameter (emtpy if there is no parameter)
         * @param explanation text explaining option and its parameter (if any)
         * @param mandatory true value if option is mandatory, false if optional
         */
        void addSupportedShortOption(char optionName, std::string optionParameterName, std::string explanation, bool mandatory);
        /**
         * @brief method for setting class fields according to run parameters read
         */
        virtual void configure();
        /**
         * @brief method for getting usage text from usageSummary and usageOption class fields
         * @return usage text
         */
        std::string getUsage();
    };
    /**
     *
     * @param argc number of program run arguments (including program name)
     * @param argv array of text strings with the names of the program run arguments
     */
    ProgramConfig(int argc, char* argv[]);
    /**
     *
     */
    virtual ~ProgramConfig();
    /**
     * @brief Returns the configuration object for this object
     *
     * @return smart pointer to the configuration object
     */
    virtual std::shared_ptr<ConfigTraits> getConfigTraits() {return pConfigTraits;}
    /**
     * @brief read program run arguments and fills up a map with their names and the name of their parameters
     *
     * @param argc number of program run arguments (including program name)
     * @param argv array of text strings with the names of the program run arguments
     * @param longOptions array of long options according to getopt standard function
     * @param shortArgs array of short options according to getopt standard function
     * @param shortRequiredArgs array with the short mandatory options
     * @param usage text with program usage explanation
     * @return a map with the program run arguments read (key is short argument name, value is the name of
     * the corresponding parameter, if any)
     */
    static ExecArgsMap readProgramArguments(int argc, char *argv[], struct option longOptions[], std::string shortArgs,
                                            std::string shortRequiredArgs, std::string usage);
    /**
     * @brief read program run short arguments and fills up a map with their names and the name of their parameters
     *
     * @param argc number of program run arguments (including program name)
     * @param argv array of text strings with the names of the program run arguments
     * @param shortArgs array of short options according to getopt standard function
     * @param shortRequiredArgs array with the short mandatory options
     * @param usage text with program usage explanation
     * @return a map with the program run arguments read (key is short argument name, value is the name of
     * the corresponding parameter, if any)
     * @return
     */
    static ExecArgsMap readProgramShortArguments(int argc, char *argv[], std::string shortArgs, std::string shortRequiredArgs,
                                                 std::string usage);
    /**
     * @brief returns the filename part of a path
     * @param path path as a text string
     * @return filename extracted from the path
     */
    static std::string fileName(std::string path);

protected:
    /**
     * Constructor without parameters
     */
    ProgramConfig();
    /**
     * @brief read program short run arguments and fills up a map with their names and the name of their parameters (the map is
     * stored in a ConfigTraits internal class field
     * @param argc number of program run arguments (including program name)
     * @param argv array of text strings with the names of the program run arguments
     */
    void readExecArgs(int argc, char* argv[]);

    /**
     * @brief check if all required run arguments are present (exits program if false)
     */
    void checkRequiredArgsPresent();
    /// Smart pointer to configuration object
    std::shared_ptr<ConfigTraits> pConfigTraits = nullptr;

private:

};
} /* namespace LPISupport */

#endif // LPISUPPORT_PROGRAMCONFIG_H
