#include <LPISupport/ProgramConfig.hpp>

// Uncomment to show class-specific debug messages
//#define PROGRAMCONFIG_DEBUG

#if !defined NDEBUG && defined PROGRAMCONFIG_DEBUG
    #define PROGRAMCONFIG_CERR(x) CERR(x)
#else
    #define PROGRAMCONFIG_CERR(x)
    #undef PROGRAMCONFIG_DEBUG
#endif

namespace LPISupport {

ProgramConfig::ProgramConfig() {

}

ProgramConfig::ProgramConfig(int argc, char* argv[]) {
    pConfigTraits = std::make_shared<ConfigTraits>();
	readExecArgs(argc, argv);
	checkRequiredArgsPresent();
	pConfigTraits->configure();
}

/**
 * @brief Destructor for class (empty)
 * 
 */
ProgramConfig::~ProgramConfig()
{

}

void ProgramConfig::ConfigTraits::addSupportedShortOption(char optionName, std::string optionParameterName, std::string explanation, bool mandatory) {
	shortArgs.append(std::string(1, optionName));
	usageOptions.append("-" + std::string(1, optionName) + "\t");
	if (!optionParameterName.empty()) {
		shortArgs.append(":");
		usageOptions.append("<" + optionParameterName + ">\t");
	}
	usageOptions.append(explanation + "\n");
	if (mandatory) {
		shortRequiredArgs.append(std::string(1, optionName));
	}
}

/**
  * @brief Reads and tests program short or long arguments and returns a map with short name of program arguments as map keys 
  * and parameter value of program arguments as map values.
  * @param[in] argc number of program arguments (incluiding program name)
  * @param[in] argv array of program arguments as strings
  * @param[in] longOptions struct with long names for options
  * @param[in] shortOptions string with a letter for every supported option
  * @param[in] shortRequiredArgs string with a letter for every required option
  * @param[in] usage string with text explaining correct program invocation
  * @return map with short program option name as a key and program option value as map value
  */
ProgramConfig::ExecArgsMap ProgramConfig::readProgramArguments(int argc, char *argv[], struct option longOptions[], 
                                                               std::string shortOptions, std::string shortRequiredArgs, 
                                                               std::string usage) {
    ExecArgsMap resultMap;
    int option;
    int option_index=0;

    resultMap["programName"] = argv[0];
    try {
        while(1) {
            option = getopt_long(argc, argv, shortOptions.c_str(),
                                 longOptions, &option_index);
            switch(option) {
            case -1: // End of  params
                return resultMap;
            case '?': /* getopt_long already printed an error message. */
                exit(-1);
                break;
            default:
                PROGRAMCONFIG_CERR("option_index: " << option_index
                     << " long_option: " << longOptions[option_index].name
                     << " option: " << (char) option << " optopt: " << (char) optopt
                     << " optarg: " << optarg << std::endl);
                resultMap[std::string(1, (char)option)] = optarg;
                break;
            }
        }
    } catch (const std::invalid_argument &ex) {
        throw std::invalid_argument (std::string(__FILE__) +
                                     ": "+std::to_string(__LINE__)+ ": " +
                                     ex.what() +
                                     ": Error reading program parameters");
        abort();
    }

    return resultMap;
}

/**
  * @brief Reads and tests program short arguments and returns a map with short name of program arguments as map keys and 
  * parameter value of program arguments as map values.
  * @param[in] argc number of program arguments (incluiding program name)
  * @param[in] argv array of program arguments as strings
  * @param[in] shortArgs string with a letter for every supported option and tags for required and optional option arguments 
  * (in standard getopt function format)
  * @param[in] shortRequiredArgs string with a letter for every required option of the supported options
  * @param[in] usage string with text explaining correct program invocation
  * @return map with program arguments short name as map keys and program arguments parameter value as map values
  */
ProgramConfig::ExecArgsMap ProgramConfig::readProgramShortArguments(int argc, char *argv[], std::string shortArgs, std::string shortRequiredArgs, 
                                                    std::string usage) {
    ExecArgsMap resultMap;
    int option;
    resultMap["programName"] = fileName(argv[0]);
    //opterr = 0; // Avoids sending message with error to stderr if option is not supported
    try {
        while ((option = getopt(argc, argv, shortArgs.c_str())) != -1) {
            switch (option) {
                case '?':
                case 'h':
                    /* getopt_long already printed an error message. */
                    std::cerr << "Usage: " << fileName(std::string(argv[0])) << " " << usage << std::endl;
                    exit(-1);
                    break;
                default:
                	if (optarg == nullptr)
                		resultMap[std::string(1, (char)option)] = "";
                	else
                		resultMap[std::string(1, (char)option)] = optarg;
                    // Remove option charater from string of required arguments characers, if present
                    if (shortRequiredArgs != "") {
                        PROGRAMCONFIG_CERR( "option read: '" << ((char) option) << "', requiredArgs before erase: " << shortRequiredArgs << std::endl);
                        size_t requiredArgPos = shortRequiredArgs.find((char) option);
                        if (requiredArgPos != std::string::npos) {
                            shortRequiredArgs.erase(requiredArgPos, 1);
                            PROGRAMCONFIG_CERR("requiredArgs after erase: " << shortRequiredArgs << std::endl);
                        } else {
                            PROGRAMCONFIG_CERR("option '" << ((char) option) << "' not in requiredArgs" << std::endl);
                        }
                    }
            }
        }
        if (shortRequiredArgs.size() != 0) {
            std::cerr << "Missing required arg(s): " << shortRequiredArgs << std::endl;
            std::cerr << "Usage: " << fileName(std::string(argv[0])) << " " << usage << std::endl;
            exit(-1);
       }
    } catch (const std::invalid_argument &ex) {
        throw std::invalid_argument (std::string(__FILE__) +
                                     ": "+std::to_string(__LINE__)+ ": " +
                                     ex.what() +
                                     ": Error reading program parameters");
        abort();
    }
    return resultMap;
}

/**
 * @brief Returns the file name part from a path
 * 
 * @param[in] path absolute or relative path of a file name
 * @return string with the file name part of the path
 */
std::string ProgramConfig::fileName(std::string path) {
    std::size_t found = path.find_last_of("/\\");
    if (found == std::string::npos) {
        return path;
    } else {
        return path.substr(found+1);
    }
}

std::string ProgramConfig::ConfigTraits::getUsage() {
    return usageSummary + "\n" + usageOptions;
}

/**
 * @brief Reads program arguments and set, depending on their values, the fields of the configuration object 
 * pConfigTraits.
 * 
 * It calls readProgramShortArguments method (that reads and tests program short arguments and returns a map with 
 * short name of program arguments as map keys and parameter value of program arguments as map values) and calls 
 * setConfig pure virtual method (that reads map with program arguments information and sets, depending on its values, 
 * fields of the class variable which stores configuration information, pConfigTraits).
 * 
 * @param[in] argc number of program arguments (incluiding program name) 
 * @param[in] argv array of program arguments as strings 
 */
void ProgramConfig::readExecArgs(int argc, char* argv[]) {
    pConfigTraits->execArgsMap = 
        ProgramConfig::readProgramShortArguments(argc, argv, pConfigTraits->shortArgs, 
                                                 pConfigTraits->shortRequiredArgs, 
                                                 pConfigTraits->getUsage());
}

void ProgramConfig::checkRequiredArgsPresent() {
    if (pConfigTraits->shortRequiredArgs.size() != 0) {
        std::cerr << "Missing required arg(s): " << pConfigTraits->shortRequiredArgs << std::endl;
        std::cerr << "Usage: " << pConfigTraits->programName << " " << pConfigTraits->getUsage() << std::endl;
        exit(-1);
    }
}

void ProgramConfig::ConfigTraits::configure() {
    // Visible ConfigTraits struct is this class ConfigTraits
    ExecArgsMap::iterator it;
    it = execArgsMap.find("programName");
    if (it != execArgsMap.end()) {
    	programName = it->second;
    	execArgsMap.erase("programName");
    }
}
}
#undef PROGRAMCONFIG_DEBUG

