/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */

#include <OpenCLIPER/OpenCLIPERDataModel.hpp>
#include "OpenCLIPER/ProgramConfig.hpp"

using namespace OpenCLIPER;

int main (int argc, char* argv[]) {

	try {
	    ProgramConfig* pProgramConfig = new ProgramConfig(argc, argv);
	    auto pConfigTraits = std::dynamic_pointer_cast<ProgramConfig::ConfigTraits>(
	        pProgramConfig->getConfigTraits());

		// Step 1: get a new OpenCLIPER app and initialize computing device
		CLapp::PlatformTraits platformTraits = pConfigTraits->platformTraits;
		CLapp::DeviceTraits deviceTraits = pConfigTraits->deviceTraits;

		auto pCLapp = CLapp::create(platformTraits,deviceTraits);

		// Step 2: load OpenCL kernel(s)
		pCLapp->loadKernels("rss.cl");

		// Step 3: load input data from Matlab file
		std::shared_ptr<KData> pInputKData(new KData(pCLapp, "MRIdataSOS.mat"));

		// Step 4: create output with suitable size and register input and output in our CL app
		std::shared_ptr<XData> pOutputXData(new XData(pCLapp, std::dynamic_pointer_cast<KData>(pInputKData)));

		// Step 5: create new process bound to our CL app
		// and set its input/output data sets
		std::unique_ptr<Process> pProcess(new SimpleMRIReconSOS(pCLapp));
		pProcess->setInput(pInputKData);
		pProcess->setOutput(pOutputXData);

		// Step 6: initialize process
		pProcess->init();

		// Step 7: launch process
		pProcess->launch();

		// Step 8 get data back from computing device (automatically called by matlabSave)
		// and store output data into matlab file
		pOutputXData->matlabSave("outputFramesSOS.mat");
        pOutputXData->show();

		// Step 9: clean up
		pProcess.reset(nullptr);
		pInputKData = nullptr;
		pOutputXData = nullptr;
		pCLapp = nullptr;
	}
	catch(cl::BuildError& e) {
		CLapp::dumpBuildError(e);
	}
	catch(cl::Error& e) {
		std::cerr << "CL Exception "<<e.err()<<" ("<<CLapp::getOpenCLErrorCodeStr(e.err())<<"): " << e.what() << std::endl;
	}
	catch (std::exception& e) {
		std::cerr << "Generic exception: " << e.what() << std::endl;
	}
}
