/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
#ifndef CLAPP_HPP
#define CLAPP_HPP

#define CL_HPP_ENABLE_EXCEPTIONS
#define CL_HPP_TARGET_OPENCL_VERSION 120
#define CL_HPP_MINIMUM_OPENCL_VERSION 120

#if defined(__APPLE__) || defined(__MACOSX)
#include<OpenCL/cl.hpp>
#else
#include<CL/cl2.hpp>
#endif

// Ensure AMD/nVidia constants/types to retrieve PCIe bus id of devices are defined
#ifndef CL_DEVICE_TOPOLOGY_AMD
    #define CL_DEVICE_TOPOLOGY_AMD 0x4037
#endif

#ifndef CL_DEVICE_TOPOLOGY_TYPE_PCIE_AMD
    #define CL_DEVICE_TOPOLOGY_TYPE_PCIE_AMD 1
#endif

#ifndef HAVE_CL_DEVICE_TOPOLOGY_AMD
    typedef union {
	struct { cl_uint type; cl_uint data[5]; } raw;
	struct { cl_uint type; cl_char unused[17]; cl_char bus; cl_char device; cl_char function; } pcie;
    } cl_device_topology_amd;
#endif

#ifndef CL_DEVICE_PCI_BUS_ID_NV
    #define CL_DEVICE_PCI_BUS_ID_NV 0x4008
#endif

#ifndef CL_DEVICE_PCI_SLOT_ID_NV
    #define CL_DEVICE_PCI_SLOT_ID_NV 0x4009
#endif

#ifdef HAVE_HIP
    #define __HIP_PLATFORM_HCC__
    #include <hip/hip_runtime_api.h>
#endif


#include<vector>
#include<map>
// For gethostname POSIX function
#include<stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>

#include<OpenCLIPER/OpenCLIPERDataModelCommonDefs.hpp>
#include<OpenCLIPER/DeviceDataProperties.hpp>
#include<LPISupport/Utils.hpp>
#include<LPISupport/InfoItems.hpp>

#include<OpenCLIPER/processes/Complex2Real.hpp>
#include<OpenCLIPER/processes/SumReduce.hpp>
#include<OpenCLIPER/processes/ScalarMultiply.hpp>
#include<OpenCLIPER/processes/NormalizeShow.hpp>

// cl2.hpp includes the | operator but not &
// We need it to test for supported command queue properties
inline cl::QueueProperties operator&(cl::QueueProperties lhs, cl::QueueProperties rhs) {
    return static_cast<cl::QueueProperties>(static_cast<cl_command_queue_properties>(lhs) & static_cast<cl_command_queue_properties>(rhs));
}

namespace OpenCLIPER {

class Data;

// List of processes for OpenCLIPER's internal use
// Note: since InternalProcess is forward-declared, a deleter must be defined manually for the unique_ptr to know how to destruct the InternalProcess object
struct InternalProcesses {
    //std::unique_ptr<InternalProcess,InternalProcessDeleter> complex2real;
    SumReduce sumReduce;
    Complex2Real complex2real;
    ScalarMultiply scalarMultiply;
    NormalizeShow normalizeShow;
};

/// Data type for a list of OpenCL kernels
typedef std::map<std::string,cl::Kernel> kernelListT;

/**
 * @brief Class representing an application with an OpenCL device bound
 *
 * This class stores:
 * * infomation about available and selected devices
 * * a list of Process subclasses objects (representing algorithms to be executed on selected device)
 * * a list of Data subclasses objects (used as input and/or output data for Process subclasses objects).
 */
class CLapp: public std::enable_shared_from_this<CLapp> {

    friend class Data;
    friend class XData;
    friend class KData;


public:

    /// Enumerated type with the device type options
    enum DeviceType {
            DEVICE_TYPE_ANY=CL_DEVICE_TYPE_ALL, ///< any device is valid CPU or GPU
            DEVICE_TYPE_CPU=CL_DEVICE_TYPE_CPU, ///< only CPU devices are valid
            DEVICE_TYPE_GPU=CL_DEVICE_TYPE_GPU ///< only GPU devices are valid
    };

    /// Struct with options for selection of OpenCL platform
    struct PlatformTraits {
        /// platform name
        std::string		     name;
        /// platform vendor
        std::string		     vendor;
        /// platform version
        std::string		     version;
        /// platform required extensions
        std::vector<std::string> extensions;
    };

    /// Struct with options for selection of OpenCL device
    struct DeviceTraits {
        /// type of device (CPU, GPU, etc.)
        DeviceType			type;
        /// device name
        std::string			name;
        /// device vendor
        std::string			vendor;
        /// device version
        std::string			version;
        /// device required extensions
        std::vector<std::string>	extensions;
        /// device queue properties
        cl::QueueProperties		queueProperties;
        /// Use HIP if available
        bool                            useHIP = false;

        /**
         * @brief Default constructor for struct fields initialization.
         *
         */
        DeviceTraits(DeviceType t=DEVICE_TYPE_ANY,cl::QueueProperties p=cl::QueueProperties::None):
            type(t), queueProperties(p) {}
    };

    /// @brief These methods substitute public constructors.
    // We use create() methods and a private constructor to avoid two-phase initializations since we need a fully constructed CLapp method
    // to assign it to processes which are internal to CLapp itself.
    static std::shared_ptr<CLapp> create(const PlatformTraits& platformTraits,const DeviceTraits& deviceTraits);
    static std::shared_ptr<CLapp> create(const PlatformTraits& platformTraits,const DeviceTraits& deviceTraits, const std::string& filename);
    static std::shared_ptr<CLapp> create(const PlatformTraits& platformTraits,const DeviceTraits& deviceTraits, const std::vector<std::string>& filenames);
    ~CLapp();

    static const char* getOpenCLErrorCodeStr(const cl_int err);
    static const std::string getOpenCLErrorInfoStr(cl::Error err, std::string fileName,
                                                         unsigned int line);
    static const std::string getOpenCLErrorExtendedInfoStr(cl::Error err,
                                                           std::string fileName,
                                                           unsigned int line,
                                                           cl::Program program,
                                                           cl::Device selected_device);
    static cl::NDRange calcLocalSize(const cl::Kernel& kernel, const cl::Device& device, const cl::NDRange& globalSize);

    /**
     * @brief Sets error text explantation from error code (integer)
     * @param[in] err error code
     * @param[out] errStr text message for the error code
     */
    void 	setOpenCLErrorCodeStr(const cl_int err,const char* errStr) { errStrings[err]=errStr; }
    static int	dumpInfo();
    void	init();
    void	init(const PlatformTraits& platformTraits, const DeviceTraits& deviceTraits, bool quiet=false);
    void	loadKernels(const std::string&		     filename, const char* compilerOptionsArg=nullptr);
    void	loadKernels(const std::vector<std::string> & filenames,const char* compilerOptionsArg=nullptr);
    static void dumpBuildError(cl::BuildError& e);
    /**
     * @brief Gets OpenCL context
     * @return reference to the OpenCL context
     */
    const cl::Context&	getContext() const {return context;}

    /**
     * @brief Gets OpenCL device from list of devices
     * @param[in] i position of device in the list of devices
     * @return reference to selected device
     */
    const cl::Device&	getDevice(const size_t i=0) const {return devices[i];}

    /**
     * @brief Gets HIP device
     * @return reference to selected HIP device
     */
#ifdef HAVE_HIP
    const hipDevice_t	getHIPDevice() const {return hipDevice;}
#endif

    /**
     * @brief Gets OpenCL command queue from list of command queues
     * @param[in] i position of queue in the list of queues
     * @return reference to selected queue
     */
    cl::CommandQueue&	getCommandQueue(const size_t i=0) {return commandQueues[i];}

    /**
     * @brief Gets OpenCL program from list of programs
     * @param[in] i position of program in the list of programs
     * @return reference to selected program
     */
    const cl::Program&	getProgram(const size_t i=0) const {return programs[i];}

    /**
     * @brief Gets kernel (type cl::Kernel) from list of kernels
     * @param[in] i index of the kernel in the kernels list
     * @return reference to selected kernel
     */
    cl::Kernel&		getKernel(const size_t i=0) {kernelListT::iterator j(kernels.begin()); std::advance(j,i); return j->second;}	// Warning: no bounds checking!

    /**
     * @brief Gets kernel (type cl::Kernel) from name
     * @param[in] name name of the kernel
     * @return reference to selected kernel
     */
    cl::Kernel&		getKernel(const char* name) {return getKernel(std::string(name));}
    cl::Kernel&		getKernel(const std::string& name);
    static 	cl_uint roundUp(cl_uint numToRound, cl_uint baseNumber);
    void                    checkDataHandle(DataHandle handle, std::string specificMessage);
    Data*	getData (DataHandle handle);
    std::shared_ptr<DeviceDataProperties>   getDeviceDataProperties (DataHandle handle);
    void                    host2Device(DataHandle handle, SyncSource host2DeviceSyncSource = SYNCSOURCEDEFAULT);
    void                    device2Host(std::shared_ptr<Data> pData, SyncSource device2HostSyncSource, bool queueFinish = true);
    void                    device2Host(DataHandle handle, SyncSource device2HostSyncSource, bool queueFinish = true);
    std::string getDeviceTypeAsString(size_t i=0);
    LPISupport::InfoItems getHWSWInfo(size_t i=0);
    std::string getDeviceName(size_t i=0);
    std::string getDeviceVendor(size_t i=0);
    cl::NDRange getMaxLocalWorkItemSizes(cl::NDRange globalSizes);

    cl::Buffer* getDeviceBuffer(DataHandle handle, dimIndexType NDArrayIndex);
    cl::Buffer* getDeviceBuffer(DataHandle handle);
    void*       getHIPDeviceBuffer(DataHandle handle);
    cl::Buffer* getCompleteDeviceBuffer(DataHandle handle);
    cl::Image* getDeviceImage(DataHandle handle, dimIndexType NDArrayIndex);
    void* getHostBuffer(DataHandle handle, dimIndexType NDArrayIndex);
    void* getHostImage(DataHandle handle, dimIndexType NDArrayIndex);
    const void* getDataDimsAndStridesHostBuffer(DataHandle handle);
    cl::Buffer* getDataDimsAndStridesDeviceBuffer(DataHandle handle);
    static long score(const cl::Device& device);
//     static void freeResources(std::unique_ptr<ProcessCore>& pProcess, std::shared_ptr<Data>& pIn, std::shared_ptr<Data>& pOut,
//                               std::shared_ptr<CLapp>& pCLapp);
protected:
    DataHandle      addData(std::shared_ptr<Data> pData, SyncSource hostDeviceSync = SYNCSOURCEDEFAULT);
    DataHandle      addData(Data* pData, SyncSource hostDeviceSync = SYNCSOURCEDEFAULT);
    void            delData(DataHandle handle);

private:
    /// @brief Default constructor (empty). Note: use create() methods instead of calling this (private) constructor.
    // The real work is done in the public create() methods
    CLapp() {}

    /// OpenCL platform
    cl::Platform			platform;

    /// OpenCL context
    cl::Context				context;

    /// List of OpenCL devices
    std::vector<cl::Device>		devices;

    /// HIP device equivalent to first OpenCL device. We should maintain a vector of pairs or so here...
#ifdef HAVE_HIP
    hipDevice_t                         hipDevice;
#endif


    /// List of OpenCL command queues
    std::vector<cl::CommandQueue>	commandQueues;

    /// List of programs
    std::vector<cl::Program>		programs;

    /// List of kernels
    kernelListT				kernels;

    /// String error
    std::string				errStr;

    /// Map with data handles as keys and smart shared pointers to DeviceDataProperties objects as values
    std::map<DataHandle, std::shared_ptr<DeviceDataProperties>>         dataMap;

    /// Current valid value for data keys (initially not valid)
    DataHandle			nextDataKey = FIRSTVALIDDATAHANDLE;
    static std::map<const cl_int,const char*>	errStrings;

    /// List of internal processes
    InternalProcesses internalProcesses;
};

} //namespace OpenCLIPER
#endif // CLAPP_HPP
