/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
#ifndef INCLUDE_OPENCLIPER_DATA_HPP_
#define INCLUDE_OPENCLIPER_DATA_HPP_

#include <OpenCLIPER/NDArray.hpp>
#include <OpenCLIPER/DeviceDataProperties.hpp>
#include <OpenCLIPER/MatVarInfo.hpp>

namespace OpenCLIPER {
class CLapp;
class NDArray;

/// Class Data - Class that includes data and properties common to k-space and x-space images.
class Data: public std::enable_shared_from_this<Data> {
    /// XData must have access to private part of Data
    friend class XData;
    /// KData must have access to private part of Data
    friend class KData;
    /// SensitivityMapsData must have access to private part of Data
    friend class SensitivityMapsData;
    /// SamplingMasksData must have access to private part of Data
    friend class SamplingMasksData;
    /// DataProperties must have access to protected part of Data
    friend class DeviceDataProperties;
    /// CLapp must have access to protected part of Data
    friend class CLapp;
    /// Trajectory must have access to private part of Data
    friend class Trajectories;
    /// MatVarDimsData must have access to private part of Data
    friend class MatVarDimsData;

public:

    // Constructors: empty data
    Data(ElementDataType elementDataType = std::type_index(typeid(complexType)));

    // Constructors: from a bunch of NDarrays
    Data(std::vector<NDArray*>*& pNDArrays, ElementDataType elementDataType = std::type_index(typeid(complexType)));
    Data(std::vector<NDArray*>*& pNDArrays, std::vector<dimIndexType>*& pDynDims, ElementDataType elementDataType = std::type_index(typeid(complexType)));

    // Constructors: from given dimensions (uninitialized)
    Data(dimIndexType width, ElementDataType elementDataType = std::type_index(typeid(complexType)));
    Data(dimIndexType width, dimIndexType height, ElementDataType elementDataType = std::type_index(typeid(complexType)));
    Data(dimIndexType width, dimIndexType height, dimIndexType depth, ElementDataType elementDataType = std::type_index(typeid(complexType)));
    Data(std::vector<dimIndexType>* pArrayDims, ElementDataType elementDataType = std::type_index(typeid(complexType)));
    Data(std::vector<std::vector<dimIndexType>*>*& pArraysDims, ElementDataType elementDataType = std::type_index(typeid(complexType)));
    Data(std::vector<std::vector<dimIndexType>*>*& pArraysDims, std::vector<dimIndexType>*& pDynDims, ElementDataType elementDataType = std::type_index(typeid(complexType)));

    // Constructors: from another Data object
    Data(const Data* sourceData, bool copyData = false);
    Data(std::shared_ptr<Data> sourceData, bool copyData = false);
    Data(std::shared_ptr<Data> sourceData, ElementDataType newElementDataType);

    virtual ~Data();
    void setApp(std::shared_ptr<CLapp> pCLapp, SyncSource host2DeviceSync = SYNCSOURCEDEFAULT);

    // Getters
    numberOfDimensionsType getNumNDArrays() const;
    std::vector<const NDArray*>* getNDArrays() const ;
    NDArray* getNDArray(dimIndexType NDArrayIndex);
    std::vector<const NDArray*>* getData() const ;
    const NDArray* getDataAtDynPos(const std::vector<dimIndexType> dynIndexes) const;
    const index1DType get1DIndexFromDynPos(const std::vector<dimIndexType> dynIndexes) const;
    /**
     * @brief Returns true of all NDArrays have the same spatial dimensions.
     * @return true if all NDArrays have the same spatial dimensions
     */
    const bool getAllSizesEqual() const {
        return allSizesEqual;
    }

    /**
     * @brief Get temporal dimensions vector.
     * @return vector of temporal dimensions
     */
    const std::vector<dimIndexType>* getDynDims() const {
        return pDynDims.get();
    }

    /**
     * @brief Get size of temporal dimensions vector.
     * @return size of vector of temporal dimensions
     */
    numberOfDimensionsType getNDynDims() const {
        return pDynDims->size();
    }

    const index1DType getDynDimsTotalSize() const;

    /**
     * @brief Gets vector of spatial and temporal data dimensions stored in host memory.
     * @return pointer to vector of spatial and temporal data dimensions
     */
    const std::vector<dimIndexType>* getDataDimsAndStridesVector() const {
        return pDataDimsAndStridesVector.get();
    }

    const void* getDataDimsAndStridesHostBuffer() const;

    /**
     * @brief Gets value describing data type used for data elments in this object.
     * @return value describing data type used
     */
    ElementDataType getElementDataType() const {
        return elementDataType;
    }

    void* getHostBuffer(dimIndexType index = 0);
    void* getHostImage(dimIndexType index);
    cl::Buffer* getCompleteDeviceBuffer();
    cl::Buffer* getDeviceBuffer();
    void*       getHIPDeviceBuffer();
    cl::Buffer* getDeviceBuffer(dimIndexType index);
    cl::Image* getDeviceImage(dimIndexType index);
    cl::Buffer* getDataDimsAndStridesDeviceBuffer();

    //Setters
    void setData(std::vector<NDArray*>*& pNDArrays);
    void setDynDims(std::vector<dimIndexType>*& pDynDims);

    // Other methods
    const std::vector<const NDArray*>* getFragment(FragmentSpecif specif);
    static std::string buildFileNamePrefix(const std::string prefix, const std::vector<dimIndexType>* pDims);
    static std::string buildFileNameSuffix(std::string suffix="", std::string fileExtension="raw");
    void saveRawHostData(const SyncSource syncSource, const std::string fileNamePrefix, std::vector<std::string> fileNameSuffixes,
                         const std::string fileNameExtension=".raw");

    // Pure virtual method, must be reimplemented (dimensions depend on the existience of number of coils class field)
    virtual void calcDataDims() = 0;
    virtual void calcDataStrides(dimIndexType deviceMemBaseAddrAlignInBytes);
    virtual void calcDimsAndStridesVector(dimIndexType deviceMemBaseAddrAlignInBytes);
    static std::map<std::string,matvar_t*>* readMatlabVariablesFromFile(std::string fileName, std::vector<std::string>
variableNames);
    static std::map<std::string,matvar_t*>* readMatlabVariablesFromFile(std::string fileName);

    /**
     * @brief Returns the size of the base element of stored data
     *
     * @return the value of the size of the base element
     */
    size_t getElementSize() {return elementSize;}
    /**
     * @brief returns the handle associated to this Data object (used in Data list managed by CLapp object)
     *
     * @return the associated handle
     */
    DataHandle getHandle() const {return myDataHandle;}
    const std::string hostImageToString(std::string title, dimIndexType index);
    const std::string hostBufferToString(std::string title, dimIndexType index);
    virtual void device2Host(SyncSource device2HostSyncSource, bool queueFinish = true);

    // Show related stuff
    struct ShowParms {
	unsigned timeDimension;	// No. of temporal (dynamic) dimension to use as time when showing videos
	float fps;		// frames per second to show, if there is at least one temporal (dynamic) dimension
	float scaleFactor;	// Window scaling factor: 1.0 = no scaling, more than 1 = upscale, less than 1 = downscale
	float normFactor;	// If normalize==true, multiply all pixels by this value before showing (0 means the average value of all NDArrays shifted to 0.5)
	float waitTime;		// milliseconds. 0 means wait forever (or key pressed, if keyEnds=true)
	ComplexPart complexPart;	// part of complex data to show (if the data to show is complex, ignored otherwise)
	std::string title;	// name and identifier of display window
	bool keyEnds;		// key pressed ends display of NDArray
	bool normalize;		// normalize data before showing it

	ShowParms(): timeDimension(0), fps(30), scaleFactor(1), normFactor(0), waitTime(0), complexPart(ComplexPart::ABS), title("Data view"), keyEnds(true), normalize(true) {}
    };

    void show(ShowParms* parms=nullptr);

protected:
    /**
     * @brief Sets handle (used in Data list managed by CLapp object) for this object.
     *
     * @param handle new handle value for the object
     */
    void setHandle(DataHandle handle) {myDataHandle = handle;}
    void loadRawHostData(const std::string fileNamePrefix, std::vector<std::vector< dimIndexType >*>*& pArraysDims,
                         std::vector <dimIndexType>*& pDynDims, std::vector<std::string> fileNameSuffixes,
                         const std::string fileNameExtension=".raw");

    void internalSetData(std::vector<NDArray*>*& pNDArrays);
    void createFromNDArraysVector(std::vector<NDArray*>*& pNDArrays, std::vector<dimIndexType>*& pDynDims);
    void createEmptyDataWithDimensions(std::vector<std::vector<dimIndexType>*>*& pArraysDims,
                                       std::vector<dimIndexType>*& pDynDims);
    void calcDataAlignedSize(dimIndexType NDArray1DIndex, dimIndexType deviceMemBaseAddrAlignInBytes);
    void checkValiditySpatialDimensions(const std::vector<dimIndexType>* pDims);
    void checkValiditySpatialDimensions(const std::vector<std::vector<dimIndexType>*>* pArraysDims);
    void loadMatlabHostData(matvar_t* matvar, dimIndexType numOfSpatialDimensions, dimIndexType numNDArraysToRead);
    void fillMatlabVarInfo(mat_t* matfp, std::string matVarname, MatVarInfo* pMatVarInfo, SyncSource syncSource);
    void readDimsMatlabVar(const matvar_t *pDimsMatVar, std::vector<dimIndexType> *pCompleteSpatialDimsVector, numCoilsType &numCoils, std::vector<dimIndexType> *pTemporalDimsVector);

    // Associations
    /** @brief Images data stored as a vector of pointers to NDArrays objects */
    std::unique_ptr<std::vector<std::unique_ptr<NDArray>>> pNDArrays = nullptr;
    /** @brief temporal field for storing data for getData() method (internally this clases uses smart pointer, externally data is
     * returned as standar pointers */
    std::vector<const NDArray*>* pNDArraysForGet = nullptr;
    /** @brief Pointer to vector of temporal dimensions of the stored group of images */ // default is empty
    std::unique_ptr<std::vector<dimIndexType>> pDynDims;
    /** @brief image spatial and temporal dimensions and their strides (field data type is valid for kernel parameters) */
    std::unique_ptr<std::vector<dimIndexType>> pDataDimsAndStridesVector;
private:
    static constexpr const char* errorPrefix = "OpenCLIPER::Data::";
    const std::string hostImageOrBufferToString(std::string title, dimIndexType index, SyncSource syncSource);
    /**
     * @brief 1 if all NDArrays have the same size for every spatial dimensions, 0 otherwise
     *
     * (bool type must not be used because is
     * not supported in OpenCL kernels, and this field has to be copied to data structures used by kernels)
     */
    dimIndexType allSizesEqual = 1;
    /// @brief Size (in bytes) of NDArray element
    size_t elementSize = 0;
    /// @brief Data type of NDArray element (see @ref ElementDataType)
    ElementDataType elementDataType = std::type_index(typeid(void));
    /// @brief format used for storing data in device memory (buffers, images or both), default value is buffers
    SyncSource syncSourceForLoading = SyncSource::BUFFER_ONLY;
    /** @brief pointer to CLapp object */
    std::shared_ptr<CLapp> pCLapp = nullptr;
    DataHandle myDataHandle = INVALIDDATAHANDLE;
    bool checkNDArrayIndex(dimIndexType index, dimIndexType nDArraySize);

    void commonFieldInitialization(ElementDataType elementDataType);
    void checkNDArraysSizesAndSetAllSizesEqual();
};
}
/* namespace OpenCLIPER */
#endif /* INCLUDE_OPENCLIPER_DATA_HPP_ */
