/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
// Avoiding compiler errors due to multiple include of header files
#ifndef PROCESSCORE_HPP
#define PROCESSCORE_HPP

#include <OpenCLIPER/OpenCLIPERDataModelCommonDefs.hpp>
#include <OpenCLIPER/Data.hpp>
// #include <OpenCLIPER/CLapp.hpp>
#include <LPISupport/SampleCollection.hpp>
#include <LPISupport/InfoItems.hpp>

#include <iostream> // std::cout, std::fixed
#include <iomanip> // std::setprecision

// Needed for kernels to find includes outside "kernels" subdirectory
#define KERNELCOMPILEOPTS "-I../include/"

namespace OpenCLIPER {

class CLapp;

/**
 * @brief Class representing common data and behaviour for all the processes in charge of operations on data.
 *
 */
class ProcessCore {
    public:
        ProcessCore();
        ProcessCore(std::shared_ptr<Data> pInputData, std::shared_ptr<Data> pOutputData);


        /// Destructor
        virtual ~ProcessCore();

        /// Structure with initialization parameters (can be redefined in the subclass, if needed)
        struct InitParameters {
            /// @brief Destructor
            virtual ~InitParameters() {}
        };

        /// Structure with launch parameters (can be redefined in the subclass, if needed)
        struct LaunchParameters {
            /// @brief Destructor
            virtual ~LaunchParameters() {}
        };

        /// Structure with profiling parameters (can be redefined in the subclass, if needed)
        struct ProfileParameters {
            /// If profiling is enabled (true)
            bool profilingEnabled = false;
            /// Number of kernel execution repetitions
            unsigned long numOfRepetitions = 1;
            /**
            * @brief Default constructor for struct fields initialization.
            *
            * @param profEnabled boolean true if profiling is enabled
            * @param numReps number of repetitions of kernel execution
            */
            ProfileParameters(bool profEnabled = false, unsigned long numReps = 1):
                profilingEnabled(profEnabled), numOfRepetitions(numReps) {}
        };

        /**
        * @brief Returns a SampleCollection object containing values of host+device execution times
        * @returns object of SampleCollection class
        */
        std::shared_ptr<LPISupport::SampleCollection> getSamplesGPU_CPUExecTime() {
            return pSamplesGPU_CPUExecutionTime;
        }

        /**
        * @brief Returns a SampleCollection object containing values of device execution times
        * @returns object of SampleCollection class
        */
        std::shared_ptr<LPISupport::SampleCollection> getSamplesGPUExecTime() {
            return pSamplesGPUExecutionTime;
        }

        /**
        * @brief Returns pointer to init parameters
        * @returns smart shared pointer to InitParameters object
        */
        virtual const std::shared_ptr<InitParameters> getInitParameters() const {
            return pInitParameters;
        }

        /**
        * @brief Returns pointer to launch parameters
        * @returns smart shared pointer to LaunchParameters object
        */
        virtual const std::shared_ptr<LaunchParameters> getLaunchParameters() const {
            return pLaunchParameters;
        }

        virtual void setApp(std::shared_ptr<CLapp> pCLapp) = 0;
        void setInput(std::shared_ptr<Data> pInputData);
        void setOutput(std::shared_ptr<Data> pOutputData);

        /**
        * @brief Sets initialization parameters for a Process object (this method can be redefined in a Process subclass).
        * @param[in] p structure with initialization parameters for a Process object
        */
        virtual void setInitParameters(const std::shared_ptr<InitParameters>& p) {
            pInitParameters=p;
        }

        /**
        * @brief Sets launch parameters for a Process object (this method can be redefined in a Process subclass).
        * @param[in] p structure with launch parameters of a Process object
        */
        virtual void setLaunchParameters(const std::shared_ptr<LaunchParameters>& p) {
            pLaunchParameters=p;
        }

        /**
        * @brief Method for Process object initialization (intialization is specific of Process subclasses).
        */
        virtual void init() {checkValidCLapp(getApp(), "ProcessCore::init");}

        /**
        * @brief Method that sets OpenCL kernel parameters and executes it (method specific of Process object subclasses)
        * @param[in] profileParameters parameters related to profiling of kernel execution
        */
        virtual void launch(ProfileParameters profileParameters = {false, 1}) = 0;

        /**
        * @brief Gets infoItems class variable value
        *
        * @return the value of infoItems class variable
        */
        LPISupport::InfoItems getInfoItems() {
        return infoItems;
        }


    protected:
        // Visible for all subclasses

        void checkValidCLapp(std::shared_ptr<OpenCLIPER::CLapp> pCLapp, std::string errorMsgPrefix);

        /**
        * @brief Get value of shared pointer to CLapp process assigned to this object (must be implemented by derived classes)
        * @return shared pointer to CLapp object
        */
        virtual const std::shared_ptr<CLapp> getApp() const = 0;

        std::shared_ptr<Data> getInput();

        /**
        * @brief Returns handle associated to input data
        * @return handle to input data
        */
        DataHandle getInHandle() {
            return inHandle;
        }

        std::shared_ptr<Data> getOutput();

        /**
        * @brief Returns handle associated to output data
        * @return handle to output data
        */
        DataHandle getOutHandle() {
            return outHandle;
        }

        void checkCommonLaunchParameters();
        void checkXDataLaunchParameters(SyncSource syncSource = SYNCSOURCEDEFAULT);
        void startProfiling(bool profilingEnabled);
        void stopProfiling(bool profilingEnabled);
        void startHostCodeProfiling(bool profilingEnabled);
        void stopHostCodeProfiling(bool profilingEnabled);
        void startKernelProfiling(bool profilingEnabled);
        void stopKernelProfiling(bool profilingEnabled);
        void buildKernelProfilingInfo(bool profilingEnabled);
        void getKernelGroupExecutionTimes(std::vector<cl::Event> eventList, std::string itemTitle, std::string totalsTitle);
        void addGlobalAndLocalWorkItemSizeInfo(cl::NDRange globalSizes, cl::NDRange localSizes, bool profilingEnabled);

        /// Smart shared pointer to init parameters
        std::shared_ptr<InitParameters> pInitParameters;
        /// Smart shared pointer to launch parameters
        std::shared_ptr<LaunchParameters> pLaunchParameters;
        /// OpenCL Command queue
        cl::CommandQueue queue;
        /// Boolean that shows if selected device supports kernel profiling
        bool profilingSupported = false;
        /// Kernel
        cl::Kernel kernel;
        /// Name of the kernel function to be executed
        std::string kernelName;
        /// Storage for error strings
        std::string errStr;
        /// Vector with InfoItems data (list of pairs title, value storing profiling information).
        LPISupport::InfoItems infoItems;
        /// Vector with values of GPU execution times
        std::shared_ptr<LPISupport::SampleCollection> pSamplesGPUExecutionTime;
        /// Vector with values of GPU+CPU execution times
        std::shared_ptr<LPISupport::SampleCollection> pSamplesGPU_CPUExecutionTime;
        /// Vector with events for profiling kernel execuion times
        std::vector<cl::Event> eventsVector;
    private:
        /// Smart shared pointer to Data object with input data
        std::shared_ptr<Data> pInputData = nullptr;
        /// Handle associated to input data
        DataHandle inHandle = INVALIDDATAHANDLE;
        /// Smart shared pointer to Data object with output data
        std::shared_ptr<Data> pOutputData = nullptr;
        /// Handle associated to output data
        DataHandle outHandle = INVALIDDATAHANDLE;
        /// Event for kernel execution start (used for profiling kernel execution times).
        cl::Event start_ev;
        /// Event for kernel execution stop (used for profiling kernel execution times).
        cl::Event stop_ev;
        /// clock at CPU starting execution of host Process code
        std::chrono::high_resolution_clock::time_point beginCPUExecTime;
        /// clock at CPU ending execution of host Process code
        std::chrono::high_resolution_clock::time_point endCPUExecTime;
};
} /* namespace OpenCLIPER */
#endif // PROCESSCORE_HPP
