/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */

#ifndef XDATA_HPP
#define XDATA_HPP

#include <OpenCLIPER/OpenCLIPERDataModelCommonDefs.hpp>
#include <OpenCLIPER/Data.hpp>
#include <OpenCLIPER/KData.hpp>
#include <OpenCLIPER/CLapp.hpp>

#include <assert.h>

namespace OpenCLIPER {
/**
 * @brief Class for storing data of images in the x-space.
 *
 * This class can be used to store a group of related images in the x-space, every image captured in a different time frame.
 * Load and save formats include standard image formats supported by DevIL library (png, jpeg, gif, tiff, etc., only one image per file),
 * matlab .mat format (one or several images per file) and OpenCLIPER raw format.
 *
 * OpenCLIPER raw format is a binary format where complex numbers of the image data are stored as a pair of real type numbers (first real
 * part, then imaginary part) using C language convention. The order for storing elements of the image data array is by columns (first columns of
 * the same row, then rows, slices and temporal dimensions by definition order).
 */
class XData: public Data {
public:

    /// @brief Name for variable inside matlab files (specific of XData class)
    static constexpr const char* matVarNameXData = "XData";

    /// Enumeration with the options for generating automatically data values for an XData object
    enum TypeOfGenData {
        /// Constant values
        CONSTANT,
        /// Consecutive values
        SEQUENTIAL
    };

    //*********************************
    // Constructors and destructor
    //*********************************

    // Empty data
    XData(std::shared_ptr<CLapp> pCLapp, ElementDataType elementDataType = std::type_index(typeid(complexType)));

    // From given dimensions (uninitialized)
    XData(std::shared_ptr<CLapp> pCLapp, dimIndexType width, ElementDataType elementDataType);
    XData(std::shared_ptr<CLapp> pCLapp, dimIndexType width, dimIndexType height, ElementDataType elementDataType);
    XData(std::shared_ptr<CLapp> pCLapp, dimIndexType width, dimIndexType height, dimIndexType depth, ElementDataType elementDataType);

    XData(std::shared_ptr<CLapp> pCLapp, std::vector<dimIndexType>* pArrayDims, ElementDataType elementDataType = std::type_index(typeid(complexType)));
    XData(std::shared_ptr<CLapp> pCLapp, std::vector<std::vector<dimIndexType>*>*& pArraysDims, std::vector<dimIndexType>*& pDynDims, ElementDataType elementDataType =
	std::type_index(typeid(complexType)));
    XData(std::shared_ptr<CLapp> pCLapp, std::vector<std::vector<dimIndexType>*>*& pArraysDims, std::vector<dimIndexType>*& pDynDims, std::vector<realType>*& pPixelSize,
	  ElementDataType elementDataType = std::type_index(typeid(complexType)));

    // From another Data or Data-derived object
    XData(std::shared_ptr<CLapp> pCLapp, const XData* sourceData, bool copyData = false);
    XData(std::shared_ptr<CLapp> pCLapp, std::shared_ptr<Data> sourceData, bool copyData = false);
    XData(std::shared_ptr<CLapp> pCLapp, std::shared_ptr<Data> sourceData, ElementDataType newElementDataType);
    XData(std::shared_ptr<CLapp> pCLapp, std::shared_ptr<XData> sourceData, bool copyData = false);
    XData(std::shared_ptr<CLapp> pCLapp, const KData* sourceData);
    XData(std::shared_ptr<CLapp> pCLapp, std::shared_ptr<KData> sourceData);

    // From one or more NDArrays
    XData(std::shared_ptr<CLapp> pCLapp, NDArray*& pData, ElementDataType elementDataType);
    XData(std::shared_ptr<CLapp> pCLapp, std::vector<NDArray*>*& pData, std::vector<dimIndexType>*& pDynDims, std::vector<realType>*& pPixelSize, ElementDataType elementDataType =
	std::type_index(typeid(complexType)));

    // From the filesystem
    //XData(std::shared_ptr<CLapp> pCLapp, const std::string matlabFileName);
    XData(std::shared_ptr<CLapp> pCLapp, const std::string fileName, ElementDataType elementDataType = std::type_index(typeid(complexType)));
    XData(std::shared_ptr<CLapp> pCLapp, const std::vector<std::string> fileNames, ElementDataType elementDataType = std::type_index(typeid(complexType)));
    XData(std::shared_ptr<CLapp> pCLapp, const std::string dataFileNamePrefix, std::vector<std::vector< dimIndexType >*>*& pArraysDims, std::vector <dimIndexType>*& pDynDims,
	const std::string framesFileNameSuffix="_frame", const std::string fileNameExtension=".raw", ElementDataType elementDataType = std::type_index(typeid(complexType)));

    virtual ~XData();

    // Inherit shared_ptr counter from base class
    std::shared_ptr<XData> shared_from_this() {
        return std::dynamic_pointer_cast<XData>(Data::shared_from_this());
    }

    // Getters
    /**
     * @brief Gets PixelSize class field.
     * @return pixel size vector
     */
    const std::vector<realType>* getPixelSize() const {
        return pPixelSize.get();
    }

    // Setters


    // Other methods
    void saveRawHostData(const SyncSource syncSource, const std::string fileNamePrefix,
                         const std::string framesFileNameSuffix="_frame",
                         const std::string fileNameExtension=".raw");
    void save(const std::string fileName);
    void save(const std::vector<std::string> fileNames);
    void save(const std::string prefixName, const std::string extension);

    void calcDataDims();
    // Save in matlab file
    void matlabSave(std::string fileName);
    static XData* genTestXData(std::shared_ptr<CLapp> pCLapp, dimIndexType width, dimIndexType height, dimIndexType numFrames, ElementDataType elementDataType, TypeOfGenData typeOfGenData = CONSTANT);
private:
    static constexpr const char* errorPrefix = "OpenCLIPER::XData::";
    void loadRawHostData(const std::string fileNamePrefix, std::vector<std::vector< dimIndexType >*>*& pArraysDims,
			 std::vector <dimIndexType>*& pDynDims,
                         const std::string framesFileNameSuffix="_frame", const std::string fileNameExtension=".raw");

    // Attributes
    /** @brief Pixel size vector (vector size equal to spatial dimensions vector size) */
    std::unique_ptr<std::vector<realType>> pPixelSize = nullptr;

    void load(const std::vector<std::string> fileNames);
    /** @brief Pointer to object containing dimensions get from matlab Dims variable */
    std::unique_ptr<MatVarDimsData> pMatVarDimsData = nullptr;
};
}
#endif // XDATA_HPP
