/**
 * @file
 * @brief File with macros and methods used by several OpenCL kernels and classes
 */

/*
 * hostKernelFunctions.h
 *
 *  Created on: 8 de nov. de 2017
 *      Author: fedsim
 */
#ifndef HOSTKERNELFUNCTIONS_H
#define HOSTKERNELFUNCTIONS_H

#include<OpenCLIPER/OpenCLIPERDataModelCommonDefs.hpp>

#ifdef __cplusplus
/// Macro used for disabling kernel-only code keyword global
#define global
/// Macro used for disabling kernel-only code keyword __global
#define __global
/// Macro for unsigned int data type name (different for kernel and non-kernel code)
#define uint cl_uint
#include <CL/cl.h>

extern "C" {
#else // Device code, C
// These functions are to be called from device code only
#define dimIndexType uint // dimIndexType defined as uint32_t only valid for C++ code
#endif // __CPLUSPLUS

//#define global

//#define DEBUG
//#define DEBUGKERNEL

#ifdef DEBUGKERNEL
/// Macro for printing debugging information
#define PRINTF(x) printf x
#else
/// Macro for printing debugging information
#define PRINTF(x) do {} while(0)
#endif // DEBUGKERNEL

// defines related to vector data types
// moved to OpenCLIPERDataModelCommonDefs.hpp because it used for checking validity of NDArray spatial dimensions
// (OpenCL vector operations are used for columns, rows or slices, so every spatial dimension has to be multiple of VECTORDATATYPESIZE/2, as
// every complex number is stored as 2 floats)
//
//#define VECTORDATATYPESIZE 16

/// Macro for defining the name of the vector type from its element data type and its number of elements
#define VECTORDATATYPEMACRO(baseType,size) {baseType ## size}
/// Data type used for vector kernel operations
#define VECTORDATATYPE float16
/// Half of the size of the vector data type defined previously
#define VECTORDATATYPEHALFSIZE (VECTORDATATYPESIZE)/2
/// Data type used for vector kernel operations on data with half size of the vector data type defined previously
#define HALFVECTORDATATYPE float8
/// OpenCL kernel operation for loading data from memory into a vector of the selected size
#define VLOADN vload16
/// OpenCL kernel operation for storing data into memory from a vector of the selected size
#define VSTOREN vstore16

/// Default kernel compile options
#define KERNELCOMPILEOPTS "-I../include/"

struct __attribute__ ((packed)) dimsAndStridesKnownFields {
    uint numSpatialDims;
    uint allSizesEqual;
    uint numCoils;
    uint numTemporalDims;
    uint firstTemporalDimSize;
};

/**
 * @brief Round up numToRound to multiple of baseNumber
 * @param[in] numToRound positive number to be rounded
 * @param[in] baseNumber value whose multiple nearest to numToRound is returned
 * @return rounded number (nearest multiple of baseNumber)
 */
uint roundUp(uint numToRound, uint baseNumber);

/**
 * @brief Return the offset (in bytes) from cl buffer start to first position of dimensions and strides array
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @return offset in bytes
 */
uint getDimsAndStridesArrayOffsetInBytes(global const void* buffer);

/**
 * @brief Return a uint pointer to the array of data dimensions and strides stored inside an OpenCL buffer
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @return pointer to the array of dimensions and strides
 */
global const uint* getDimsAndStridesArrayInBuffer(global const void* buffer);

/**
 * @brief Return a pointer to the struct of data dimensions and strides stored inside an OpenCL buffer
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @return pointer to the array of dimensions and strides
 */
global const struct dimsAndStridesKnownFields* getDimsAndStridesStruct(global const void* buffer);

/**
 * @brief Returns the number of spatial dimensions stored in dimensions info array
 *
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @return the number of spatial dimensions for the data
 */
uint getNumSpatialDims(global const void* buffer);

/**
 * @brief Returns the number of coils stored in dimensions info array
 *
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @return the number of coils for the data
 */
uint getNumCoils(global const void* buffer);


/**
 * @brief Returns the number of temporal dimensions stored in dimensions info array
 *
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @return the number of of temporal dimensions for the data
 */
uint getNumTemporalDims(global const void* buffer);

/**
 * @brief returns a pointer to the start of strides array (included in dimsAndStridesInfo)
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @return the pointer to the start of the strides information of the dimsAndStridesInfo array
 */
global const uint* getStridesArray(global const void* buffer);

/**
 * @brief Returns the size of a temporal dimension stored in dimsensions info array
 *
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @param[in] temporalDimIndex index of the temporal dimension
 * @return size of the specified temporal dimension
 */
uint getTemporalDimSize(global const void* buffer, uint temporalDimIndex);

/**
 * @brief Returns the size of a specific spatial dimension of a NDArray element (can be different for every NDArray)
 *
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @param[in] spatialDimIndex the spatial dimension to be get (enum knownSpatialDimPos values can be used for
 * this parameter)
 * @param[in] NDArray1DIndex 1D-index for the NDArray
 * @return the size of the selected spatial dimension and NDArray
 */
uint getSpatialDimSize(global const void* buffer, uint spatialDimIndex, uint NDArray1DIndex);

/**
 * @brief Returns the size of a specific dimension (be it spatial, coil or temporal) of a NDArray element (can be different for every NDArray)
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @param[in] dimIndex the dimension to get, spatial dimensions first, then coils, then temporal dimensions
 * @param[in] NDArray1DIndex 1D-index of the NDArray
 * @return the size of the selected dimension in the chosen NDArray
 */
uint getDimSize(global const void* buffer, uint dimIndex, uint NDArray1DIndex);

/**
 * @brief Returns the size (product of sizes of spatial dimensions) of a specific NDArray (can be different for every NDArray)
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @param[in] NDArray1DIndex 1D-index for the NDArray
 * @return the size of the selected NDArray
 */
uint getNDArrayTotalSize(global const void* buffer, uint NDArray1DIndex);

/**
 * @brief ...
 *
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @return cl_uint
 */
uint getNDArrayStride(global const void* buffer);

/**
 * @brief Returns the coil stride (number of real NDArray elements between the first NDArrays associated to two consecutive coils)
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @param[in] NDArray1DIndex 1D-index for the NDArray
 * @return the coil stride for the specific NDArray
 */
uint getCoilStride(global const void* buffer, uint NDArray1DIndex);

/**
 * @brief Returns the frame stride (number of real NDArray elements between the first NDArrays associated to two consecutive frames)
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @param[in] temporalDimIndex index of the selected frame dimension
 * @param[in] NDArray1DIndex 1D-index for the NDArray
 * @return the frame stride for the specific frame dimension and NDArray
 */
uint getTemporalDimStride(global const void* buffer, uint temporalDimIndex, uint NDArray1DIndex);

/**
 * @brief Returns the spatial stride (number of real elements between the first data element associated to two consecutive cols, rows, slices, etc.)
 * for a specific NDArray
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @param[in] spatialDimIndex the spatial dimension to be get (enum knownSpatialDimPos values can be used for
 * this parameter)
 * @param[in] NDArray1DIndex 1D-index for the NDArray
 * @return the spatial stride for the specific spatial dimension and NDArray
 */
uint getSpatialDimStride(global const void* buffer, uint spatialDimIndex, uint NDArray1DIndex);

/**
 * @brief Returns the stride for the first element of a NDArray object
 *
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @param[in] NDArray1DIndex 1-dimensional index for selected NDArray (obtained from temporal dimensions indexes and coil index)
 * @return the stride for the first element of a NDArray object
 */
uint getElementStride(global const void* buffer, uint NDArray1DIndex);

/**
 * @brief Returns the stride (number of real elements between the first data element associated with two consecutive cols, rows, slices, coils (if any), frames, etc.)
 * for a specific NDArray
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @param[in] dimIndex the spatial dimension to be get (enum knownSpatialDimPos values can be used for
 * this parameter)
 * @param[in] NDArray1DIndex 1D-index for the NDArray
 * @return the spatial stride for the specific spatial dimension and NDArray
 */
uint getDimStride(global const void* buffer, uint dimIndex, uint NDArray1DIndex);

/**
 * Calculates a 1-dimensional index for accessing a specific NDArray from the vector of NDArrays corresponding to the specified coil
 * index and the indexes for the temporal dimensions
 * @param[in] buffer OpenCL buffer storing data and their dimensions and strides array
 * @param[in] coilIndex index for the coils
 * @param[in] temporalDimIndexes vector of indexes for the selected frame
 * @return the 1-dimensional index for a NDArray
 */
uint getNDArray1DIndex(global const void* buffer, uint coilIndex, uint temporalDimIndexes[]);

global const void* getNextElement(global const void* buffer, dimIndexType nDim, dimIndexType curNDArray, dimIndexType curOffset);

// These functions are to be called from device code only
#ifdef __OPENCL_C_VERSION
void printComplex(float2 complex, char* name);
void printVector(__constant char name[], float *vector, uint numberOfElements);
#endif //__OPENCL_C_VERSION

#ifdef __cplusplus
    }  //extern "C"
#undef global
#undef __global
#undef uint
#endif // __CPLUSPLUS

#endif // HOSTKERNELFUNCTIONS_H
