/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel RodrÃ­guez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena MartÃ­n GonzÃ¡lez,
 *                    Elisa Moya SÃ¡ez,
 *                    Marcos MartÃ­n FernÃ¡ndez and
 *                    Carlos Alberola LÃ³pez
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. TelecomunicaciÃ³n
 *  Universidad de Valladolid
 *  Paseo de BelÃ©n 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
#include <OpenCLIPER/OpenCLIPERDataModel.hpp>
#include <OpenCLIPER/PerfTestConfResult.hpp>

// Uncomment to show specific debug messages
//#define MRIRECONEND2ENDPERFTEST_DEBUG

#if !defined NDEBUG && defined MRIRECONEND2ENDPERFTEST_DEBUG
#define MRIRECONEND2ENDPERFTEST_CERR(x) CERR(x)
#else
#define MRIRECONEND2ENDPERFTEST_CERR(x)
#undef MRIRECONEND2ENDPERFTEST_DEBUG
#endif


using namespace OpenCLIPER;
int main(int argc, char *argv[]) {
    std::shared_ptr<LPISupport::SampleCollection> pSamples = std::make_shared<LPISupport::SampleCollection>("execution time");
    try{
        CLapp::PlatformTraits platformTraits;
        CLapp::DeviceTraits deviceTraits;
        PerfTestConfResult* pPerfTest = new PerfTestConfResult(argc, argv);
        auto pConfigTraits = std::dynamic_pointer_cast<PerfTestConfResult::ConfigTraits>(pPerfTest->getConfigTraits());
        unsigned int numberOfIterations = pConfigTraits->repetitions;
        std::cerr << "Number of iterations: " << numberOfIterations << "\n";

        deviceTraits = pConfigTraits->deviceTraits;
        platformTraits = pConfigTraits->platformTraits;
        // CL_QUEUE_PROFILING_ENABLE slows down the queue. Beware!
        // deviceTraits.queueProperties = cl::QueueProperties(CL_QUEUE_PROFILING_ENABLE);
        auto pCLapp=CLapp::create(platformTraits, deviceTraits);

        // Load OpenCL kernel(s)
        std::vector<std::string> kernelFileNames = {"complexElementProd.cl", "xImageSum.cl", "blankDataRows_kernel.cl",
            "creabspline.cl", "initialize.cl", "optimizer.cl", "reductions.cl",
            "temporalTV.cl", "vectorNormalization_kernel.cl"};
            pCLapp->loadKernels(kernelFileNames);

            // Load input data from Matlab file
            std::string filename = "../../tests/matlabKImages/data_160x160_coils23_frames20_single.mat";
            std::shared_ptr<Data> pInputKData = std::make_shared<KData>(pCLapp, filename);
            std::shared_ptr<XData> pOutputXData;
            for (unsigned int i = 0; i < numberOfIterations; i++) {
                // Create output with suitable size
                pOutputXData = nullptr;
                pOutputXData = std::make_shared<XData>(pCLapp, std::dynamic_pointer_cast<KData>(pInputKData));

                // Parameters for NESTA execution
                // -------------------------------------------------------------------------------------------------
                float lambda_i = 3E-2;	//3E-2
                float mu_f = 1E-4;	//1E-10
                float La = 1;
                uint maxIntIter = 4; 	// 6
                float tolVar = 1E-3;	//1E-4
                int verbose = 10;
                int maxIter = 3000;	// 3000 // 300
                uint stoptest = 1;

                // Parameters for GROUPWISE REGISTRATION execution
                // -------------------------------------------------------------------------------------------------
                uint MOTION_ITERS = 3; 	// Number of iterations to perform GroupwiseRegistration + Nesta
                uint nmax = 100; 	// Maximum number of iterations
                float et = 0.001f; 	// Transformation norm threshold
                float eh = 0.001f; 	// Metric variation threshold
                std::vector<float> lambda = {1.f,6.f,1.f,5.f}; // Smoothnes/Regularization weights (1st spatial, 2nd spatial, 1st temp, 2nd temp)
                int radius = 30; 	// Radius of the circular ROI (0 if 'wholebox')
                int E = 3; 		// E: spline order
                int Dp[2] = {4,4}; 	// Dp: point density
                float W = 1.f; 		// Weight for each point displacement
                bool flagW = true; 	// Flag for adaptative set of W



                // Create new process, set its input/output data sets and bind it to our CL app
                // Reuse process NestaUp
                std::unique_ptr<Process> nestaUpProcess(new NestaUp(pCLapp));
                nestaUpProcess->setInput(pInputKData);//setInHandle(inHandle);
                nestaUpProcess->setOutput(pOutputXData);//setOutHandle(outHandle);
                nestaUpProcess->init();

                // Reuse process GroupwiseRegistration
                //DataHandle TDataHandle = INVALIDDATAHANDLE;
                std::shared_ptr<Data> TData;
                Bspline* ts = new Bspline();

                std::unique_ptr<Process> GWRegistrationProcess(new GroupwiseRegistration(pCLapp));
                GWRegistrationProcess->setInput(pOutputXData); // Only needs input
                GWRegistrationProcess->init();
                std::shared_ptr<GroupwiseRegistration::LaunchParameters> paramsGW;

		// Start measuring execution time now
		std::chrono::high_resolution_clock::time_point beginReconExecTime = std::chrono::high_resolution_clock::now();

                ///////////////////////////////////////////////////////////////////////////////////////////
                // NESTA - First reconstruction                                                          //
                ///////////////////////////////////////////////////////////////////////////////////////////

                auto paramsNestaUp = std::make_shared<NestaUp::LaunchParameters>(lambda_i, mu_f, La, maxIntIter, tolVar, verbose, maxIter, stoptest, nullptr);
                nestaUpProcess->setLaunchParameters(paramsNestaUp);
                nestaUpProcess->launch();

                //pOutputXData->device2Host(SyncSource::BUFFER_ONLY);
                #if !defined NDEBUG && defined MRIRECONEND2ENDPERFTEST_DEBUG
		    (std::dynamic_pointer_cast<XData>(pOutputXData))->matlabSave("../../output/NESTA_nomotion_recon.mat", SyncSource::BUFFER_ONLY);
                #endif

                for(unsigned int i=0; i<MOTION_ITERS; i++){

                    ///////////////////////////////////////////////////////////////////////////////////////////
                    // Groupwise Registration to estimate cardiac motion                                     //
                    ///////////////////////////////////////////////////////////////////////////////////////////
                    ArgumentsMotionCompensation *argsMC = new ArgumentsMotionCompensation();
                    paramsGW = std::make_shared<GroupwiseRegistration::LaunchParameters>(argsMC, nmax, et, eh, lambda, radius, E, Dp, W, flagW, TData, ts, i);
                    GWRegistrationProcess->setLaunchParameters(paramsGW);
                    GWRegistrationProcess->launch();
                    TData = argsMC->TData; // Reuse T data obtained in current iteration as initialization for next iteration

                    #if !defined NDEBUG && defined MRIRECONEND2ENDPERFTEST_DEBUG
			std::shared_ptr<XData> pOutputRegData = std::make_shared<XData>(pCLapp, pOutputXData);
			std::unique_ptr<Process> MCProcess(new MotionCompensation(pCLapp));
			MCProcess->setLaunchParameters(std::make_shared<MotionCompensation::LaunchParameters>(argsMC));
			MCProcess->setInput(pOutputXData);
			MCProcess->setOutput(pOutputRegData);
			MCProcess->launch();
			MCProcess.reset(nullptr);
			//pOutputRegData->device2Host(SyncSource::BUFFER_ONLY);
			pOutputRegData->matlabSave("../../output/registration_"+ std::to_string(i+1) + "recon.mat", SyncSource::BUFFER_ONLY);
			pOutputRegData = nullptr;
                    #endif

                    ///////////////////////////////////////////////////////////////////////////////////////////
                    // NESTA - Reconstruction with motion compensation                                       //
                    ///////////////////////////////////////////////////////////////////////////////////////////
                    paramsNestaUp = std::make_shared<NestaUp::LaunchParameters>(lambda_i, mu_f, La, maxIntIter, tolVar, verbose, maxIter, stoptest, argsMC);
                    nestaUpProcess->setLaunchParameters(paramsNestaUp);
                    nestaUpProcess->launch();
                    delete argsMC;

                    #if !defined NDEBUG && defined MRIRECONEND2ENDPERFTEST_DEBUG
			//pOutputXData->device2Host(SyncSource::BUFFER_ONLY); // automatically called by matlabSave
			pOutputXData->matlabSave("../../output/NESTA_motion"+ std::to_string(i+1) + "recon.mat", SyncSource::BUFFER_ONLY);
                    #endif
                }

                delete ts;

		// Execution time measurement ends now
                std::chrono::high_resolution_clock::time_point endReconExecTime = std::chrono::high_resolution_clock::now();
                TIME_DIFF_TYPE elapsedTimeRecon = (std::chrono::duration_cast<std::chrono::nanoseconds>(endReconExecTime - beginReconExecTime).count()) / 1e9;

		std::ostringstream ostream2;
                ostream2 << std::fixed << std::setprecision(PROFILINGTIMESPRECISION) << elapsedTimeRecon;
                std::cout << " --> Time elapsed in reconstruction end to end (s): " << ostream2.str() << "\n" << std::endl;
                pSamples->appendSample(elapsedTimeRecon);
                GWRegistrationProcess.reset(nullptr);
                nestaUpProcess.reset(nullptr);
                TData = nullptr;
            }
            pPerfTest->buildTestInfo(pSamples, &pCLapp);
            pPerfTest->saveOrPrint();

            // Show reconstructed data
            pOutputXData->show();

            // Save reconstructed data
            pOutputXData->matlabSave("MRIReconEnd2EndResult.mat");

            pInputKData = nullptr;
            pOutputXData = nullptr;
    }
    catch(cl::BuildError& e) {
        CLapp::dumpBuildError(e);
    }
    catch(cl::Error& e) {
        std::cerr << "CL Exception "<<e.err()<<" ("<<CLapp::getOpenCLErrorCodeStr(e.err())<<"): " << e.what() << std::endl;
    }
    catch (std::exception& e) {
        std::cerr << "Error: " << e.what() << "\tfile: " << __FILE__ << "\tline: " << __LINE__  << std::endl;
    }
}

#undef MRIRECONEND2ENDPERFTEST_DEBUG
