// Extraído y adaptado de
// http://developer.amd.com/tools-and-sdks/opencl-zone/opencl-resources/introductory-tutorial-to-opencl/

#include "vectorUtils.hpp"
#include "commonArrayMult.hpp"
#include <OpenCLIPER/processes/performanceTests/ArrayMultProcess.hpp>
#include <cstdio>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <string>
#include <iterator>
#include <vector>
#include <array>
#include <chrono> // Para medir tiempos de ejecución
using namespace std;
using namespace std::chrono;
#include <iomanip> // Para std::setprecision
#include <OpenCLIPER/OpenCLIPERDataModel.hpp>
#include "PerformanceTestArrayOpParallel.hpp"
#include <utility>
#include <omp.h>
//#define __NO_STD_VECTOR // Use cl::vector instead of STL version

int main (int argc, char* argv[]) {

    unsigned long numberOfIterations = 1;
    unsigned int size = 2048, blockSize = 0;
    unsigned int RowsA;
    unsigned int ColsA;
    unsigned int RowsB;
    unsigned int ColsB;
    unsigned int RowsC;
    unsigned int ColsC;
    std::shared_ptr<LPISupport::SampleCollection> pSamples = std::make_shared<LPISupport::SampleCollection>("execution time");

    try {
        PerformanceTestArrayOpParallel* pPerfTest = new PerformanceTestArrayOpParallel(argc, argv);
        auto pConfigTraits = std::dynamic_pointer_cast<PerformanceTestArrayOpParallel::ConfigTraits>(pPerfTest->getConfigTraits());
        size = pConfigTraits->size;
        numberOfIterations = pConfigTraits->repetitions;
        blockSize = pConfigTraits->dimBlockOrLocalSize;
        std::cout << "read size: " << size << std::endl;
        std::cout << "read blockSize: " << blockSize << std::endl;
        RowsA = size;
        ColsA = size;
        RowsB = ColsA;
        ColsB = size;
        RowsC = RowsA;
        ColsC = ColsB;

        // Number of simple calculations for array multiplication is
        // ColsA products of two float numbers plus ColsA-1 sums of two float numbers
        // for every element of output array (which has RowsC*ColsC elements)
        // IMPORTANT!!: at least one expression operand must be casted to unsigned long type for the result being unsigned long;
        // otherwise, in spite of operand values being less than integer type maximum value, the result may exceed unsigned int maximum value and be truncated
        pConfigTraits->numOpsPerCalc = ((unsigned long) (RowsC)) * ColsC * (ColsA*2-1);
        cerr << "Number of operations: " << ((unsigned long)RowsC*ColsC * (ColsA * 2 - 1)) << std::endl;
        cerr << "Number of operations stored in pConfigTraits->numOpsPerCalc: " << pConfigTraits->numOpsPerCalc << std::endl;

	// Step 1: get a new OpenCLIPER app and initialize computing device
        CLapp::PlatformTraits platformTraits;
        CLapp::DeviceTraits deviceTraits;
#ifdef USE_GPU
        deviceTraits.type=CLapp::DEVICE_TYPE_GPU;
        if (!pConfigTraits->deviceName.empty()) {
            deviceTraits.name = pConfigTraits->deviceName;
        }
#else
        deviceTraits.type=CLapp::DEVICE_TYPE_CPU;
#endif
        deviceTraits.queueProperties = cl::QueueProperties(CL_QUEUE_PROFILING_ENABLE);
        auto pCLapp=CLapp::create(platformTraits,deviceTraits);

        // Step 2: load OpenCL kernel(s)
        pCLapp->loadKernels("performanceTests/arrayMult.cl");

        const unsigned int SHOW_SIZE = 10;
        const unsigned int PRECISION_DIGITS = 10;

        cerr << argv[0] << " performance measurement" << std::endl;
        cout << "Starting program... " << flush;

        // Step 3: load input data
        cout << "Creating and filling arrays ... " << flush;
        std::shared_ptr<Data> XDataA(pCLapp, XData::genTestXData(pCLapp, ColsA, RowsA, 1, type_index(typeid(realType)), XData::CONSTANT));
        std::shared_ptr<Data> XDataB(pCLapp, XData::genTestXData(pCLapp, ColsB, RowsB, 1, type_index(typeid(realType)), XData::CONSTANT));

        // Step 4: create output with same size as input
        std::vector<dimIndexType>* pArrayDims = new std::vector<dimIndexType>({ColsC, RowsC});
        std::vector<std::vector<dimIndexType>*>* pArraysDims = new std::vector<std::vector<dimIndexType>*>;
        pArraysDims->push_back(pArrayDims);
        std::vector<dimIndexType>* pDynDims = new std::vector<dimIndexType>(); // Constructor value only sets vector size not value, that is 0 by default
        pDynDims->push_back(1); // Sets value
        std::shared_ptr<Data> XDataC(pCLapp, new XData(pCLapp, pArraysDims, pDynDims, type_index(typeid(realType))));
        //std::shared_ptr<Data> XDataC(new XData((dynamic_pointer_cast<XData>(XDataA)), false));

        // Set 5: register input and output in our CL app
        // automatic
        cout << "Done." << endl;

        /*
        cout << XDataA->getData()->at(0)->hostDataToString("a");
        cout << XDataB->getData()->at(0)->hostDataToString("b");
        cout << endl;
        */

        // Step 6: create new process bound to our CL app
        // and set its input/output data sets
        std::unique_ptr<Process> pProcess(new ArrayMultProcess(pCLapp));
        pProcess->setInput(XDataA);
        pProcess->setOutput(XDataC);

        // Set parameters: handle of second array to be multiplied
        auto launchParamsArrayMultProcess = make_shared<ArrayMultProcess::LaunchParameters>(XDataB->getHandle(), RowsA, ColsA, ColsB, blockSize);
        pProcess->setLaunchParameters(launchParamsArrayMultProcess);

        // Step 7: initialize process
        pProcess->init();

        cerr << "Executing " << numberOfIterations << " iteration(s)\n";
        cerr << "- Matrix [1 .. " << size << ", 1 .. " << size << "]" << endl;
        LPISupport::InfoItems infoItems;
        cerr << pCLapp->getHWSWInfo().to_string(pConfigTraits->outputFormat);
        //TIME_DIFF_TYPE diffT2T1 = 0;
        std::stringstream strstr;
        strstr << setprecision(PRECISION_DIGITS);
        cout << "Starting product... " << endl;
        /*
        for (unsigned long iteration = 0; iteration < numberOfIterations; iteration++) {
            cout << "Iteration #" << iteration << std::endl;
            BEGIN_TIME(t1);
        */
        // Step 7.2 launch process
            Process::ProfileParameters profileParameters;
            // Enable gpu profiling
            profileParameters.profilingEnabled = true;
            profileParameters.numOfRepetitions = numberOfIterations;
            pProcess->launch(profileParameters);

        /*
            END_TIME(t2);
            TIME_DIFF(diffT2T1, t1, t2);
            pSamples->appendSample(diffT2T1);
        }
        */
            // Step 8: get data back from computing device
            pCLapp->device2Host(XDataC, SyncSource::BUFFER_ONLY);

        cout << "Product finished." << endl;
        if (size <= SHOW_SIZE) {
            cout << XDataC->hostBufferToString("c",0);
        }

        LPISupport::InfoItems infoItemsProfilingGPU;
        infoItemsProfilingGPU.append(pProcess->getSamplesGPU_CPUExecTime()->to_infoItems(PRECISION_DIGITS));
        /*
         for (unsigned int i = 0; i < infoItemsProfilingGPU.size(); i++) {
            infoItems.push_back(infoItemsProfilingGPU.at(i));
        }
        */
        cerr << infoItemsProfilingGPU.to_string(pConfigTraits->outputFormat);
        pConfigTraits->deviceType = pCLapp->getDeviceTypeAsString();
        pConfigTraits->deviceName = pCLapp->getDeviceVendor() +  " " + pCLapp->getDeviceName();
#if USE_GPU
        pPerfTest->buildTestInfo(pProcess->getSamplesGPUExecTime());
#else
        //pPerfTest->buildTestInfo(pProcess->getSamplesGPU_CPUExecTime());
        pPerfTest->buildTestInfo(pProcess->getSamplesGPUExecTime());
#endif
        pPerfTest->saveOrPrint();

        // Step 10: clean up
        pProcess.reset(nullptr);
        XDataA = nullptr;
        XDataB = nullptr;
        XDataC = nullptr;
        pCLapp = nullptr;
        pSamples = nullptr;
    } catch(cl::BuildError& e) {
    	CLapp::dumpBuildError(e);
    }
    catch(cl::Error& e) {
        std::cerr << CLapp::getOpenCLErrorInfoStr(e, __FILE__, __LINE__);
    }
    catch (std::exception& e) {
    	std::cerr << "Generic exception: " << e.what() << std::endl;
    }
}
