#include<fstream>
#include<sstream>
#include<string>
#include <OpenCLIPER/Data.hpp>
#include <OpenCLIPER/XData.hpp>
#include <OpenCLIPER/CLapp.hpp>
#include <OpenCLIPER/processes/Complex2Real.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/imgproc.hpp>
#include <OpenCLIPER/MatVarDimsData.hpp>

// Uncomment to show class-specific debug messages
//#define DATA_DEBUG

#if !defined NDEBUG && defined DATA_DEBUG
#define DATA_CERR(x) CERR(x)
#else
#define DATA_CERR(x)
#undef DATA_DEBUG
#endif

namespace OpenCLIPER {
/**
 * @brief Constructor that creates an empty Data object.
 * @param[in] elementDataType data type of base element
 */
Data::Data(ElementDataType elementDataType) {
	commonFieldInitialization(elementDataType);
	pNDArraysForGet = new std::vector<const NDArray*>;
	pNDArrays = nullptr;
}

/**
 * @brief Constructor that creates a Data object from a vector of NDArrays.
 *
 * All parameters of type *& (reference to pointer) have move semantics: ownership of memory is moved from parameter to this
 * object, and parameter value is set to nullptr after method completion. Temporal dimensions vector is initialized to a 1D vector
 * with a value equal to the number of NDArrays.
 * @param[in,out] pNDArrays pointer to vector of pointers to NDArray objects
 * @param[in] elementDataType data type of base element
 */
Data::Data(std::vector<NDArray*>*& pNDArrays, ElementDataType elementDataType) {
	commonFieldInitialization(elementDataType);
	// As there is no explicit pDynDims vector, we create a 1D-vector with a value
	// equal to the number of NDArrays
	// Error: SenstivityMaps (subclass of Data) does not have temporal dimensions (number of temporal dimensions must be 0)
	//std::vector<dimIndexType>* pDynDims = new std::vector<dimIndexType>({pNDArrays->size()});
	std::vector<dimIndexType>* pDynDims = new std::vector<dimIndexType>();
	createFromNDArraysVector(pNDArrays, pDynDims);
}

/**
 * @brief Constructor that creates a Data object from a vector of NDArrays and a vector with temporal dimensions (see @ref
 * createFromNDArraysVector).
 *
 * All parameters of type *& (reference to pointer) have move semantics: ownership of memory is moved from parameter to this
 * object, and parameter value is set to nullptr after method completion.
 * @param[in,out] pNDArrays pointer to vector of pointers to NDArray objects
 * @param[in,out] pDynDims pointer to vector of temporal dimensions
 * @param[in] elementDataType data type of base element
 */
Data::Data(std::vector<NDArray*>*& pNDArrays, std::vector<dimIndexType>*& pDynDims, ElementDataType elementDataType) {
	commonFieldInitialization(elementDataType);
	createFromNDArraysVector(pNDArrays, pDynDims);
}

/**
 * @brief Creates a Data object from a vector of NDArrays and a vector of temporal dimensions.
 *
 * All parameters of type *& (reference to pointer) have move semantics: ownership of memory is moved from parameter to this
 * object, and parameter value is set to nullptr after method completion.
 * @param[in,out] pNDArrays pointer to vector of pointers to NDArray objects
 * @param[in,out] pDynDims pointer to vector of temporal dimensions
 */
void Data::createFromNDArraysVector(std::vector<NDArray*>*& pNDArrays, std::vector<dimIndexType>*& pDynDims) {
	constexpr const char* methodErrorPrefix = "createFromNDArraysVector: ";
	std::ostringstream outputstream;
	outputstream << errorPrefix << methodErrorPrefix;
	pNDArraysForGet = new std::vector<const NDArray*>;
	setDynDims(pDynDims);
	internalSetData(pNDArrays);
}

/**
 * @brief Constructor that creates an uninitialized Data object containing a single 1-dimensional NDArray with given spatial dimensions and no temporal dimensions.
 *
 * @param[in] width length of the NDArray
 * @param[in] elementDataType data type of base element
 */
Data::Data(dimIndexType width, ElementDataType elementDataType) {
	commonFieldInitialization(elementDataType);
	pNDArraysForGet = new std::vector<const NDArray*>;

	auto pArrayDims = new std::vector<dimIndexType>(1, width);
	auto pArraysDims = new std::vector<std::vector<dimIndexType>*>(1, pArrayDims);
	auto pDynDims = new std::vector<dimIndexType>();
	createEmptyDataWithDimensions(pArraysDims, pDynDims);
}

/**
 * @brief Constructor that creates an uninitialized Data object containing a single 2-dimensional NDArray with given spatial dimensions and no temporal dimensions.
 *
 * @param[in] width length of the NDArray along the first dimension
 * @param[in] height length of the NDArray along the second dimension
 * @param[in] elementDataType data type of base element
 */
Data::Data(dimIndexType width, dimIndexType height, ElementDataType elementDataType) {
	commonFieldInitialization(elementDataType);
	pNDArraysForGet = new std::vector<const NDArray*>;

	auto pArrayDims = new std::vector<dimIndexType>(2);
	*pArrayDims = {width, height};
	auto pArraysDims = new std::vector<std::vector<dimIndexType>*>(1, pArrayDims);
	auto pDynDims = new std::vector<dimIndexType>();
	createEmptyDataWithDimensions(pArraysDims, pDynDims);
}

/**
 * @brief Constructor that creates an uninitialized Data object containing a single 3-dimensional NDArray with given spatial dimensions and no temporal dimensions.
 *
 * @param[in] width length of the NDArray along the first dimension
 * @param[in] height length of the NDArray along the second dimension
 * @param[in] depth length of the NDArray along the third dimension
 * @param[in] elementDataType data type of base element
 */
Data::Data(dimIndexType width, dimIndexType height, dimIndexType depth, ElementDataType elementDataType) {
	commonFieldInitialization(elementDataType);
	pNDArraysForGet = new std::vector<const NDArray*>;

	auto pArrayDims = new std::vector<dimIndexType>(3);
	*pArrayDims = {width, height, depth};
	auto pArraysDims = new std::vector<std::vector<dimIndexType>*>(1, pArrayDims);
	auto pDynDims = new std::vector<dimIndexType>();
	createEmptyDataWithDimensions(pArraysDims, pDynDims);
}

/**
 * @brief Constructor that creates an uninitialized Data object containing a single NDArray with given spatial dimensions and no temporal dimensions.
 *
 * @param[in,out] pArraysDims pointer to vector of pointers to spatial dimensions vectors
 * @param[in] elementDataType data type of base element
 */
Data::Data(std::vector<dimIndexType>* pArrayDims, ElementDataType elementDataType) {
	commonFieldInitialization(elementDataType);
	pNDArraysForGet = new std::vector<const NDArray*>;

	auto pArraysDims = new std::vector<std::vector<dimIndexType>*>(1, pArrayDims);
	auto pDynDims = new std::vector<dimIndexType>();
	createEmptyDataWithDimensions(pArraysDims, pDynDims);
}

/**
 * @brief Constructor that creates an uninitialized Data object with given spatial dimensions and no temporal dimensions.
 *
 * All parameters of type *& (reference to pointer) have move semantics: ownership of memory is moved from parameter to this
 * object, and parameter value is set to nullptr after method completion. Temporal dimensions vector is initialized to a 1D
 * vector with a value equal to the number of spatial dimensions vectors.
 * @param[in,out] pArraysDims pointer to vector of pointers to spatial dimensions vectors
 * @param[in] elementDataType data type of base element
 */
Data::Data(std::vector<std::vector<dimIndexType>*>*& pArraysDims, ElementDataType elementDataType) {
	commonFieldInitialization(elementDataType);
	pNDArraysForGet = new std::vector<const NDArray*>;

	// As there is no explicit pDynDims vector, we create a 1D-vector with a value
	// equal to the number of NDArrays
	// Error: this pDynDims data is not valid for SensitivityMapsData (several NDArrays, 1 per number of coils, but number of
	// temporal dimensions is 0
	std::vector<dimIndexType>* pDynDims = new std::vector<dimIndexType>();
	createEmptyDataWithDimensions(pArraysDims, pDynDims);
}

/**
 * @brief Constructor that creates an uninitialized Data object with given spatial and temporal dimensions.
 *
 * All parameters of type *& (reference to pointer) have move semantics: ownership of memory is moved from parameter to this
 * object, and parameter value is set to nullptr after method completion.
 * @param[in,out] pArraysDims pointer to vector of pointers to spatial dimensions vectors
 * @param[in,out] pDynDims pointer to vector of temporal dimensions
 * @param[in] elementDataType data type of base element
 */
Data::Data(std::vector<std::vector<dimIndexType>*>*& pArraysDims, std::vector<dimIndexType>*& pDynDims,
		ElementDataType elementDataType) {
	commonFieldInitialization(elementDataType);
	createEmptyDataWithDimensions(pArraysDims, pDynDims);
}

/**
 * @brief Constructor that creates a Data object with spatial and temporal dimensions set, dimensions are got from
 * another Data object.
 *
 * @param[in] sourceData Data object source of data, spatial and temporal dimensions
 * @param[in] copyData if true also data (not only dimensions) are copied from NDArray sourceData object to this object
 */
Data::Data(const Data* sourceData, bool copyData) {
	commonFieldInitialization(sourceData->getElementDataType());
	pNDArraysForGet = new std::vector<const NDArray*>;
	std::vector<NDArray*>* pLocalData = new std::vector<NDArray*>;
	NDArray* pLocalNDArray;
	std::vector<dimIndexType>* pLocalDynDims = new std::vector<dimIndexType>(*(sourceData->getDynDims()));
	setDynDims(pLocalDynDims);
	for (dimIndexType i = 0; i <  sourceData->getData()->size(); i++) {
		pLocalNDArray = NDArray::createNDArray(sourceData->getData()->at(i), copyData, elementDataType);
		pLocalData->push_back(pLocalNDArray);
	}
	setData(pLocalData);
}

/**
 * @brief Constructor that creates a Data object with spatial and temporal dimensions set, from another Data object.
 * Data element type is the same as in the source Data object.
 *
 * @param[in] sourceData Data object source of data, spatial and temporal dimensions
 * @param[in] copyData if true also data (not only dimensions) are copied from NDArray sourceData object to this object
 */
Data::Data(std::shared_ptr<Data> sourceData, bool copyData) {
	commonFieldInitialization(sourceData->getElementDataType());
	pNDArraysForGet = new std::vector<const NDArray*>;
	std::vector<NDArray*>* pLocalData = new std::vector<NDArray*>;
	NDArray* pLocalNDArray;
	std::vector<dimIndexType>* pLocalDynDims = new std::vector<dimIndexType>(*(sourceData->getDynDims()));
	setDynDims(pLocalDynDims);
	for (dimIndexType i = 0; i <  sourceData->getData()->size(); i++) {
		pLocalNDArray = NDArray::createNDArray(sourceData->getData()->at(i), copyData, elementDataType);
		pLocalData->push_back(pLocalNDArray);
	}
	setData(pLocalData);
}

/**
 * @brief Constructor that creates an empty Data object with spatial and temporal dimensions set, from another Data object.
 * Data element type is specified by the caller.
 *
 * @param[in] sourceData Data object source of data, spatial and temporal dimensions
 * @param[in] newElementDataType Data type of the newly created object
 */
Data::Data(std::shared_ptr<Data> sourceData, ElementDataType newElementDataType) {
	commonFieldInitialization(newElementDataType);
	pNDArraysForGet = new std::vector<const NDArray*>;
	std::vector<NDArray*>* pLocalData = new std::vector<NDArray*>;
	NDArray* pLocalNDArray;
	std::vector<dimIndexType>* pLocalDynDims = new std::vector<dimIndexType>(*(sourceData->getDynDims()));
	setDynDims(pLocalDynDims);
	for (dimIndexType i = 0; i < sourceData->getData()->size(); i++) {
		pLocalNDArray = NDArray::createNDArray(sourceData->getData()->at(i), false, elementDataType);
		pLocalData->push_back(pLocalNDArray);
	}
	setData(pLocalData);
}


/**
 * @brief Load contents of matlab variable into a Data object (containing a group of NDArray objects)
 * @param[in] matvar matlab array variable previously read from file
 * @param[in] numOfSpatialDimensions number of spatial dimensions in matlab variable (rest of dimensions can be temporal dimensions or
 * number of coils)
 * @param[in] numNDArraysToRead number of NDArrays to be read from matlab variable
 * @throw std::invalid_argument if element datatype is not supported
 */
void Data::loadMatlabHostData(matvar_t* matvar, dimIndexType numOfSpatialDimensions, dimIndexType numNDArraysToRead) {
#ifdef DATA_DEBUG
	BEGIN_TIME(bTReadingKDataMatVar);
	DATA_CERR("  reading data from matlab variable... ");
#endif
	switch(matvar->data_type) {
	case MAT_T_SINGLE:
		if (matvar->isComplex) {
			elementDataType = std::type_index(typeid(complexType));
		} else {
			elementDataType = std::type_index(typeid(realType));
		}
		break;
	case MAT_T_UINT32:
		if(!matvar->isComplex) {
			elementDataType = std::type_index(typeid(dimIndexType));
		} else {
			BTTHROW(std::invalid_argument("Unsupported element data type in matlab data"));
		}
		break;
	default:
		BTTHROW(std::invalid_argument("Unsupported element data type in matlab data"));
	}

	// Note: commonFieldInitialization resets pDynDims, which is already set. Set elementDataType/elementSize here of calling it
	this->elementDataType = elementDataType;
	elementSize = NDArray::getElementSize(elementDataType);

	pNDArrays = std::unique_ptr <std::vector<std::unique_ptr<NDArray>>>(new std::vector<std::unique_ptr<NDArray>> );

	NDArray* pNDArray;
	dimIndexType nDArrayNumElems = 1;
	for (dimIndexType i = 0; i < numOfSpatialDimensions; i ++) {
		nDArrayNumElems *= matvar->dims[i];
	}
	dimIndexType nDArrayOffset = 0;
	for (unsigned int i = 0; i < numNDArraysToRead; i++) {
		pNDArray = NDArray::createNDArray(matvar, numOfSpatialDimensions, nDArrayOffset);
		pNDArrays->push_back(std::unique_ptr < NDArray >(pNDArray));
		// If number of spatial dimensions is 1, the offset for the next NDArray in elements is
		// the number of NDArrays (matlab saves arrays by columns)
		nDArrayOffset += nDArrayNumElems;
	}
	Mat_VarFree(matvar);
	matvar = NULL;
#ifdef DATA_DEBUG
	DATA_CERR("done, ");
	END_TIME(eTReadingKDataMatVar);
	TIME_DIFF_TYPE diffTReadingKDataMatVar;
	TIME_DIFF(diffTReadingKDataMatVar, bTReadingKDataMatVar, eTReadingKDataMatVar);
	DATA_CERR("elapsed time: " << diffTReadingKDataMatVar << " s" << std::endl);
#endif
}

/**
 * @brief Read matlab variable with information about spatial and temporal dimensions of a KData or XData
 *
 * @param[in] pDimsMatVar matlab variable with information about dimensions
 * @param[out] pCompleteSpatialDimsVector vector with spatial dimensions sizes
 * @param[out] nTemporalDims number of temporal dimensions
 * @throw std::invalid_argument if element datatype of dimensions array is not supported
 */
void Data::readDimsMatlabVar(const matvar_t *pDimsMatVar, std::vector<dimIndexType> *pCompleteSpatialDimsVector, numCoilsType &numCoils, std::vector<dimIndexType> *pTemporalDimsVector) {
	if ((pDimsMatVar->isComplex) || (pDimsMatVar->data_type != MAT_T_UINT32)) {
		BTTHROW(std::invalid_argument("Matlab dims variable should be an array of 32 bits unsigned integers"));
	}
	char *data = (char*)pDimsMatVar->data;
	dimIndexType offsetInBytes = Mat_SizeOf(pDimsMatVar->data_type);
	dimIndexType nSpatialDims, nTemporalDims, spatialDimSize, temporalDimSize; //, numOfDimensions;
	dimIndexType firstSpatialDimIndex, lastSpatialDimIndex, firstTemporalDimIndex, lastTemporalDimIndex;
	// Minimum number of dimensions of matlab variable (matlab rank) is 2, but Dims variable is a 1-dimensional array (first dimension is 1, second dimension is Dims vector size)
	//numOfDimensions = pDimsMatVar->dims[1];
	// First numOfDimensions-2 positions (indexes from 0 to numOfDimensions-3) are sizes of spatial dimensions
	// The position with index numOfDimensions-2) is the number of coils
	// The last position (index numOfDimensions-1) is the number of temporal dimensions
	nSpatialDims = *((dimIndexType *)(data + offsetInBytes * (MatVarDimsData::NSD_POS)));
	numCoils = *((dimIndexType *)(data + offsetInBytes * (MatVarDimsData::NCOILS_POS)));
	nTemporalDims = *((dimIndexType *)(data + offsetInBytes * (MatVarDimsData::NTD_POS)));
	firstTemporalDimIndex = MatVarDimsData::NTD_POS + 1;
	lastTemporalDimIndex = firstTemporalDimIndex + nTemporalDims - 1;
	firstSpatialDimIndex = lastTemporalDimIndex + 1;
	lastSpatialDimIndex = firstSpatialDimIndex + nSpatialDims -1;
	for (auto i = firstSpatialDimIndex; i <= lastSpatialDimIndex; i++) {
		spatialDimSize = *((dimIndexType *)(data + (offsetInBytes * i)));
		pCompleteSpatialDimsVector->push_back(spatialDimSize);
	}
	for (auto i = firstTemporalDimIndex; i <= lastTemporalDimIndex; i++) {
		temporalDimSize = *((dimIndexType *)(data + (offsetInBytes * i)));
		pTemporalDimsVector->push_back(temporalDimSize);
	}
}

/**
 * @brief Creates empty Data object with spatial and temporal dimensions.
 * @param[in,out] pArraysDims pointer to a vector with pointers to vectors with the spatial dimensions of every contained NDArray
 * @param[in,out] pDynDims pointer to vector of temporal dimensions
 */
void Data::createEmptyDataWithDimensions(std::vector<std::vector<dimIndexType>*>*& pArraysDims, std::vector<dimIndexType>*& pDynDims) {
	constexpr const char* methodErrorPrefix = "createFromNDArraysVector: ";
	std::ostringstream outputstream;
	outputstream << errorPrefix << methodErrorPrefix;
	pNDArraysForGet = new std::vector<const NDArray*>;
	setDynDims(pDynDims);
	pNDArrays = std::unique_ptr <std::vector<std::unique_ptr<NDArray>>>(new std::vector<std::unique_ptr<NDArray>>);
	std::vector <dimIndexType> *pDimsLocal;
	std::unique_ptr<NDArray> unique_ptr_NDArray;

	for (dimIndexType nDArrayIndex = 0; nDArrayIndex < pArraysDims->size(); nDArrayIndex++) {
		pDimsLocal = pArraysDims->at(nDArrayIndex);
		index1DType nDArray1DSize = 1;
		for (index1DType dimsIndex = 0; dimsIndex < pArraysDims->at(nDArrayIndex)->size(); dimsIndex ++) {
			nDArray1DSize *= pArraysDims->at(nDArrayIndex)->at(dimsIndex);
		}
		NDArray* pNDArray = NDArray::createNDArray(pDimsLocal, elementDataType);
		unique_ptr_NDArray = std::unique_ptr < NDArray > (pNDArray);
		pNDArrays->push_back(std::move(unique_ptr_NDArray));
	}
	delete(pArraysDims);
	pArraysDims = nullptr;
	checkNDArraysSizesAndSetAllSizesEqual();
}

/**
 * @brief Initialices Data class fields.
 * @param[in] elementDataType data type of base element
 */
void Data::commonFieldInitialization(ElementDataType elementDataType) {
	pDynDims = std::unique_ptr<std::vector<dimIndexType>>(new std::vector<dimIndexType>());
	pDataDimsAndStridesVector = std::unique_ptr<std::vector<index1DType>>(new std::vector<index1DType>(NUMINITIALPOSITIONSDIMSINFO,0));
	this->elementDataType = elementDataType;
	elementSize = NDArray::getElementSize(elementDataType);
}

/**
 * @brief Checks if all NDArrays have the same spatial dimensions and set accordingly <i>allSizesEqual</i> class field.
 * @throw std::invalid_argument if data is empty
 */
void Data::checkNDArraysSizesAndSetAllSizesEqual() {
	std::vector<dimIndexType> NDArrayDims;
	if (pNDArrays == nullptr) {
		BTTHROW(std::invalid_argument("Empty data, cannot check if all NDArrays have same size, aborting"));
	}
	if (pNDArrays->size() == 0) {
		BTTHROW(std::invalid_argument("Empty data, cannot check if all NDArrays have same size, aborting"));
	}
	NDArrayDims = *(pNDArrays->at(0)->pDims);
	for (dimIndexType i = 1; i < pNDArrays->size(); i ++) {
		if (NDArrayDims != *(pNDArrays->at(i)->pDims)) {
			allSizesEqual = 0;
			return;
		}
	}
	allSizesEqual = 1;
}


/**
 * @brief Destructor, frees previously reserved memory for data structures.
 */
Data::~Data() {
	DATA_CERR("~Data() begins..." << std::endl);
	if (myDataHandle != INVALIDDATAHANDLE) {
		pCLapp->delData(myDataHandle);
	}
	// Free host data vectors
	if (pNDArrays != nullptr) {
		for (dimIndexType i = 0; i < pNDArrays->size(); i++) {
			pNDArrays->at(i) = nullptr;
		}
	}

	// Free vector of dimensions and strides
	pDataDimsAndStridesVector = nullptr;

	pCLapp = nullptr;
	DATA_CERR("~Data() ends..." << std::endl);
}

/**
 * @brief Binds CLapp object to this Data object and adds itself to the list of Data objects managed by the CLapp object.
 *
 * @param[in] pCLapp pointer to CLapp object (contains an initialized OpenCL environment)
 * @param[in] host2DeviceSync format used for storing data in device memory (buffers, images or both)
 */
void Data::setApp(std::shared_ptr<CLapp> pCLapp, SyncSource host2DeviceSync) {
	try {
		this->pCLapp = pCLapp;
		this->syncSourceForLoading = host2DeviceSync;
		if (pCLapp != nullptr) {
			myDataHandle = pCLapp->addData(this, host2DeviceSync);
		}
	} catch (cl::Error& err) {
		/////////////////////////////////////////////////////////////////
		// Catch OpenCL errors and print log if it is a build error
		/////////////////////////////////////////////////////////////////
		DATA_CERR(CLapp::getOpenCLErrorInfoStr(err, __FILE__, __LINE__));
		BTTHROW(std::invalid_argument(CLapp::getOpenCLErrorInfoStr(err, __FILE__, __LINE__)));
	} catch (std::exception& e) {
		BTTHROW(std::invalid_argument(e.what()));
	}
}

// Getters
/**
 * @brief Get number of images stored in the object.
 * @return number of images stored in the object
 */
numberOfDimensionsType Data::getNumNDArrays() const {
	if (pNDArrays == nullptr)
		return 0;
	else
		return pNDArrays->size();
}

/**
 * @brief Gets a read-only pointer to images data.
 * @return pointer to a vector of images (objects of NDArray type)
 */
std::vector<const NDArray*>* Data::getNDArrays() const {
	// pNDArraysForGet stores a COPY of pHostData,
	// it must not be deallocated explicitly
	// (their pointed data is managed by pHostData),
	// but the vector can be emptied (vector data pointers are
	// deallocated, but not the data, which is still pointed
	// from pHostData)
	pNDArraysForGet->resize(0);
	if (pNDArrays == nullptr)
		return pNDArraysForGet;
	for (auto i = pNDArrays->begin(); i < pNDArrays->end(); ++i) {
		pNDArraysForGet->push_back((*i).get());
	}
	//return pResult.get();
	return pNDArraysForGet;
}

/**
 * @brief Gets a read-only pointer to stored data.
 * @return pointer to a vector of objects of NDArray type (every object contains an array of related elements, pixels of an image, for example)
 */
std::vector<const NDArray*>* Data::getData() const {
	return (std::vector<const NDArray*>*) (getNDArrays());
}

/**
 * @brief Gets a read-only pointer to one image data.
 * @param[in] NDArrayIndex NDarray index for the vector of NDArrays
 * @return const pointer to a NDArray object
 */
NDArray* Data::getNDArray(dimIndexType NDArrayIndex) {
	return pNDArrays->at(NDArrayIndex).get();
}

/**
 * @brief Gets a pointer to one NDArray object (image data) from the vector of NDArray objects.
 * @param[in] dynIndexes vector with values for the temporal indexes used for indexing the NDArray objects vector
 * @return pointer to NDArray object at the specified position
 */
const NDArray* Data::getDataAtDynPos(const std::vector<dimIndexType> dynIndexes) const {
	index1DType index1D = get1DIndexFromDynPos(dynIndexes);
	return pNDArrays->at(index1D).get();
}

/**
 * @brief Gets a 1D index for the NDArray objects vector from the vector with temporal indexes.
 * @param[in] dynIndexes vector with values for the temporal indexes used for indexing the NDArray objects vector
 * @return 1D index with the position of the NDArray at the specified temporal indexes
 */
const index1DType Data::get1DIndexFromDynPos(const std::vector<dimIndexType> dynIndexes) const {
	index1DType index1D = 0, mult = 1;
	for (index1DType dynVectorIndex = 0; dynVectorIndex < dynIndexes.size(); dynVectorIndex++) {
		mult = 1;
		if (dynVectorIndex > 0) {
			for (index1DType dynDimsIndex = 0; dynDimsIndex <= dynVectorIndex-1; dynDimsIndex++) {
				mult = mult * (pDynDims->at(dynDimsIndex));
			}
			index1D += dynIndexes.at(dynVectorIndex) * mult;
		}
	}
	printf("index1D: %d\n", index1D);
	return index1D;
}

/**
 * @brief Get number of temporal frames (product of values of each DynDims vector position).
 * @return number of temporal frames
 */
const index1DType Data::getDynDimsTotalSize() const {
	index1DType size = 1;
	for (index1DType i = 0; i<pDynDims->size(); i++) {
		size = size * pDynDims->at(i);
	}
	return size;
}

/**
 * @brief Gets pointer to data of a NDArray stored in host memory as a buffer
 * (pointer to object of class cl::Buffer). Delgates task to CLapp object.
 * @param[in] index NDarray index for the vector of NDArrays
 * @return raw pointer to object of class cl::Buffer
 */
void* Data::getHostBuffer(dimIndexType index) {
	// If pCLapp is not valid or this Data has not been bound to pCLapp return the
	// pointer to the data array inisde pHostData vector (as a generic pointer)
	if ((pCLapp == nullptr) || (myDataHandle == INVALIDDATAHANDLE)) {
		return getNDArray(index)->getHostDataAsVoidPointer();
	}
	return pCLapp->getHostBuffer(myDataHandle, index);
}

/**
 * @brief Gets pointer to data of a NDArray stored in host memory as a image
 * (pointer to object of class cl::Image). Delgates task to CLapp object.
 * @param[in] index NDarray index for the vector of NDArrays
 * @return raw pointer to object of class cl::Image
 */
void* Data::getHostImage(dimIndexType index) {
	// If pCLapp is not valid or this Data has not been bound to pCLapp return the
	// pointer to the data array inisde pHostData vector (as a generic pointer)
	if ((pCLapp == nullptr) || (myDataHandle == INVALIDDATAHANDLE)) {
		return getNDArray(index)->getHostDataAsVoidPointer();
	}
	return pCLapp->getHostImage(myDataHandle, index);
}

/**
 * @brief Gets pointer to array of dimensions and strides and data of all NDArrays of a Data object stored in contiguous device memory as a buffer
 * (pointer to object of class cl::Buffer). Delegates task to CLapp object.
 * @return raw pointer to object of class cl::Buffer
 */
cl::Buffer* Data::getCompleteDeviceBuffer() {
	return pCLapp->getCompleteDeviceBuffer(myDataHandle);
}

/**
 * @brief Gets pointer to data of all NDArrays of a Data object stored in device memory as a buffer
 * (pointer to object of class cl::Buffer). Delegates task to CLapp object.
 * @return raw pointer to object of class cl::Buffer
 */
cl::Buffer* Data::getDeviceBuffer() {
	return pCLapp->getDeviceBuffer(myDataHandle);
}

/**
 * @brief Gets HIP pointer to data of all NDArrays of a Data object stored in device memory as a buffer
 * (raw void* pointer). Delegates task to CLapp object.
 * @return raw void* HIP pointer to device data
 */
void* Data::getHIPDeviceBuffer() {
	return pCLapp->getHIPDeviceBuffer(myDataHandle);
}

/**
 * @brief Gets pointer to data of all NDArrays of a Data object stored in device memory as a buffer
 * (pointer to object of class cl::Buffer). Delegates task to CLapp object.
 * @param[in] index NDarray index for the vector of NDArrays
 * @return raw pointer to object of class cl::Buffer
 */
cl::Buffer* Data::getDeviceBuffer(dimIndexType index) {
	return pCLapp->getDeviceBuffer(myDataHandle, index);
}

/**
 * @brief Gets pointer to data of a NDArray stored in device memory as a image
 * (pointer to object of class cl::Image). Delegates task to CLapp object.
 * @param[in] index NDarray index for the vector of NDArrays
 * @return raw pointer to object of class cl::Image
 */
cl::Image* Data::getDeviceImage(dimIndexType index) {
	return pCLapp->getDeviceImage(myDataHandle, index);
}

/**
 * @brief Gets pointer to spatial and temporal data dimensions stored in host memory as an array.
 * @return generic pointer to array of spatial and temporal data dimensions
 */
const void* Data::getDataDimsAndStridesHostBuffer() const {
	return pCLapp->getDataDimsAndStridesHostBuffer(myDataHandle);
}

/**
 * @brief Gets pointer to spatial and temporal data dimensions stored in device memory as an array.
 * @return generic pointer to array of spatial and temporal data dimensions
 */
cl::Buffer* Data::getDataDimsAndStridesDeviceBuffer() {
	return pCLapp->getDataDimsAndStridesDeviceBuffer(myDataHandle);
}

/**
 * @brief Converts data of pHostImage field (pointer to raw data) to a text representation.
 *
 * @param[in] title title for data text representation
 * @param[in] index index of the NDArray to get representation for
 * @return string with text representation of data in pHostImage
 */
const std::string Data::hostImageToString(std::string title, dimIndexType index) {
	return hostImageOrBufferToString(title, index, SyncSource::IMAGE_ONLY);
}

/**
 * @brief Converts data of pHostBuffer field (pointer to raw float data) to a text representation.
 *
 * @param[in] title title for data text representation
 * @param[in] index index of the NDArray to get representation for
 * @return string with text representation of data in pHostBuffer
 */
const std::string Data::hostBufferToString(std::string title, dimIndexType index) {
	return hostImageOrBufferToString(title, index, SyncSource::BUFFER_ONLY);
}

/**
 * @brief Converts data of pHostImage or pHostBuffer, depending on syncSource parameter value, to a text representation.
 *
 * @param[in] title title for data text representation
 * @param[in] index index of the NDArray to get representation for
 * @param[in] syncSource type of data source: image (SyncSource::IMAGE_ONLY value) or buffer (SyncSource::BUFFER_ONLY value)
 * @return string with text representation of data in pHostImage
 * @throw std::invalid_argument if value of syncSource parameter is invalid
 */
const std::string Data::hostImageOrBufferToString(std::string title, dimIndexType index, SyncSource syncSource) {
	void* pElementsArray;
	if (syncSource == SyncSource::IMAGE_ONLY) {
		pElementsArray = getHostImage(index);
	} else {
		if (syncSource == SyncSource::BUFFER_ONLY) {
			pElementsArray = getHostBuffer(index);
		} else {
			BTTHROW(std::invalid_argument("SyncSource::ALL not supported for hostImageOrBufferToString method"));
		}
	}
	return getNDArray(index)->nDArrayElementsToString(title, pElementsArray);
}


/**
 * @brief Stores vector of standard pointers to NDArrays objects as smart pointer to vector of smart pointers
 * @param[in,out] pNDArrays vector of standard pointers to NDArray objects
 */
void Data::internalSetData(std::vector<NDArray*>*& pNDArrays) {
	if (this->pNDArrays != nullptr) {
		this->pNDArrays->resize(0); // empty the pNDArrays attribute
	} else {
		this->pNDArrays =
				std::unique_ptr <std::vector<std::unique_ptr<NDArray>>>
				(new std::vector<std::unique_ptr<NDArray>> );
	}
	for (auto i = pNDArrays->begin(); i < pNDArrays->end(); ++i) {
		for (auto j = pNDArrays->begin(); j < i; j++) {
			assert (*i != *j);
		}
		this->pNDArrays->push_back(std::unique_ptr < NDArray > (*i));
	}
	delete (pNDArrays);
	pNDArrays = nullptr;
	checkNDArraysSizesAndSetAllSizesEqual();
}

/**
 * @brief Checks if index is valid for NDArray spatial dimensions.
 * @param[in] index index to be checked
 * @param[in] nDArraySize vector with spaial dimensions of NDArray
 */
bool Data::checkNDArrayIndex(dimIndexType index, dimIndexType nDArraySize) {
	if ((index < 0) || (index > nDArraySize)) {
		std::cerr << "NDArray index out of range: " << index << "(NDArray size: " << std::to_string(nDArraySize) << ")\n";
		return false;
	}
	return true;
}

/**
 * @brief Store image data (vector to NDArray pointers) in the class field pData.
 * @param[in,out] pData image data as a vector of NDArrays. All parameters of type *& (reference to pointer) have move semantics:
 * ownership of memory is moved from parameter to this object, and parameter value is set to nullptr after method completion.
 */
void Data::setData(std::vector<NDArray*>*& pNDArrays) {
	internalSetData(pNDArrays);
}

/**
 * @brief Sets vector of temporal dimensions. All parameters of type *& (reference to pointer) have move semantics: ownership of
 * memory is moved from parameter to this object, and parameter value is set to nullptr after method completion.
 * @param[in,out] pDynDims pointer to vector of temporal dimensions
 */
void Data::setDynDims(std::vector<dimIndexType>*& pDynDims) {
	this->pDynDims.reset(pDynDims);
	pDynDims = nullptr;

	// Understand nullptr temporal dimensions as a single, one-length temporal dimension
	if(this->pDynDims == nullptr) {
		this->pDynDims = std::unique_ptr<std::vector<dimIndexType>>(new std::vector<dimIndexType>({1}));
		CERR("Data::setDynDims: nullptr pDynDims given; set to single temporal dimension of length 1\n");
	}

	// Check for zero-length dimensions and set them to 1 if found
	auto& dynDims = *this->pDynDims;
	for(auto&& i:dynDims)
		if(i == 0) {
			i = 1;
			CERR("Data::setDynDims: zero-length temporal dimension found; set to length=1\n");
		}

}

/**
 * @brief Build a file name prefix of the form \<prefix\>\<width\>x\<height\> (for a 2D dimensions vector) or
 * \<prefix\>\<width\>x\<height\>x\<depth\> (for a 3D dimensions vector).
 * @param[in] prefix string with the prefix part of the name
 * @param[in] pDims vector of dimensions for extracting widht, height and depth (if present) values
 * @return string with the format \<prefix\>\<rows\>x\<cols\> (for a 2D pDims) or
 * \<prefix\>\<width\>x\<height\>x\<depth\> (for a 3D pDims)
 */
std::string Data::buildFileNamePrefix(const std::string prefix, const std::vector<dimIndexType>* pDims) {
	std::ostringstream fileNamePrefix;
	fileNamePrefix << prefix;
	if (pDims != nullptr) {
		dimIndexType dimsSize = pDims->size();
		for (dimIndexType i = 0; i < dimsSize; i++) {
			fileNamePrefix << pDims->at(i);
			if (i != (dimsSize-1)) {
				fileNamePrefix << "x";
			}
		}
	}
	return fileNamePrefix.str();
}

/**
 * @brief Build a file name sufix part of the form \<suffix\>.\<fileExtension\>.
 * @param[in] suffix string with the suffix part of the name (default value is empty string)
 * @param[in] fileExtension string with de extension part of the name (default value is "raw")
 * @return string with the format \<suffix\>.\<fileExtension\>
 */
std::string Data::buildFileNameSuffix(const std::string suffix, const std::string fileExtension) {
	return suffix + "." + fileExtension;
}


/**
 * @brief Load data of a group of files to hostData (every file contains one image and it is stored into a NDArray object).
 *
 * Files of this group have names with the format \<fileNamePrefix\>\<fileNameSuffix(i)\>\<fileNameExtension\>,
 * where \<i\> is the index for indentifying the suffix of every image.
 * @param[in] fileNamePrefix fixed part of the file name before the variable part
 * @param[in,out] pArraysDims pointer to a vector with pointers to vectors with the dimensions of every image
 * @param[in,out] pDynDims pointer to vector of temporal dimensions
 * @param[in] fileNameSuffixes vector with the variable part of the name for every image
 * @param[in] fileNameExtension extension of the file name
 */
void Data::loadRawHostData(const std::string fileNamePrefix, std::vector<std::vector< dimIndexType >*>*& pArraysDims,
		std::vector <dimIndexType>*& pDynDims,
		std::vector<std::string> fileNameSuffixes, const std::string fileNameExtension) {
	setDynDims(pDynDims);
	std::vector<dimIndexType>* pTempDims;
	NDArray* pTempNDArray;
	std::vector<NDArray*>* pNDArrays;
	pNDArrays = new std::vector<NDArray*>;
	// if file name suffixes vector is empty we must add an empty string value (without suffixes no file is read)
	if (fileNameSuffixes.size() == 0) {
		fileNameSuffixes.push_back("");
	}
	DATA_CERR ("Reading files with prefix: " << fileNamePrefix << " (" << fileNameSuffixes.size() << " file(s))... ");
	for (dimIndexType i = 0; i < fileNameSuffixes.size() ; i++) {
		std::ostringstream fileNameStream;
		pTempDims = new std::vector<dimIndexType>(*(pArraysDims->at(i))); // parameter pDims contents are copied to new vector
		fileNameStream << fileNamePrefix;
		if (fileNameSuffixes.at(i) != "") {
			fileNameStream << fileNameSuffixes.at(i);
		}
		fileNameStream << fileNameExtension;
		std::string completeFileName = fileNameStream.str();
		pTempNDArray = NDArray::createNDArray(completeFileName, pTempDims, elementDataType);
		pNDArrays->push_back(pTempNDArray); // copy NDArray pointer and add it to vector
	}
	DATA_CERR("done" << std::endl);
	this->setData(pNDArrays);
	delete(pArraysDims);
	pArraysDims = nullptr;
}

/**
 * @brief Save hostBuffer or hostImage data (depending on the parameter syncSource) of every NDArray
 * contained object to a group of files.
 *
 * File names have format \<fileNamePrefix\>\<fileNameSuffix(i)\>\<fileNameExtension\>,
 * where \<i\> is the index for indentifying the suffix of every image.
 * @param[in] syncSource set the data source used for saving (IMAGE_ONLY or BUFFER_ONLY)
 * @param[in] fileNamePrefix fixed part of the file name before the variable part
 * @param[in] fileNameSuffixes vector with the variable part of the name for every image
 * @param[in] fileNameExtension extension of the file name
 */
void Data::saveRawHostData(const SyncSource syncSource, const std::string fileNamePrefix, std::vector<std::string> fileNameSuffixes,
		const std::string fileNameExtension) {
	dimIndexType index(0);
	if (pNDArrays->size() > fileNameSuffixes.size()) {
		std::ostringstream errorInfo;
		errorInfo << "number fileNameSuffixes (" << fileNameSuffixes.size() << ") must be equal to number of NDArrays ("
				<< pNDArrays->size() << ")\n";
		BTTHROW(std::invalid_argument(errorInfo.str()));
	}
	//for (auto i = pNDArrays->begin(); i != pNDArrays->end(); i++) {
	for (size_t i = 0; i < pNDArrays->size(); i++) {
		std::ostringstream fileNameStream;
		fileNameStream << fileNamePrefix << fileNameSuffixes.at(index);
		fileNameStream << fileNameExtension;

		std::ofstream f(fileNameStream.str());
		void* data;
		if(syncSource == SyncSource::BUFFER_ONLY) {
			data = getHostBuffer(i);
		} else {//image
			data = getHostImage(i);
		}
		f.write(reinterpret_cast<const char*>(data), getNDArray(i)->size() * elementSize);
		f.close();
		fileNameStream.clear();
		index++;
	}
}

/**
 * @brief Get a fragment from stored images.
 * @param[in] specif fragment specification
 * @return group of images (pointer to read-only vectors to NDArrays)
 */
const std::vector<const NDArray*>* Data::getFragment(FragmentSpecif specif) {
	std::vector<const NDArray*>* pResult = new std::vector<const NDArray*>;
	for (auto i = pNDArrays->begin(); i < pNDArrays->end(); ++i) {
		pResult->push_back((*i).get());
	}
	return pResult;
}

/**
 * @brief Calculates image/volume size in number of real elements rounded up to the nearest multiple of device memory base
 * address alignment, and stores it in dataStridesVector field after storing spatial dimensions strides.
 * @param[in] NDArray1DIndex index of the NDArray to be used for getting spatial dimensions strides
 * @param deviceMemBaseAddrAlignInBytes device memory alignment in bytes
 */
void Data::calcDataAlignedSize(dimIndexType NDArray1DIndex, dimIndexType deviceMemBaseAddrAlignInBytes) {
	dimIndexType numSpatialDims = pNDArrays->at(NDArray1DIndex)->getNDims();
	dimIndexType lastSpatialDimSize = pNDArrays->at(NDArray1DIndex)->getDims()->at(numSpatialDims-1);
	std::vector<dimIndexType>* pSpatialDimsStridesVector =
			pNDArrays->at(NDArray1DIndex)->calcUnaligned1DArrayStridesFromNDArrayDims();
	// NDArray data size unaligned measured in kernel known elements (float -not complex-,  uint, etc.)
	dimIndexType NDArrayDataUnalignedSizeInElements = pSpatialDimsStridesVector->at(numSpatialDims-1) * lastSpatialDimSize;
	// Memory alignment measured in kernel known elements (memory aligment in bytes / size of kernel known element in bytes).
	// Element size is 8 bytes for complex data NDArray, but 4 for floats, known by kernels.
	// This way, effective elementSize is stored elemens size / first spatial stride value (2 for complex data, 1 for float, uint, etc).
	dimIndexType memoryAlignmentInElements = deviceMemBaseAddrAlignInBytes / (elementSize / pSpatialDimsStridesVector->at(ElementStridePos));
	dimIndexType NDArrayDataAlignedSize = CLapp::roundUp(NDArrayDataUnalignedSizeInElements, memoryAlignmentInElements);
	pDataDimsAndStridesVector->insert(pDataDimsAndStridesVector->end(), pSpatialDimsStridesVector->begin(),
			pSpatialDimsStridesVector->end());
	delete(pSpatialDimsStridesVector);
	pDataDimsAndStridesVector->push_back(NDArrayDataAlignedSize);
}

/**
 * @brief Stores image/volume spatial and temporal dimensions in dataDims field (dataDims data type is valid to be used for
 * kernel parameter).
 * @throw std::invalid_argument if data is empty
 */
void Data::calcDataDims() {
	std::string errorMessage;
	if (pDataDimsAndStridesVector->size() > NUMINITIALPOSITIONSDIMSINFO) { // DataDimsAndStridesVector already set, must be reset
		// pDataDimsAndStridesVector is initialized with NUMINITIALPOSITIONSDIMSINFO positions (the unique known and constant positions)
		pDataDimsAndStridesVector->resize(NUMINITIALPOSITIONSDIMSINFO);
	}
	errorMessage.append(errorPrefix);
	errorMessage.append("data empty, cannot calculate spatial dimensions");

	if (pNDArrays == nullptr) {
		BTTHROW(std::invalid_argument(errorMessage.c_str()));
	}
	if (pNDArrays->size() == 0) {
		BTTHROW(std::invalid_argument(errorMessage.c_str()));
	}
	// number of spatial dims for the first NDArray (the same for all NDArrays if allSizesEqual class field is 1)
	uint numSpatialDims = pNDArrays->at(0)->getNDims();
	pDataDimsAndStridesVector->at(NumSpatialDimsPos) = numSpatialDims;
	pDataDimsAndStridesVector->at(AllSizesEqualPos) = allSizesEqual;
	uint numTemporalDims = getNDynDims();
	pDataDimsAndStridesVector->at(NumTemporalDimsPos) = numTemporalDims;
	for (uint temporalDimId = 0; temporalDimId < numTemporalDims; temporalDimId++) {
		pDataDimsAndStridesVector->push_back(getDynDims()->at(temporalDimId));
	}
	//uint FirstSpatialDimsPos = FirstTemporalDimPos + numTemporalDims;
	for (uint spatialDimId = 0; spatialDimId < numSpatialDims; spatialDimId++) {
		pDataDimsAndStridesVector->push_back(pNDArrays->at(0)->getDims()->at(spatialDimId));
	}
	if (!allSizesEqual) {// store information of spatial dimensions for last (pNDArrays->size())-1 NDArrays
		for (uint NDArrayIndex = 1; NDArrayIndex < pNDArrays->size(); NDArrayIndex++) {
			numSpatialDims = pNDArrays->at(NDArrayIndex)->getNDims();
			for (uint spatialDimId = 0; spatialDimId < numSpatialDims; spatialDimId++) {
				pDataDimsAndStridesVector->push_back(pNDArrays->at(NDArrayIndex)->getDims()->at(spatialDimId));
			}
		}
	}
}

/**
 * @brief Calculates strides associated to temporal dimensions and first following stride, for accessing data stored as 1D arrays of real
 * elements, and stores them in dataStridesVector field.
 *
 * The first following stride will be the coil stride or the first frame dimension stride
 * depending on the subclass of Data (coil stride for KData and SensitivityMapsData, first frame dimension stride for XData
 * and SamplingMaskData).
 * @param deviceMemBaseAddrAlignInBytes device memory alignment in bytes
 */
void Data::calcDataStrides(dimIndexType deviceMemBaseAddrAlignInBytes) {
	dimIndexType NumNDArrays;
	if (allSizesEqual) {
		NumNDArrays = 1;
	} else {
		NumNDArrays = pNDArrays->size();
	}
	for (dimIndexType NDArray1DIndex = 0; NDArray1DIndex < NumNDArrays; NDArray1DIndex++) {
		// Calculate all spatial strides and the following one (rounded up to multiple of deviceMemBaseAddrAlignInBytes/8
		// bytes per complex element (2 floats)
		calcDataAlignedSize(NDArray1DIndex, deviceMemBaseAddrAlignInBytes);
		//dimIndexType lastStride = pDataStridesVector->at(pDataStridesVector->size()-1);
		dimIndexType lastStride = pDataDimsAndStridesVector->at(pDataDimsAndStridesVector->size()-1);
		dimIndexType numCoils = pDataDimsAndStridesVector->at(NumCoilsPos);
		if (numCoils == 0) {// XData
			// Last stride is first temporal dimension stride for XData
			for (dimIndexType temporalDimId = 1; temporalDimId < getNDynDims(); temporalDimId++) {
				lastStride *= pDynDims->at(temporalDimId-1);  // new stride is equal to previous stride * previous dimension
				pDataDimsAndStridesVector->push_back(lastStride);
			}
		} else {
			// Last stride is coil stride for KData or SensitivityMapsData
			// stride for first temporal dimension is equal to coils stride * number_of_coils
			lastStride *= numCoils;
			for (dimIndexType temporalDimId = 0; temporalDimId < getNDynDims(); temporalDimId++) {
				pDataDimsAndStridesVector->push_back(lastStride);
				lastStride *= pDynDims->at(temporalDimId);  // new stride is equal to previous stride * previous dimension
			}
		}
	}
}

/**
 * Stores image/volume spatial and temporal dimensions in DataDimsAndStrides field
 * @brief Calculation of data spatial and temporal dimensions and associated strides for accessing data stored as 1D arrays of real
 * elements (see @ref calcDataDims and @ref calcDataStrides).
 * @param deviceMemBaseAddrAlignInBytes device memory alignment in bytes
 */
void Data::calcDimsAndStridesVector(dimIndexType deviceMemBaseAddrAlignInBytes) {
	calcDataDims();
	calcDataStrides(deviceMemBaseAddrAlignInBytes);
	DATA_CERR("dimsAndStridesVector: ");
	for (size_t i=0; i < pDataDimsAndStridesVector->size() - 1; i++) {
		DATA_CERR(pDataDimsAndStridesVector->at(i) << ", ");
	}
	DATA_CERR(pDataDimsAndStridesVector->at(pDataDimsAndStridesVector->size() - 1) << std::endl);
}

/**
 * @brief Checks if every spatial dimension is multiple of VECTORDATATYPESIZE/2.
 * @param[in] pDims pointer to vector with spatial dimensions
 * @throw invalid_argument if checks failed for at least one dimension
 */
void Data::checkValiditySpatialDimensions(const std::vector<dimIndexType>* pDims) {
	std::ostringstream errorMsgStream;
	if (pDims == nullptr) {
		BTTHROW(std::invalid_argument("NDArray dimensions not set yet"));
	}
	dimIndexType dimsSize = 1; //pDims->size();
	for (dimIndexType i = 0; i < dimsSize; i++) {
		if ((pDims->at(i)) % (VECTORDATATYPESIZE/2) != 0) {
			errorMsgStream << "NDArray dimension " << i << " is not multiple of " << (VECTORDATATYPESIZE/2) << " (is " << pDims->at(i) << ")";
			BTTHROW(std::invalid_argument(errorMsgStream.str()));
		}
	}
}

/**
 * @brief Checks if every spatial dimension of every NDArray is multiple of VECTORDATATYPESIZE/2.
 * @param[in] pArraysDims pointer to vector of pointerts to vectors with spatial dimensions
 * @throw invalid_argument if checks failed for at least one dimension of one NDArray
 */
void Data::checkValiditySpatialDimensions(const std::vector<std::vector<dimIndexType>*>* pArraysDims) {
	for (dimIndexType i = 0; i < pArraysDims->size(); i++) {
		checkValiditySpatialDimensions(pArraysDims->at(i));
	}
}

/**
 * @brief Read one or more matlab variables from a file with matlab format
 * @param[in] fileName name of the file in matlab format
 * @param[in] variableNames vector with name of variables to be read
 * @return map of matlab variables (keys are variable names, values are matlab variables)
 * @throw std::invalid_argument if matlab file cannot be opened or variable cannot be found in matlab file
 */
std::map<std::string,matvar_t*>* Data::readMatlabVariablesFromFile(std::string fileName, std::vector<std::string> variableNames) {
#ifdef DATA_DEBUG
	BEGIN_TIME(bTReadingMatlabFile);
	DATA_CERR("Reading matlab file...\n");
#endif
	mat_t *matfp;
	matvar_t *matvar;
	std::map<std::string,matvar_t*>* pMatlabVariablesMap = new std::map<std::string,matvar_t*>;
	matfp = Mat_Open(fileName.c_str(), MAT_ACC_RDONLY);
	if ( NULL == matfp ) {
		BTTHROW(std::invalid_argument("Error opening MAT file '" + fileName + "'!\n"));
		return pMatlabVariablesMap;
	}
	for (auto variableName : variableNames) {
		matvar = Mat_VarRead(matfp, variableName.c_str());
		if ( NULL == matvar ) {
			BTTHROW(std::invalid_argument("Variable '" + variableName + "' not found, or error reading MAT file\n"));
		} else {
			(*pMatlabVariablesMap)[variableName] = matvar;
#ifdef DATA_DEBUG
			Mat_VarPrint(matvar, 1);
#endif
		}
	}
	Mat_Close(matfp);
#ifdef DATA_DEBUG
	DATA_CERR("Done\n");
	END_TIME(eTReadingMatlabFile);
	TIME_DIFF_TYPE diffTReadingMatlabFile;
	TIME_DIFF(diffTReadingMatlabFile, bTReadingMatlabFile, eTReadingMatlabFile);
	DATA_CERR("Elapsed time: " << diffTReadingMatlabFile << " s" << std::endl);
#endif
	return pMatlabVariablesMap;
}

/**
 * @brief Read all matlab variables stored in a file with matlab format
 * @param[in] fileName name of the file in matlab format
 * @return map of matlab variables (keys are variable names, values are matlab variables)
 * @throw std::invalid_argument if matlab file cannot be opened
 */
std::map<std::string,matvar_t*>* Data::readMatlabVariablesFromFile(std::string fileName) {
	mat_t *matfp;
	matvar_t *matvar;
	std::map<std::string,matvar_t*>* pMatlabVariablesMap = new std::map<std::string,matvar_t*>;
	matfp = Mat_Open(fileName.c_str(), MAT_ACC_RDONLY);
	if ( NULL == matfp ) {
		BTTHROW(std::invalid_argument("Error opening MAT file '" + fileName + "'!\n"));
		return pMatlabVariablesMap;
	}

	while ( (matvar = Mat_VarReadNext(matfp)) != NULL ) {
		// sanity check: since rank is a signed int in matio and we use unsigned vars to store dimensions, we could be fooled by a corrupted matlab file
		if( matvar->rank < 2 )
			BTTHROW(std::invalid_argument("Data::readMatlabVariablesFromFile: invalid rank in matlab variable"));

		(*pMatlabVariablesMap)[matvar->name] = matvar;
#ifdef DATA_DEBUG
		Mat_VarPrint(matvar, 1);
#endif
	}
	Mat_Close(matfp);
	return pMatlabVariablesMap;
}

/**
 * @brief Fills a previously created matlab variable with data from this object data
 *
 * @param[in] matfp pointer to open matlab format file
 * @param[in] matVarName of the matlab variable to be filled with data
 * @param[in] pMatVarInfo pointer to a MatVarInfo object (storing matlab info about data: class_type, data_type, data values and options)
 * @param[in] syncSource source of data in device memory (buffers or images)
 * @throw std::invalid_argument if syncSource value is not supported or matlab variable cannot be created
 */
void Data::fillMatlabVarInfo(mat_t* matfp, std::string matVarName, MatVarInfo* pMatVarInfo, SyncSource syncSource) {
	matvar_t *matvar;
	dimIndexType numNDArrayElements = getNDArray(0)->size();
	dimIndexType matlabNDArrayOffset = 0;
	void *pNDArrayData;
	for (dimIndexType i = 0; i < getNDArrays()->size(); i++) {
		switch (syncSource) {
		case SyncSource::BUFFER_ONLY:
			pNDArrayData = getHostBuffer(i);
			break;
		case SyncSource::IMAGE_ONLY:
			pNDArrayData = getHostImage(i);
			break;
		default:
			BTTHROW(std::invalid_argument("Unsupported SyncSource in dumpMatlabElement"));
		};

		pNDArrays->at(i)->fillMatlabVarInfo(pMatVarInfo, (const dimIndexType*) pDataDimsAndStridesVector.get()->data(),
				matlabNDArrayOffset, pNDArrayData);
		matlabNDArrayOffset += numNDArrayElements;
	}
	matvar = Mat_VarCreate(matVarName.c_str(),pMatVarInfo->getClassType(), pMatVarInfo->getDataType(), pMatVarInfo->getRank(), pMatVarInfo->getDims(), pMatVarInfo->getData(),pMatVarInfo->getOpt());
	if ( NULL == matvar ) {
		BTTHROW(std::invalid_argument(std::string("Error creating variable ") + matVarName));
	} else {
		Mat_VarWrite(matfp, matvar, MAT_COMPRESSION_ZLIB);
		Mat_VarFree(matvar);
	}
}

/**
 * @brief Copy data stored in device memory to host memory (from buffers or images). Delegates task to CLapp object.
 * @param[in] device2HostSyncSource data source in device memory (buffers, images, both not supported)
 * @param[in] queueFinish true if queue.finish() method must be called to guarantee that kernel execution has finished before
 * copying data back to host memory
 */
void Data::device2Host(SyncSource device2HostSyncSource, bool queueFinish) {
	pCLapp->device2Host(getHandle(), device2HostSyncSource, queueFinish);
}

// Show-related stuff

void Data::show(ShowParms* inParms) {
	// if the caller does not specify parameters, set default ones
	std::unique_ptr<ShowParms> parms;
	if(inParms)
		parms = std::unique_ptr<ShowParms>(new ShowParms(*inParms));
	else
		parms=std::unique_ptr<ShowParms>(new ShowParms());

	// Some sanity checks
	if(allSizesEqual!=1)
		BTTHROW(std::invalid_argument("show: Every NDArray in the Data object must be the same size"));

	auto numNDArrays=getNumNDArrays();
	if(numNDArrays==0)
		BTTHROW(std::invalid_argument("show: the Data object to show does not contain any NDArrays"));

	// All NDArrays are equal in size if we are here, so take spatial dimensions from the first one
	auto nDArray=getNDArray(0);
	auto nDims=nDArray->getNDims();
	auto dims=nDArray->getDims();

	// Get the number of frames to show
	auto tempDims=getDynDims();
	unsigned int nFrames=tempDims->at(parms->timeDimension);

	// Transform data if necessary (complex to real, normalize)
	// and select OpenCV data type accordingly
	auto pIn = shared_from_this();
	std::shared_ptr<XData> pShow = std::dynamic_pointer_cast<XData>(pIn);
	int opencvDataType;
	if(elementDataType == std::type_index(typeid(complexType)) ||
			elementDataType == std::type_index(typeid(realType))) {

		// If data is complex, convert it to real first
		if(elementDataType == std::type_index(typeid(complexType))) {
			auto& complex2real=pCLapp->internalProcesses.complex2real;
			pShow = std::make_shared<XData>(pCLapp, pIn, std::type_index(typeid(realType)));
			auto initParms=std::make_shared<Complex2Real::InitParameters>(parms->complexPart);
			complex2real.setInitParameters(initParms);
			complex2real.setInput(pIn);
			complex2real.setOutput(pShow);
			complex2real.init();
			complex2real.launch();

			// Set input for the (possible) next transformation
			pIn=pShow;
		}

		if(parms->normalize) {
			// If we are told to normalize but not a normalizing factor, guess it from the data
			if(parms->normFactor==0) {
				auto& sumReduce = pCLapp->internalProcesses.sumReduce;
				auto pSum = std::make_shared<XData>(pCLapp, 1 * sizeof(realType), std::type_index(typeid(realType)));
				sumReduce.setInput(pIn);
				sumReduce.setOutput(pSum);
				sumReduce.init();
				sumReduce.launch();

				auto& normalizeShow = pCLapp->internalProcesses.normalizeShow;
				if(pShow.get()==this)
					auto pShow = std::make_shared<XData>(pCLapp, 1 * sizeof(realType), std::type_index(typeid(realType)));
				normalizeShow.setInput(pIn);
				normalizeShow.setOutput(pShow);
				normalizeShow.init();
				auto launchParms=std::make_shared<NormalizeShow::LaunchParameters>(pSum);
				normalizeShow.setLaunchParameters(launchParms);
				normalizeShow.launch();

				// Set input for the (possible) next transformation
						pIn=pShow;
			}
			// if we are given a normalizing factor, just multiply the data with it
			else {
				auto& scalarMultiply=pCLapp->internalProcesses.scalarMultiply;
				if(pShow.get()==this)
					pShow = std::make_shared<XData>(pCLapp, pIn, std::type_index(typeid(realType)));
				auto launchParms=std::make_shared<ScalarMultiply::LaunchParameters>(parms->normFactor);
				scalarMultiply.setLaunchParameters(launchParms);
				scalarMultiply.setInput(pIn);
				scalarMultiply.setOutput(pShow);
				scalarMultiply.init();
				scalarMultiply.launch();
			}
		}

		// Get data from device
		pCLapp->getCommandQueue().finish();
		pShow->device2Host(SyncSource::BUFFER_ONLY);

		// Set OpenCV data type to 32bit-float-1channel
		opencvDataType=CV_32FC1;
	}
	else if(elementDataType == std::type_index(typeid(dimIndexType))) {
		if(parms->normalize)
			DATA_CERR("show: NDArrays of integral type can't be normalized. Showing unnormalized.");

		// Set OpenCV data type to 32bit-signed-1channel
		opencvDataType=CV_32SC1;
	}
	else
		BTTHROW(std::invalid_argument("show: unsupported Data type"));

	// We need as many cv::Mat objects as the number of frames to show
	std::vector<cv::Mat> mats(nFrames);

	// If the NDArray is 2D, we can construct a cv::Mat object from our host buffer in one zero-copy shot
	// Note: we bet that every dimension will fit in a const int... (getDims returns unsigned int)
	if(nDims==2) {
		// Build vector of mat objects from respective host buffers
		for(dimIndexType numNDArray=0; numNDArray<nFrames; numNDArray++)
			mats[numNDArray]=cv::Mat(nDims,reinterpret_cast<const int*> (dims->data()),opencvDataType,pShow->getHostBuffer(numNDArray));
	}
	else {
		BTTHROW(std::invalid_argument("non-2D NDArrays can't be shown yet"));
	}

	// Create cv window
	cv::namedWindow(parms->title, cv::WINDOW_NORMAL);
	cv::resizeWindow(parms->title, parms->scaleFactor * (*dims)[WIDTHPOS], parms->scaleFactor * (*dims)[HEIGHTPOS]);

	// Show!
	int key;
	for(;;) {
		for(dimIndexType numNDArray=0; numNDArray<nFrames; numNDArray++) {
			cv::Point textOrg(20,25);
			std::stringstream s;
			s.clear();
			s<<numNDArray;
			cv::putText(mats[numNDArray], s.str(), textOrg, cv::FONT_HERSHEY_SCRIPT_SIMPLEX, 1, cv::Scalar::all(255), 1, 8, false);
			cv::imshow(parms->title, mats[numNDArray]);

			// Note: CV docs say waitKey returns (int)(-1) if no key pressed but actually returns (int)255
			if(((key=cv::waitKey(1000/parms->fps))!=255) && parms->keyEnds)
				break;
		}
		if((key!=255) && parms->keyEnds)
			break;
	}
}

} /* namespace OpenCLIPER */

#undef DATA_DEBUG
