/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
#include <OpenCLIPER/MatVarInfo.hpp>

namespace OpenCLIPER {
    /**
     * @brief Default constructor, sets all the class variables with values passed as method parameters
     *
     * @param[in] class_type matlab class type
     * @param[in] data_type matlab data type
     * @param[in] data array of data in matlab format
     * @param[in] opt matlab options
     */
    MatVarInfo::MatVarInfo(enum matio_classes class_type, enum matio_types data_type, void *data, int opt) {
    this->class_type = class_type;
    this->data_type = data_type;
    this->data = data;
    this->opt = opt;
}

/**
 * @brief Updates dimensions and rank class variables from a vector of dimensions
 *
 * @param[in] pDimsVectorArg pointer to vector of data dimensions
 * @param[in] swapColsAndRows true if columns and rows must be swapped (matlab arrays are stored by columns, C/C++ convention is store them by rows)
 */
void MatVarInfo::updateDimsAndRank(const std::vector<dimIndexType>* pDimsVectorArg) {
    // pDimsVector field still empty, must guarantee minimum rank of 2
    // and swap first and second dimensions, if both exist
    if (pDimsVector->size() == 0) {
        // data with 1 spatial dimension, minimum matlab rank is 2 (vector is an array of 1xN dimensions)
        if (pDimsVectorArg->size() == 1) {
            pDimsVector->push_back(1);
            pDimsVector->push_back(pDimsVectorArg->at(WIDTHPOS));
        } else { // 2 or more spatial dimensions, first 2 swapped (matlab stores data by columns, OpenCLIPER by rows)
            pDimsVector->push_back(pDimsVectorArg->at(HEIGHTPOS));
            pDimsVector->push_back(pDimsVectorArg->at(WIDTHPOS));
        }
        // append third and following dimensions
        for (dimIndexType i = 2; i < pDimsVectorArg->size(); i++) {
            pDimsVector->push_back(pDimsVectorArg->at(i));
        }
    } else {
        // pDimsVector field already contains some dimensions info, only
        // append new ones without reordering
        for (dimIndexType i = 0; i < pDimsVectorArg->size(); i++) {
            pDimsVector->push_back(pDimsVectorArg->at(i));
        }
    }
    rank = pDimsVector->size();
}

/**
 * @brief Gets number of data dimensions
 *
 * @return the number of data dimensions
 */
size_t* MatVarInfo::getDims() {
    size_t *pDims;
    pDims = new size_t[rank];
    std::copy(pDimsVector->begin(), pDimsVector->end(), pDims);
    return pDims;
}

MatVarInfo::~MatVarInfo() {
}
}
