/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
/*
 * SamplingMasksData.cpp
 *
 *  Created on: 28 de oct. de 2016
 *      Author: manrod
 */

#include <OpenCLIPER/SamplingMasksData.hpp>
#include <OpenCLIPER/ConcreteNDArray.hpp>

// Uncomment to show class-specific debug messages
#define SAMPLINGMASKSDATA_DEBUG

#if !defined NDEBUG && defined SAMPLINGMASKSDATA_DEBUG
    #define SAMPLINGMASKSDATA_CERR(x) CERR(x)
#else
    #define SAMPLINGMASKSDATA_CERR(x)
    #undef SAMPLINGMASKSDATA_DEBUG
#endif

namespace OpenCLIPER {

/**
 * @brief Constructor that create a SamplingMasksData object from a vector of Data (containing NDArrays, each one contains the
 * sampling mask used for images captured at the same time frame).
 *
 * @param[in] pCLapp pointer to CLapp object (contains an initialized OpenCL environment)
 * @param[in] pMasks pointer to Data object containing sampling mask information (the format used in this input data is the
 * external format for sampling masks, see <a href="#details">detailed description of the class</a>)
 */
SamplingMasksData::SamplingMasksData(std::shared_ptr<CLapp> pCLapp, Data*& pMasks) :
    Data(std::type_index(typeid(dimIndexType))) {
	this->pMasks.reset(pMasks);
	pMasks = nullptr;
        externalToInternalFormat();
        setApp(pCLapp, SYNCSOURCEDEFAULT);
}

/**
 * @brief Constructor tha create a SamplingMasksData object from a vector of NDArrays (each one contains the sampling mask of one frame)
 *
 * @param[in] pCLapp pointer to CLapp object (contains an initialized OpenCL environment)
 * @param[in] pData pointer to vector of NDArrays containing sampling mask information (the format used in this input data is the
 * external format for sampling masks, see <a href="#details">detailed description of the class</a>)
 * @param[in] pDynDims pointer to vector of tempoeral dimensions
 */
SamplingMasksData::SamplingMasksData(std::shared_ptr<CLapp> pCLapp, std::vector<NDArray*>*& pData,
				     std::vector<dimIndexType>*& pDynDims) : Data(pData, pDynDims,
				     std::type_index(typeid(dimIndexType))) {
    externalToInternalFormat();
    setApp(pCLapp, SYNCSOURCEDEFAULT);
}

/**
 * @brief Constructor that creates a SamplingMaskData object from a group if files in raw format (the format used in these files is
 * the external format for sampling masks, see <a href="#details">detailed description of the class</a>)
 *
 * Files of this group have names with the format
 * \<dataFileNamePrefix\>\<dims\>\<framesFileNameSuffix\>\<frameNumber\>\<fileNameExtension\>,
 * where \<dims\> is a string with the format \<width\>x\<height\>x\<depth\> and \<frameNumber\> is the identifier
 * of the temporal frame of image adquisition.

 * @param[in] pCLapp pointer to CLapp object (contains an initialized OpenCL environment)
 * @param[in] dataFileNamePrefix name of the file in raw format
 * @param[in,out] pArraysDims pointer to vectors with spatial dimensions
 * @param[in,out] pDynDims pointer to vector of tempoeral dimensions
 * @param[in] framesFileNameSuffix name suffix for name part depending on frame index
 * @param[in] fileNameExtension extension for the name of the file
 */
SamplingMasksData::SamplingMasksData(std::shared_ptr<CLapp> pCLapp, const std::string dataFileNamePrefix,
				     std::vector<std::vector< dimIndexType >*>*& pArraysDims,
                                     std::vector <dimIndexType>*& pDynDims, const std::string framesFileNameSuffix,
                                     const std::string fileNameExtension) :
    Data(std::type_index(typeid(dimIndexType))) {
    loadRawHostData(dataFileNamePrefix, pArraysDims, pDynDims, framesFileNameSuffix, fileNameExtension);
    externalToInternalFormat();
    setApp(pCLapp, SYNCSOURCEDEFAULT);
}

/**
 * @brief Constructor that creates a SamplingMaskData object from a matlab format file (the format used in this file is the
 * external format for sampling masks, see <a href="#details">detailed description of the class</a>)
 * @param[in] pCLapp pointer to CLapp object (contains an initialized OpenCL environment)
 * @param[in] pMatlabVar pointer to matlab var containing data of the sampling masks
 * @param[in] numOfSpatialDimensions number of spatial dimensions of data
 */
SamplingMasksData::SamplingMasksData(std::shared_ptr<CLapp> pCLapp, matvar_t* pMatlabVar, dimIndexType numOfSpatialDimensions):
    Data(std::type_index(typeid(dimIndexType))) {
    if (pMatlabVar == nullptr) {
	BTTHROW(std::invalid_argument("pointer to matlab variable for sampling masks is nullptr"));
    }
    // Vector of temporal dimensions vector for Data object
    std::vector<dimIndexType>* pDynDims = new std::vector<dimIndexType>;
    // Loop over temporal dimensions for creating temporal dimensions vector for Data object, first temporal dimension comes after
    // last spatial dimension
    for (int i = numOfSpatialDimensions; i < pMatlabVar->rank; i++) {
        pDynDims->push_back(pMatlabVar->dims[i]);
    }
    setDynDims(pDynDims);
    // Number of spatial dimensions data to be read (NDArrays) is the product of all temporal dimensions values
    dimIndexType numOfNDArraysToBeRead = getDynDimsTotalSize();
    SAMPLINGMASKSDATA_CERR("SamplingMasks reading data...\n");
    ((Data *)(this))->loadMatlabHostData(pMatlabVar, numOfSpatialDimensions, numOfNDArraysToBeRead);
    SAMPLINGMASKSDATA_CERR("Done.\n");
    externalToInternalFormat();
    setApp(pCLapp, SYNCSOURCEDEFAULT);
}

/**
 * @brief Constructor that creates a SamplingMasksData object from another SamplingMasksData object (at least temporal
 * dimensions are copied, data is copied if copyData parameter is true).
 * @param[in] pCLapp pointer to CLapp object (contains an initialized OpenCL environment)
 * @param[in] sourceData source SensitivityMapsData object
 * @param[in] copyData if true also data (not only dimensions) are copied from sourceData object to this object
 */
SamplingMasksData::SamplingMasksData(std::shared_ptr<CLapp> pCLapp, const std::shared_ptr<SamplingMasksData> sourceData, bool copyData) :
    Data (sourceData, copyData) {
    if (sourceData->getMasksFormat() == EXTERNAL) {
        externalToInternalFormat();
    }
    setApp(pCLapp, SYNCSOURCEDEFAULT);
}

/**
 * @brief Destructor.
 *
 */
SamplingMasksData::~SamplingMasksData() {
#ifdef SamplingMasksData_DEBUG
    SAMPLINGMASKSDATA_CERR("~SamplingMasksData() begins..." << std::endl);
    SAMPLINGMASKSDATA_CERR("~SamplingMasksData() ends..." << std::endl);
#endif
}

/**
 * @brief Load data of a group of files to NDArray objects (every file contains one sampling mask and is stored into a NDArray object).
 *
 * Files of this group have names with the format
 * \<dataFileNamePrefix\>\<dims\>\<framesFileNameSuffix\>\<frameNumber\>\<fileNameExtension\>,
 * where \<dims\> is a string with the format \<width\>x\<height\>x\<depth\> and \<frameNumber\> is the identifier
 * of the temporal frame of image adquisition.
 * @param[in] fileNamePrefix fixed part of the file name before the variable part
 * @param[in] pArraysDims pointer to a vector of pointers to vectors with the spatial dimensions of every sampling mask
 * @param[in] pDynDims pointer to vector with temporal dimensions (number of frames)
 * @param[in] framesFileNameSuffix part of the file name before the frame number
 * @param[in] fileNameExtension extension for file names
 */
void SamplingMasksData::loadRawHostData(const std::string fileNamePrefix, std::vector<std::vector< dimIndexType >*>*& pArraysDims,
                                        std::vector <dimIndexType>*& pDynDims, const std::string framesFileNameSuffix,
                                        const std::string fileNameExtension) {
    index1DType numFrames = 1;
    for (index1DType i = 0; i<pDynDims->size();i++) {
        numFrames = numFrames * pDynDims->at(i);
    }

    std::stringstream variableSuffixStream;
    std::vector<std::string> fileNameSuffixes;
    for (dimIndexType frame = 0; frame < numFrames; frame++) {
        variableSuffixStream << framesFileNameSuffix << std::setfill('0') << std::setw(2) << frame ; // _frameFF
        fileNameSuffixes.push_back(variableSuffixStream.str());
        variableSuffixStream.str(""); // emtpy string, we don't want to accumulate strings among iterations
    }
    std::string dataFileNamePrefixWithDims = OpenCLIPER::Data::buildFileNamePrefix(fileNamePrefix, pArraysDims->at(0));
    (reinterpret_cast<Data*>(this))->loadRawHostData(dataFileNamePrefixWithDims, pArraysDims, pDynDims, fileNameSuffixes,
                                                     fileNameExtension);
    SAMPLINGMASKSDATA_CERR("SamplingMasksData size: " << getData()->size() << std::endl);
}

/**
 * @brief Converts sampling mask information as a vector of 0 or 1 values (1 if line has been captured, 0 otherwise) to a
 * group of line numbers (only numbers of the lines to be blanked, i.e., marked as not used)
 *
 * External format example for a image of 8 lines: {1, 0, 1, 1, 0, 0, 0, 1}, corresponding internal format {1, 4, 5, 6}.
 */
void SamplingMasksData::externalToInternalFormat() {
    if (masksFormat != EXTERNAL) {
        return;
    }
    //pRowNumbersToBeBlankedVector->resize(0);
    std::vector<dimIndexType>* pListOfRawsToBlankInFrame;
    dimIndexType elementData;
    std::vector<NDArray*>* pNDArraysBlankingInfo = new std::vector<NDArray*>;
    NDArray* pNDArrayBlankingInfo;
    std::vector<dimIndexType>* NDArrayBlankingInfoDims;
    // Number of image lines is the first spatial dimension of the first NDArray (all NDArray with the same number of lines)
    numberOfImageLines = getData()->at(0)->getDims()->at(0);
    for (dimIndexType i = 0; i < getDynDimsTotalSize(); i++) { // loop for all frames
        pListOfRawsToBlankInFrame= new std::vector<dimIndexType>;
        // Only 1 spatial dimension per NDArray
        for (dimIndexType elementIndex = 0; elementIndex < getData()->at(i)->getDims()->at(0); elementIndex ++) {
            elementData = static_cast<const ConcreteNDArray<dimIndexType>*>(getData()->at(i))->getHostData()->at(elementIndex);
            switch (elementData) {
                case 0: // if 0 line has not been captured, it must be blanked
                    pListOfRawsToBlankInFrame->push_back(elementIndex);
                    break;
                case 1: // if 1 line has been captured, it must not be blanked
                    break;
                default:
                    BTTHROW(std::invalid_argument("sampling maks data element must be 0 or 1 (it is " +
                        std::to_string(elementData) + ")"));
            };
        }
        NDArrayBlankingInfoDims = new std::vector<dimIndexType>;
        NDArrayBlankingInfoDims->push_back(pListOfRawsToBlankInFrame->size());
        pNDArrayBlankingInfo = NDArray::createNDArray<dimIndexType>(NDArrayBlankingInfoDims, pListOfRawsToBlankInFrame);
        pNDArraysBlankingInfo->push_back(pNDArrayBlankingInfo);
    }
    // Only data field (group of NDArrays containing data and spatial dimensions) is overwritten, temporal dimensions are not changed.
    setData(pNDArraysBlankingInfo);
    masksFormat = INTERNAL;
}

/**
 * @brief Converts sampling mask information as group of line numbers, starting at 0, (only numbers of the lines to be blanked,
 * marked as not used) to a vector of 0 or 1 values (1 if line has been captured, 0 otherwise).
 *
 * External format example for a image of 8 lines: {1, 0, 1, 1, 0, 0, 0, 1}, corresponding internal format {1, 4, 5, 6}.
 *
 */
void SamplingMasksData::internalToExternalFormat() {
    if (masksFormat != INTERNAL) {
        return;
    }
    std::vector<dimIndexType>* pListOfStatusOfLines;
    dimIndexType elementData, lastLineNotProcessedId;
    std::vector<NDArray*>* pNDArraysExternalFormatBlankingInfo = new std::vector<NDArray*>;
    NDArray* pNDArrayExternalFormatBlankingInfo;
    std::vector<dimIndexType>* NDArrayExternalFormatBlankingInfoDims;
    for (dimIndexType i = 0; i < getDynDimsTotalSize(); i++) { // loop for all frames
        pListOfStatusOfLines = new std::vector<dimIndexType>;
        lastLineNotProcessedId = 0;
        // Only 1 spatial dimension per NDArray
        for (dimIndexType elementIndex = 0; elementIndex < getData()->at(i)->getDims()->at(0); elementIndex++) {
            elementData = static_cast<const ConcreteNDArray<dimIndexType>*>(getData()->at(i))->getHostData()->at(elementIndex);
            for (dimIndexType lineIndex = lastLineNotProcessedId; lineIndex < elementData; lineIndex++) {
                pListOfStatusOfLines->push_back(1); // all lines captured previous to a non-captured line
            }
            pListOfStatusOfLines->push_back(0); // line with number equal to elementData has not been captured;
            lastLineNotProcessedId = elementData + 1;
        }
        /*
         * If the list of ids of lines not captured has been processed and the id of last line not processed is
         * less than number of image lines - 1, we have to add 1s for the range
         * [last_line_not_processed_id, number_of_image_lines-1] to show that this lines have been captured
         */
        for (dimIndexType lineIndex = lastLineNotProcessedId; lineIndex < numberOfImageLines; lineIndex++) {
            pListOfStatusOfLines->push_back(1);
        }
        NDArrayExternalFormatBlankingInfoDims = new std::vector<dimIndexType>;
        NDArrayExternalFormatBlankingInfoDims->push_back(numberOfImageLines);
        pNDArrayExternalFormatBlankingInfo = NDArray::createNDArray<dimIndexType>(NDArrayExternalFormatBlankingInfoDims, pListOfStatusOfLines);
        pNDArraysExternalFormatBlankingInfo->push_back(pNDArrayExternalFormatBlankingInfo);
    }
    // Only data field (group of NDArrays containing data and spatial dimensions) is overwritten, temporal dimensions are not changed.
    setData(pNDArraysExternalFormatBlankingInfo);
    masksFormat = EXTERNAL;
}

} /* namespace OpenCLIPER */
#undef SAMPLINGMASKSDATA_DEBUG
