/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
#include <OpenCLIPER/kernels/hostKernelFunctions.h>

#ifdef __cplusplus
#define global
#define __global
#define uint cl_uint
extern "C" {
#endif  // __CPLUSPLUS

uint roundUp(uint numToRound, uint baseNumber) {
	//assert(baseNumber);
	uint remainder = numToRound % baseNumber;
	PRINTF(("remainder: %d\n", remainder));
	if (remainder == 0) {
		return numToRound;
	} else {
		return numToRound + baseNumber - remainder;
	}
}

inline uint getDimsAndStridesArrayOffsetInBytes(global const void* buffer) {
    // first uint element of input buffer before data start position is offset in bytes for dimensions and strides vector
    uint dataDimsAndStridesArrayOffset = *((global uint*) buffer-1);
    return dataDimsAndStridesArrayOffset;
}

inline global const uint* getDimsAndStridesArrayInBuffer(global const void* buffer) {
    uint dimsAndStridesArrayOffsetInBytes = getDimsAndStridesArrayOffsetInBytes(buffer);
    global const uint* dimsAndStridesArray;
    dimsAndStridesArray = ((global const uint*) buffer) - (dimsAndStridesArrayOffsetInBytes / sizeof(uint));
    return dimsAndStridesArray;
}

inline global const struct dimsAndStridesKnownFields* getDimsAndStridesStruct(global const void* buffer) {
    return (global const struct dimsAndStridesKnownFields*) getDimsAndStridesArrayInBuffer(buffer);
}

inline uint getNumSpatialDims(global const void* buffer) {
    return getDimsAndStridesStruct(buffer)->numSpatialDims;
}

inline uint getNumCoils(global const void* buffer) {
    return getDimsAndStridesStruct(buffer)->numCoils;
}

inline uint getNumTemporalDims(global const void* buffer) {
    return getDimsAndStridesStruct(buffer)->numTemporalDims;
}

inline uint getNumOfNDArrays(global const void* buffer) {
    uint numOfNDArrays = 1;
    uint numCoils = getNumCoils(buffer);
    if (numCoils > 0) {
        numOfNDArrays *= numCoils;
    }
    uint numSpatialDims = getNumTemporalDims(buffer);
    if (numSpatialDims > 0) {
        for (uint i = 0; i < numSpatialDims; i++) {
            numOfNDArrays *= getTemporalDimSize(buffer, i);
        }
    }
    return numOfNDArrays;
}

inline global const uint* getStridesArray(global const void* buffer) {
    global const struct dimsAndStridesKnownFields* pDimsAndStridesStruct = getDimsAndStridesStruct(buffer);
    uint numOfNDArrays = getNumOfNDArrays(buffer);
    uint numOfTempDims = getNumTemporalDims(buffer);
    uint dimsArrayNumOfPos = NUMINITIALPOSITIONSDIMSINFO + numOfTempDims;
    if (pDimsAndStridesStruct->allSizesEqual == 1) {
        dimsArrayNumOfPos += getNumSpatialDims(buffer);
    } else {
        dimsArrayNumOfPos += getNumSpatialDims(buffer) * numOfNDArrays;
    }
    return ((global const uint*) pDimsAndStridesStruct) + dimsArrayNumOfPos;
}

uint getTemporalDimSize(global const void* buffer, uint temporalDimIndex) {
    uint numTemporalDims = getNumTemporalDims(buffer);
    if (temporalDimIndex >= numTemporalDims) {
        // invalid frameDimPos, return 0 as size for it
        return 0;
    }
   uint temporalDimSizePos = FirstTemporalDimPos+temporalDimIndex;
   return getDimsAndStridesArrayInBuffer(buffer)[temporalDimSizePos];
}

uint getSpatialDimSize(global const void* buffer, uint spatialDimIndex, uint NDArray1DIndex) {
    global const struct dimsAndStridesKnownFields* pDimsAndStridesStruct = getDimsAndStridesStruct(buffer);
    uint numSpatialDims = pDimsAndStridesStruct->numSpatialDims;
    if (spatialDimIndex >= numSpatialDims) { // XData
        // invalid spatialDimPos, return 0 as size for it
        return 0;
    }
    uint numTemporalDims = pDimsAndStridesStruct->numTemporalDims;
    uint spatialDimSizePos = FirstTemporalDimPos + numTemporalDims + spatialDimIndex;
    if (pDimsAndStridesStruct->allSizesEqual == 0) {
        spatialDimSizePos += numSpatialDims * NDArray1DIndex;
    }
    return ((global const uint*) pDimsAndStridesStruct)[spatialDimSizePos];
}

uint getDimSize(global const void* buffer, uint dimIndex, uint NDArray1DIndex) {
    uint numSpatialDims = getNumSpatialDims(buffer);
    if (dimIndex < numSpatialDims)
        return getSpatialDimSize(buffer,dimIndex,NDArray1DIndex);
    else {
        dimIndex-=numSpatialDims;
        uint nCoils=getNumCoils(buffer);
        if(nCoils) {
            if(dimIndex==0)
            return nCoils;
            else
            dimIndex--;
        }
        if(dimIndex<getNumTemporalDims(buffer))
            return getTemporalDimSize(buffer,dimIndex);
        else
            return 0; // invalid dimIndex, return 0 as size for it
    }
}

uint getNDArrayTotalSize(global const void* buffer, uint NDArray1DIndex) {
    uint numSpatialDims = getNumSpatialDims(buffer);
    uint acum = 1;
    for (uint spatialDimPos = 0; spatialDimPos < numSpatialDims; spatialDimPos ++) {
        acum *= getSpatialDimSize(buffer, spatialDimPos, NDArray1DIndex);
    }
    return acum;
}

uint getNDArrayStride(global const void* buffer) {
    global const struct dimsAndStridesKnownFields* pDimsAndStridesStruct = getDimsAndStridesStruct(buffer);
    if (pDimsAndStridesStruct->allSizesEqual == 1) { // All NDArrays have the same size, only 1 group of strides valid for every NDArray
        return 0;
    }
    else {
        uint NDArrayStride;
        NDArrayStride = pDimsAndStridesStruct->numSpatialDims + pDimsAndStridesStruct->numTemporalDims;
        if (pDimsAndStridesStruct->numCoils != 0) {
            NDArrayStride += 1;
        }
        return NDArrayStride;
    }
}

uint getCoilStride(global const void* buffer, uint NDArray1DIndex) {
    if (getNumCoils(buffer) == 0) {
        // if number of coills is 0, coil stride does not exist, method returns 0
        return 0;
    }
    uint numSpatialDims = getNumSpatialDims(buffer);
    uint coilStridePos = numSpatialDims + getNDArrayStride(buffer) * NDArray1DIndex;
    return getStridesArray(buffer)[coilStridePos];
}

uint getTemporalDimStride(global const void* buffer, uint temporalDimIndex, uint NDArray1DIndex) {
    global const struct dimsAndStridesKnownFields* pDimsAndStridesStruct = getDimsAndStridesStruct(buffer);
    uint firstTemporalStridePos;
    uint numCoils = pDimsAndStridesStruct->numCoils;
    uint numSpatialDims = pDimsAndStridesStruct->numSpatialDims;
    uint numTemporalDims = pDimsAndStridesStruct->numTemporalDims;
    // Invalid temporalDimIndex, return 0 as stride for it
    if(temporalDimIndex >= numTemporalDims) {
	return 0;
    }
    firstTemporalStridePos = FirstSpatialStridePos + numSpatialDims;
    if (numCoils != 0) {
        firstTemporalStridePos++; // first temporal stride is stored after coil stride
    }
    uint temporalStridePosition = firstTemporalStridePos + temporalDimIndex + getNDArrayStride(buffer) * NDArray1DIndex;
    return getStridesArray(buffer)[temporalStridePosition];
}

uint getSpatialDimStride(global const void* buffer, uint spatialDimIndex, uint NDArray1DIndex) {
    uint spatialDimStridePos = FirstSpatialStridePos + spatialDimIndex + getNDArrayStride(buffer) * NDArray1DIndex;
    return getStridesArray(buffer)[spatialDimStridePos];
}

uint getElementStride(global const void* buffer, uint NDArray1DIndex) {
    uint elementStridePos = ElementStridePos + getNDArrayStride(buffer) * NDArray1DIndex;
    return getStridesArray(buffer)[elementStridePos];
}

uint getDimStride(global const void* buffer, uint dimIndex, uint NDArray1DIndex) {
    global const struct dimsAndStridesKnownFields* pDimsAndStridesStruct = getDimsAndStridesStruct(buffer);    
    uint numSpatialDims = pDimsAndStridesStruct->numSpatialDims;
    if (dimIndex < numSpatialDims)
        return getSpatialDimStride(buffer,dimIndex,NDArray1DIndex);
    else {
	dimIndex-=numSpatialDims;
	uint nCoils=getNumCoils(buffer);
	if(nCoils) {
	    if(dimIndex==0)
		return getCoilStride(buffer, NDArray1DIndex);
	    else
		dimIndex--;
	}
	if(dimIndex<pDimsAndStridesStruct->numTemporalDims)
	    return getTemporalDimStride(buffer,dimIndex,NDArray1DIndex);
	else
	    return 0; // invalid dimIndex, return 0 as stride for it
    }
}

uint getNDArray1DIndex(global const void* buffer, uint coilIndex, uint temporalDimIndexes[]) {
    global const struct dimsAndStridesKnownFields* pDimsAndStridesStruct = getDimsAndStridesStruct(buffer);    
    uint numCoils = pDimsAndStridesStruct->numCoils;
    // coilIndex value 0 can be used to ignore coil dimensions (if numCoils is 0)
    // but coilIndex must be < numCoils if numCoils != 0
    if ((numCoils != 0) && (coilIndex >= numCoils)) {
        // Invalid coil index, returns invalid NDArray 1D-index,
        return -1;
    }
    uint numTemporalDims = pDimsAndStridesStruct->numTemporalDims;
    for (uint temporalDimId = 0; temporalDimId < numTemporalDims; temporalDimId++) {
        if (temporalDimIndexes[temporalDimId] >= getTemporalDimSize(buffer, temporalDimId)) {
            // Invalid temporal dim index for dimension number temporalDimId
            return -1;
        }
    }
    uint index = 0, stride = 1;
    for (uint i = 0; i < numTemporalDims; i++) {
        index += temporalDimIndexes[i] * stride;
        stride *= (&pDimsAndStridesStruct->firstTemporalDimSize)[i];
    }
    if (numCoils != 0) {
        index = coilIndex + (numCoils * index);
    }
    return index;
}

global const void* getNextElement(global const void* buffer, dimIndexType dimIndex, dimIndexType curNDArray, dimIndexType curOffset) {
    global const char * charBuffer = (global const char *) buffer;
    dimIndexType stride = getDimStride(buffer, dimIndex, curNDArray);
    return charBuffer + curOffset + stride;
}


// These functions are to be called from device code only
#ifdef __OPENCL_C_VERSION
void printComplex(float2 complex, char* name) {
        printf("%s: %f + %f j\n", name, complex.s0, complex.s1);
}

void printVector(__constant char name[], float *vector, uint numberOfElements) {
        printf("%s printVector starts...\n", name);
        printf("Number of elements: %u\n", numberOfElements);
        for (uint i = 0; i < numberOfElements - 1; i++) {
                //printf("i: %u\n", i);
                printf("%f, ", vector[i]);
        }
        printf("%f)\n", vector[numberOfElements-1]);
        printf("printVector ends\n");
}
#endif // __OPENCL_C_VERSION

#ifdef __cplusplus
}  //extern "C"
#endif // __CPLUSPLUS
