#include <OpenCLIPER/processes/BlankDataRows.hpp>
#include <OpenCLIPER/CLapp.hpp>

// Uncomment to show class-specific debug messages
//#define BLANKDATAROWS_DEBUG

#if !defined NDEBUG && defined BLANKDATAROWS_DEBUG
    #define BLANKDATAROWS_CERR(x) CERR(x)
#else
#define BLANKDATAROWS_CERR(x)
    #undef BLANKDATAROWS_DEBUG
#endif

namespace OpenCLIPER {

  void BlankDataRows::init() {
    kernel = getApp()->getKernel("blankDataRows_kernel");
  }

  void BlankDataRows::launch(ProfileParameters profileParameters) {
    auto pLP=std::dynamic_pointer_cast<LaunchParameters>(pLaunchParameters);

    cl::Program program;
    cl::Device selected_device;
    cl::CommandQueue queue;
    checkCommonLaunchParameters();
    infoItems.addInfoItem("Title", "BlankDataRows info");

    startProfiling(profileParameters.profilingEnabled);
    try {
        std::shared_ptr<SamplingMasksData> pSamplingMasksData;
        std::vector<cl::Event> kernelsExecEventList;
        selected_device = getApp()->getDevice();
        //cl::Context context = getApp()->getContext();
        queue = getApp()->getCommandQueue();
        cl::Buffer* pInputBuffer;
        cl::Buffer* pSamplingMasksBuffer;
        cl::Event event;

        if (pLP->samplingMasksData == nullptr) {
            BTTHROW(std::invalid_argument ("Empty sampling masks data"));
        }
        pInputBuffer = getInput()->getDeviceBuffer();
        pSamplingMasksData = pLP->samplingMasksData;
        pSamplingMasksBuffer = pSamplingMasksData->getDeviceBuffer();

        uint numFrames = (getInput())->getDynDimsTotalSize();

        BLANKDATAROWS_CERR("NSD AllSE NCoils NTD, TD(0) ... TD(NTD-1) SD(0) ... SD(NSD-1)" << std::endl);
#if !defined NDEBUG && defined BLANKDATAROWS_DEBUG
        PRINTVECTOR("inputDataDims", *getInput()->getDataDimsAndStridesVector(), uint);
#endif

        uint numberOfColumnGroups, numberOfRowsToBeBlanked, numberOfCoils, inputDataNumberOfColumns;
        inputDataNumberOfColumns = NDARRAYWIDTH(getInput()->getData()->at(0));
        for (uint frameId = 0; frameId < numFrames; frameId++) {
            BLANKDATAROWS_CERR("NSD AllSE NCoils NTD, TD(0) ... TD(NTD-1) SD(0) ... SD(NSD-1)" << std::endl);
#if !defined NDEBUG && defined BLANKDATAROWS_DEBUG
            PRINTVECTOR("samplingMasksDimsAndStrides", *pSamplingMasksData->getDataDimsAndStridesVector(), uint);
#endif

            kernel.setArg(0, *pInputBuffer);
            kernel.setArg(1, *pSamplingMasksBuffer);
            kernel.setArg(2, frameId);

            //cl::NDRange globalSizes = {1, 1, 1};

            // Num of groups of columns of size equal to OpenCL vector data type size / 2 (two floats per element)
            numberOfColumnGroups = inputDataNumberOfColumns / 8;
            BLANKDATAROWS_CERR("inputDataNumberOfColumns: " << inputDataNumberOfColumns << std::endl);
            BLANKDATAROWS_CERR("numberOfColumnGroups: " << numberOfColumnGroups << std::endl);
            // Number of columns of the samplingk mask to be zeroed
            numberOfRowsToBeBlanked = NDARRAYWIDTH(pSamplingMasksData->getData()->at(0));
            // Number of coils
            numberOfCoils = (std::dynamic_pointer_cast<KData>(getInput()))->getNCoils();
            cl::NDRange globalSizes = {numberOfColumnGroups, numberOfRowsToBeBlanked, numberOfCoils};
            cl::NDRange localSizes = cl::NDRange();
            BLANKDATAROWS_CERR("enqueueNDRangeKernel " << kernelName << "..." << std::endl);
            queue.enqueueNDRangeKernel(kernel, cl::NullRange, globalSizes, localSizes, NULL, &event);
        }
        BLANKDATAROWS_CERR("done." << std::endl);
        stopProfiling(profileParameters.profilingEnabled);
        if (profileParameters.profilingEnabled) {
            if (profilingSupported) {
            getKernelGroupExecutionTimes(kernelsExecEventList, "OpenCLIPER::DataAndSensitivityMapsProduct::launch kernel",
                                        "OpenCLIPER::DataAndSensitivityMapsProduct::launch group of kernels");
            }
        }
    } catch (cl::Error& err) {
      /////////////////////////////////////////////////////////////////
      // Catch OpenCL errors and print log if it is a build error
      /////////////////////////////////////////////////////////////////
      std::cerr << "ERROR: " << err.what() << " (" << err.err() << ", "
      << CLapp::getOpenCLErrorCodeStr(err.err()) << ")"
      << "\tfile: " << __FILE__ << "\tline: " << __LINE__
      << std::endl;
      if ((err.err() == CL_BUILD_PROGRAM_FAILURE)     || (err.err() == CL_INVALID_KERNEL)) {
	std::cerr << "Extended info: ";
	std::string str = program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(selected_device);
	std::cerr << "Program Info: " << str << std::endl << std::flush;
      }
      throw;
    } catch (std::string msg) {
      std::cerr << "Exception caught in DataAndSensitivityMapsProduct(): " << msg << std::endl;
      throw;
    }
  }
}
