/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belón 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 *
 * FFT.cpp
 *
 *  Created on: 23 de nov. de 2016
 *      Author: fedsim
 */

#include <OpenCLIPER/processes/FFT.hpp>
#include <OpenCLIPER/CLapp.hpp>
#include <OpenCLIPER/hostKernelFunctions.hpp>
#include <LPISupport/InfoItems.hpp>
#include <OpenCLIPER/KData.hpp>
#include <OpenCLIPER/processes/ScalarMultiply.hpp>

// Uncomment to show class-specific debug messages
#define FFT_DEBUG

#if !defined NDEBUG && defined FFT_DEBUG
    #define FFT_CERR(x) CERR(x)
#else
    #define FFT_CERR(x)
    #undef FFT_DEBUG
#endif

namespace OpenCLIPER {

void FFT::init() {

FFT_CERR("FFT::init()\n");
    // Add clFFT error codes to OpenCLIPER error handling. Ideally this should be done
    // only once per CLapp object, not once per FFT object but, anyway...
    getApp()->setOpenCLErrorCodeStr(CLFFT_BUGCHECK,"CLFFT_BUGCHECK");
    getApp()->setOpenCLErrorCodeStr(CLFFT_NOTIMPLEMENTED,"CLFFT_NOTIMPLEMENTED");
    getApp()->setOpenCLErrorCodeStr(CLFFT_TRANSPOSED_NOTIMPLEMENTED,"CLFFT_TRANSPOSED_NOTIMPLEMENTED");
    getApp()->setOpenCLErrorCodeStr(CLFFT_FILE_NOT_FOUND,"CLFFT_FILE_NOT_FOUND");
    getApp()->setOpenCLErrorCodeStr(CLFFT_FILE_CREATE_FAILURE,"CLFFT_FILE_CREATE_FAILURE");
    getApp()->setOpenCLErrorCodeStr(CLFFT_VERSION_MISMATCH,"CLFFT_VERSION_MISMATCH");
    getApp()->setOpenCLErrorCodeStr(CLFFT_INVALID_PLAN,"CLFFT_INVALID_PLAN");
    getApp()->setOpenCLErrorCodeStr(CLFFT_DEVICE_NO_DOUBLE,"CLFFT_DEVICE_NO_DOUBLE");
    getApp()->setOpenCLErrorCodeStr(CLFFT_DEVICE_MISMATCH,"CLFFT_DEVICE_MISMATCH");

#ifdef HAVE_ROCFFT
    // rocfft status codes are positive whereas CL codes are negative. Hopefully they won't clash!
    getApp()->setOpenCLErrorCodeStr(rocfft_status_success,"rocfft_status_success");
    getApp()->setOpenCLErrorCodeStr(rocfft_status_failure,"rocfft_status_failure");
    getApp()->setOpenCLErrorCodeStr(rocfft_status_invalid_arg_value,"rocfft_status_invalid_arg_value");
    getApp()->setOpenCLErrorCodeStr(rocfft_status_invalid_dimensions,"rocfft_status_invalid_dimensions");
    getApp()->setOpenCLErrorCodeStr(rocfft_status_invalid_array_type,"rocfft_status_invalid_array_type");
    getApp()->setOpenCLErrorCodeStr(rocfft_status_invalid_strides,"rocfft_status_invalid_strides");
    getApp()->setOpenCLErrorCodeStr(rocfft_status_invalid_distance,"rocfft_status_invalid_distance");
    getApp()->setOpenCLErrorCodeStr(rocfft_status_invalid_offset,"rocfft_status_invalid_offset");
#endif

    if(!getInput())
	BTTHROW(cl::Error(CL_INVALID_MEM_OBJECT,"OpenCLIPER::FFT::init(): init() called before setInputData()"));

    if(!getOutput())
	BTTHROW(cl::Error(CL_INVALID_MEM_OBJECT,"OpenCLIPER::FFT::init(): init() called before setOutputData()"));

    if (getInput()->getData()->size() != getOutput()->getData()->size())
	BTTHROW(cl::Error(CL_INVALID_MEM_OBJECT,"OpenCLIPER::FFT::launch(): inputData and outputData must have the same number of images"));

    if(!getInput()->getAllSizesEqual())
	BTTHROW(cl::Error(CLFFT_NOTIMPLEMENTED,"OpenCLIPER::FFT::init(): FFT for variable-size data objects is not implemented at this time"));


    const std::vector<dimIndexType>& dimsVector=*(getInput()->getDataDimsAndStridesVector());
    dimIndexType nSpatialDims=dimsVector[NumSpatialDimsPos];
    dimIndexType nTotalDims=nSpatialDims + (dimsVector[NumCoilsPos]>=2?1:0) + dimsVector[NumTemporalDimsPos];

    size_t fftDataSize[3];
    clfftDim clFFTnDims;
    size_t rocFFTnDims;
    unsigned batchSize=1;
    switch(nSpatialDims) {
	case 1:
	    clFFTnDims=CLFFT_1D;
            rocFFTnDims=1;
	    break;
	case 2:
	    clFFTnDims=CLFFT_2D;
            rocFFTnDims=2;
	    break;
	case 3:
	    clFFTnDims=CLFFT_3D;
            rocFFTnDims=3;
	    break;
	default:
	    BTTHROW(cl::Error(CL_INVALID_WORK_DIMENSION,"OpenCLIPER::FFT::init: Only 1, 2 and 3-dimensional FFTs are supported"));
    }

    for(unsigned i=0;i<nSpatialDims;i++)
	/* clFFTDataSize[i]=getSpatialDimSize(static_cast<const cl_uint*> (getInput()->getDataDimsAndStridesHostBuffer()),i,0); */
	fftDataSize[i]=getSpatialDimSize(static_cast<const cl_uint*> (getInput()->getHostBuffer()),i,0);

    for(unsigned i=nSpatialDims;i<nTotalDims;i++)
	/* batchSize*=getDimSize(static_cast<const cl_uint*> (getInput()->getDataDimsAndStridesHostBuffer()),i,0); */
	batchSize*=getDimSize(static_cast<const cl_uint*> (getInput()->getHostBuffer()),i,0);

    //clFFT expects strides in units of the FFT elements (No. of complex numbers), so divide by 2 the value returned by getDimStride
    //getDimStride will return batchDistance=0 if there are spatial dimensions only
    /* size_t batchDistance=getDimStride(static_cast<const cl_uint*> (getInput()->getDataDimsAndStridesHostBuffer()),nSpatialDims,0)/2; */
    size_t batchDistance=getDimStride(static_cast<const cl_uint*> (getInput()->getHostBuffer()),nSpatialDims,0)/2;

    // Use rocFFT if available
#ifdef HAVE_ROCFFT
    if(getApp()->getHIPDevice() != -1) {
        rocfft_result_placement rocfftPlace;
        if(getInput()==getOutput())
            rocfftPlace=rocfft_placement_inplace;
        else
            rocfftPlace=rocfft_placement_notinplace;

        rocfft_status err;

	if((err=rocfft_setup())!=rocfft_status_success) {
            errStr="OpenCLIPER::FFT::init::rocfft_setup: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            BTTHROW(cl::Error(err,errStr.c_str()));
        }

	if((err=rocfft_plan_description_create(&rocfftPlanDescription))!=rocfft_status_success) {
            errStr="OpenCLIPER::FFT::init::rocfft_plan_description_create: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            BTTHROW(cl::Error(err,errStr.c_str()));
        }

	if((err=rocfft_plan_description_set_data_layout(rocfftPlanDescription, rocfft_array_type_complex_interleaved, rocfft_array_type_complex_interleaved, nullptr, nullptr,
                                                0, nullptr, batchDistance,
                                                0, nullptr, batchDistance)
           )!=rocfft_status_success) {
            errStr="OpenCLIPER::FFT::init::rocfft_plan_description_set_data_layout: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            BTTHROW(cl::Error(err,errStr.c_str()));
        }

        // In rocFFT, the transform direction is embedded in the plan, so we create two plans instead of one
	if( ((err=rocfft_plan_create(&rocPlanHandleFW, rocfftPlace, rocfft_transform_type_complex_forward, OPENCLIPER_ROCFFT_PRECISION, rocFFTnDims, fftDataSize, batchSize, rocfftPlanDescription))!=rocfft_status_success) ||
            ((err=rocfft_plan_create(&rocPlanHandleBW, rocfftPlace, rocfft_transform_type_complex_inverse, OPENCLIPER_ROCFFT_PRECISION, rocFFTnDims, fftDataSize, batchSize, rocfftPlanDescription))!=rocfft_status_success)) {
            errStr="OpenCLIPER::FFT::init::rocfft_plan_create: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            BTTHROW(cl::Error(err,errStr.c_str()));
        }

	if( ((err=rocfft_plan_get_work_buffer_size(rocPlanHandleFW, &rocWorkBufferBytesFW))!=rocfft_status_success) ||
            ((err=rocfft_plan_get_work_buffer_size(rocPlanHandleBW, &rocWorkBufferBytesBW))!=rocfft_status_success)) {
            errStr="OpenCLIPER::FFT::init::rocfft_plan_get_work_buffer_size: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            BTTHROW(cl::Error(err,errStr.c_str()));
        }

        // See if a work buffer is required and allocate it. NOTE: if one is required but not allocated and passed to rocfft, a segfault occurs!
        if(rocWorkBufferBytesFW != 0) {
            if((err=rocfft_execution_info_create(&rocExecInfoFW))!=rocfft_status_success) {
                errStr="OpenCLIPER::FFT::init::rocfft_execution_info_create: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
                BTTHROW(cl::Error(err,errStr.c_str()));
            }

            hipError_t hipErr;
            if((hipErr=hipMalloc(&rocWorkBufferFW, rocWorkBufferBytesFW))!=hipSuccess) {
                errStr="OpenCLIPER::FFT::init::hipMalloc: "; errStr+=hipGetErrorString(hipErr);
                BTTHROW(cl::Error(err,errStr.c_str()));
            }

            if((err=rocfft_execution_info_set_work_buffer(rocExecInfoFW, rocWorkBufferFW, rocWorkBufferBytesFW))!=rocfft_status_success) {
                errStr="OpenCLIPER::FFT::init::rocfft_execution_info_set_work_buffer: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
                BTTHROW(cl::Error(err,errStr.c_str()));
            }
        }
        else
            rocExecInfoFW = nullptr;

        if(rocWorkBufferBytesBW != 0) {
            if((err=rocfft_execution_info_create(&rocExecInfoBW))!=rocfft_status_success) {
                errStr="OpenCLIPER::FFT::init::rocfft_execution_info_create: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
                BTTHROW(cl::Error(err,errStr.c_str()));
            }

            hipError_t hipErr;
            if((hipErr=hipMalloc(&rocWorkBufferBW, rocWorkBufferBytesBW))!=hipSuccess) {
                errStr="OpenCLIPER::FFT::init::hipMalloc: "; errStr+=hipGetErrorString(hipErr);
                BTTHROW(cl::Error(err,errStr.c_str()));
            }

            if((err=rocfft_execution_info_set_work_buffer(rocExecInfoBW, rocWorkBufferBW, rocWorkBufferBytesBW))!=rocfft_status_success) {
                errStr="OpenCLIPER::FFT::init::rocfft_execution_info_set_work_buffer: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
                BTTHROW(cl::Error(err,errStr.c_str()));
            }
        }
        else
            rocExecInfoBW = nullptr;

        // Apparently, rocfft generates the plan in the first invocation of rocfft_execute. Force plan creation here so that the first
        // call to launch() does not lag horribly
        void* tmp;
        void* tmp2;
        hipMalloc(&tmp, batchDistance * batchSize * sizeof(cl_float2));
        if(rocfftPlace==rocfft_placement_notinplace)
            hipMalloc(&tmp2, batchDistance * batchSize * sizeof(cl_float2));
        else
            tmp2=nullptr;

        rocfft_execute(rocPlanHandleFW, &tmp, &tmp2, rocExecInfoFW);
        rocfft_execute(rocPlanHandleBW, &tmp, &tmp2, rocExecInfoBW);

        hipFree(tmp);
        if(tmp2) hipFree(tmp2);

        // rocFFT returns values unnormalized (inly in the inverse transform). We need to divide them by the number of elements in each batch
        scalarMultiply=std::make_shared<ScalarMultiply>(getApp());
        scalarMultiply->setInput(getOutput());
        scalarMultiply->setOutput(getOutput());
        scalarMultiply->init();
        scalarMultiplyLP=std::make_shared<ScalarMultiply::LaunchParameters>(1.0/batchDistance);
        scalarMultiply->setLaunchParameters(scalarMultiplyLP);
    }

    // If HIP is unavailable, use plain clFFT
    else {
#endif // HAVE_ROCFFT
        clfftResultLocation clfftPlace;
        if(getInput()==getOutput())
            clfftPlace=CLFFT_INPLACE;
        else
            clfftPlace=CLFFT_OUTOFPLACE;

        /* Setup clFFT. */
        clfftSetupData fftSetup;
        cl_int err;

        if((err=clfftInitSetupData(&fftSetup))!=CL_SUCCESS) {
            errStr="OpenCLIPER::FFT::init: clfftInitSetupData: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            BTTHROW(cl::Error(err,errStr.c_str()));
        }

        if((err=clfftSetup(&fftSetup))!=CL_SUCCESS) {
            errStr="OpenCLIPER::FFT::init: clfftSetup: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            BTTHROW(cl::Error(err,errStr.c_str()));
        }

        //Create a default plan
        if((err=clfftCreateDefaultPlan(&clPlanHandle,(getApp()->getContext())(),clFFTnDims,fftDataSize))!=CL_SUCCESS) {
            errStr="OpenCLIPER::FFT::init: clfftCreateDefaultPlan: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            if(err==CLFFT_NOTIMPLEMENTED)
                errStr+=". Hint: data dimensions must be combinations of powers of 2, 3, 5, and 7";
            BTTHROW(cl::Error(err,errStr.c_str()));
        }

        //Set plan parameters: precision
        if((err=clfftSetPlanPrecision(clPlanHandle,OPENCLIPER_CLFFT_PRECISION))!=CL_SUCCESS) {
            errStr="OpenCLIPER::FFT::init: clfftSetPlanPrecision: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            BTTHROW(cl::Error(err,errStr.c_str()));
        }

        //Set plan parameters: data layout
        if((err=clfftSetLayout(clPlanHandle, CLFFT_COMPLEX_INTERLEAVED, CLFFT_COMPLEX_INTERLEAVED))!=CL_SUCCESS) {
            errStr="OpenCLIPER::FFT::init: clfftSetLayout: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            BTTHROW(cl::Error(err,errStr.c_str()));
        }

        //Set plan parameters: in place/out of place
        if((err=clfftSetResultLocation(clPlanHandle,clfftPlace))!=CL_SUCCESS) {
            errStr="OpenCLIPER::FFT::init: clfftSetResultLocation: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            BTTHROW(cl::Error(err,errStr.c_str()));
        }

        //Set plan parameters: batch size
        if((err=clfftSetPlanBatchSize(clPlanHandle,batchSize))!=CL_SUCCESS) {
            errStr="OpenCLIPER::FFT::init: clfftSetPlanBatchSize: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            BTTHROW(cl::Error(err,errStr.c_str()));
        }

        //Set plan parameters: batch distance
        if((err=clfftSetPlanDistance(clPlanHandle,batchDistance,batchDistance))!=CL_SUCCESS) {
            errStr="OpenCLIPER::FFT::init: clfftSetPlanDistance: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            BTTHROW(cl::Error(err,errStr.c_str()));
        }

        // Bake the plan
        if((err=clfftBakePlan(clPlanHandle, 1, &(getApp()->getCommandQueue(0))(), NULL, NULL))!=CL_SUCCESS) {
            errStr="OpenCLIPER::FFT::init: clfftBakePlan: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            BTTHROW(cl::Error(err,errStr.c_str()));
        }
#ifdef HAVE_ROCFFT
    }
#endif
}

void FFT::launch(ProfileParameters profileParameters) {
    auto pLP=std::dynamic_pointer_cast<LaunchParameters>(pLaunchParameters);
    checkCommonLaunchParameters();
    if(!pLP)
	pLP=std::unique_ptr<LaunchParameters>(new LaunchParameters());

    /* Execute the plan. */

    infoItems.addInfoItem("Title", "FFT info");
    BEGIN_TIME(beginTime);
    std::vector<cl::Event> kernelsExecEventList;
    startKernelProfiling(profileParameters.profilingEnabled);

    // If HIP is available, use it
#ifdef HAVE_ROCFFT
    if(getApp()->getHIPDevice() != -1) {
        rocfft_status err;
        void* inputData=getInput()->getHIPDeviceBuffer();
        void* outputData=getOutput()->getHIPDeviceBuffer();

        // Wait for CL queue before using HIP
        getApp()->getCommandQueue().finish();
        if(pLP->dir == FORWARD) {
            if((err=rocfft_execute(rocPlanHandleFW, &inputData, &outputData, rocExecInfoFW))!=rocfft_status_success) {
                errStr="OpenCLIPER::FFT::launch: clfftEnqueueTransform: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
                BTTHROW(cl::Error(err,errStr.c_str()));
            }
            // Wait for HIP queue before continuing with CL
            hipDeviceSynchronize();
        }
        else {
            if((err=rocfft_execute(rocPlanHandleBW, &inputData, &outputData, rocExecInfoBW))!=rocfft_status_success) {
                errStr="OpenCLIPER::FFT::launch: clfftEnqueueTransform: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
                BTTHROW(cl::Error(err,errStr.c_str()));
            }

            // Wait for HIP queue before continuing with CL
            hipDeviceSynchronize();
            scalarMultiply->setInput(getOutput());
            scalarMultiply->setOutput(getOutput());
            scalarMultiply->launch();
        }
    }
    // If no HIP, use clFFT
    else {
#endif // HAVE_ROCFFT
        cl_int err;
        cl_mem inputData=(*(getInput()->getDeviceBuffer()))();
        cl_mem outputData=(*(getOutput()->getDeviceBuffer()))();

        if((err=clfftEnqueueTransform(clPlanHandle,static_cast<clfftDirection> (pLP->dir),1, &(getApp()->getCommandQueue(0))(),0,nullptr,nullptr,&inputData, &outputData,nullptr))!=CL_SUCCESS) {
            errStr="OpenCLIPER::FFT::launch: clfftEnqueueTransform: "; errStr+=getApp()->getOpenCLErrorCodeStr(err);
            BTTHROW(cl::Error(err,errStr.c_str()));
        }
#ifdef HAVE_ROCFFT
    }
#endif

    stopKernelProfiling(profileParameters.profilingEnabled);
    if (profileParameters.profilingEnabled) {
        END_TIME(endTime);
        TIME_DIFF_TYPE elapsedTime;
        TIME_DIFF(elapsedTime, beginTime, endTime);
        std::ostringstream ostream;
        ostream << std::fixed << std::setprecision(PROFILINGTIMESPRECISION) << elapsedTime;
        infoItems.addInfoItem("Total (host+device) FFT time (s)", ostream.str());
        if (profilingSupported) {
            getKernelGroupExecutionTimes(kernelsExecEventList, "OpenCLIPER::FFT::launch kernel", "OpenCLIPER::FFT::launch group of kernels");
        }
    }
}

FFT::~FFT() {
#ifdef HAVE_ROCFFT
    if(getApp()->getHIPDevice() != -1) {
	//Release plans
        rocfft_plan_destroy(rocPlanHandleFW);
        rocfft_plan_destroy(rocPlanHandleBW);

        if(rocExecInfoFW != nullptr) {
            hipFree(rocWorkBufferFW);
            rocfft_execution_info_destroy(rocExecInfoFW);
        }

        if(rocExecInfoBW != nullptr) {
            hipFree(rocWorkBufferBW);
            rocfft_execution_info_destroy(rocExecInfoBW);
        }

        //Release rocFFT library
	rocfft_cleanup();
    }
    else {
#endif
	//Release the plan
	clfftDestroyPlan(&clPlanHandle);

	//Release clFFT library
	clfftTeardown();
#ifdef HAVE_ROCFFT
    }
#endif
}

} /* namespace OpenCLIPER */

#undef FFT_DEBUG
