/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */

#include <OpenCLIPER/processes/GroupwiseRegistration.hpp>
#define GROUPWISEREGISTRATION_DEBUG

#if !defined NDEBUG && defined GROUPWISEREGISTRATION_DEBUG
    #define GROUPWISEREGISTRATION_CERR(x) CERR(x)
#else
    #define GROUPWISEREGISTRATION_CERR(x)
    #undef GROUPWISEREGISTRATION_DEBUG
#endif

namespace OpenCLIPER {

  void GroupwiseRegistration::init() {

    // Creation and initialization of processes executed in creabspline
    pControlPoints.reset(new ControlPointsLocation(getApp()));
    pCreateBB.reset(new CreateBB(getApp()));
    pPermuteBB.reset(new PermuteBB(getApp()));
    pCreateBBg.reset(new CreateBBg(getApp()));
    pRepmatBBg.reset(new RepmatBBg(getApp()));
    pPermuteBBg.reset(new PermuteBBg(getApp()));

    pControlPoints->init();
    pCreateBB->init();
    pPermuteBB->init();
    pCreateBBg->init();
    pRepmatBBg->init();
    pPermuteBBg->init();

    // Creation and initialization of processes executed in optimizer
    pInitZero.reset(new InitZero(getApp()));
    pInitIT.reset(new CopyDataGPU(getApp()));
    pMetric.reset(new Metric(getApp()));
    pSumLambdaMetric.reset(new SumLambdaMetric(getApp()));
    pGradient.reset(new Gradient(getApp()));
    pPermutedV.reset(new PermutedV(getApp()));
    pUpdateTransformation.reset(new UpdateTransformation(getApp()));
    pCost6D.reset(new Cost6D(getApp()));
    pGradientMetric.reset(new GradientMetric(getApp()));
    pGradientJointAux.reset(new GradientJointAux(getApp()));
    pGradientWithSmoothTerms.reset(new GradientWithSmoothTerms(getApp()));
    pReduction7Dto6D.reset(new Reduction7Dto6D(getApp()));
    pReduction6Dto5D.reset(new ReductionSum(getApp()));
    pReduction5Dto4D.reset(new ReductionSum(getApp()));
    pAuxiliarMask.reset(new AuxiliarMask(getApp()));
    pAuxiliarCost.reset(new AuxiliarCost(getApp()));
    pCost6DReduction.reset(new Cost6DReduction(getApp()));
    pCostReduction.reset(new CostReduction(getApp()));
    pTransformationAux.reset(new TransformationAux(getApp()));
    pDeformation.reset(new Deformation(getApp()));
    pDeformationAdjust.reset(new DeformationAdjust(getApp()));
    pInitdx.reset(new Initdx(getApp()));
    pGradientInterpolator.reset(new GradientInterpolator(getApp()));
    pPermuteTAux.reset(new PermuteTAux(getApp()));
    pShiftTAux.reset(new ShiftTAux(getApp()));
    pRegularization.reset(new Regularization(getApp()));
    pInterpolator.reset(new Interpolator(getApp()));
    pGradientRegularization.reset(new GradientRegularization(getApp()));

    pInitZero->init();
    pInitIT->init();
    pMetric->init();
    pGradient->init();
    pPermutedV->init();
    pUpdateTransformation->init();
    pCost6D->init();
    pGradientMetric->init();
    pGradientJointAux->init();
    pGradientWithSmoothTerms->init();
    pReduction7Dto6D->init();
    pReduction6Dto5D->init();
    pReduction5Dto4D->init();
    pAuxiliarMask->init();
    pAuxiliarCost->init();
    pCost6DReduction->init();
    pCostReduction->init();
    pDeformationAdjust->init();
    pInitdx->init();
    pPermuteTAux->init();
    pShiftTAux->init();
    pGradientRegularization->init();

  }

  GroupwiseRegistration::~GroupwiseRegistration() {
    pControlPoints.reset(nullptr);
    pCreateBB.reset(nullptr);
    pPermuteBB.reset(nullptr);
    pCreateBBg.reset(nullptr);
    pRepmatBBg.reset(nullptr);
    pPermuteBBg.reset(nullptr);
    pInitZero.reset(nullptr);
    pMetric.reset(nullptr);
    pCostReduction.reset(nullptr);
    pGradient.reset(nullptr);
    pGradientMetric.reset(nullptr);
    pTransformationAux.reset(nullptr);
    pDeformation.reset(nullptr);
    pDeformationAdjust.reset(nullptr);
    pInitdx.reset(nullptr);
    pGradientInterpolator.reset(nullptr);
    pGradientRegularization.reset(nullptr);
    pReduction7Dto6D.reset(nullptr);
    pReduction6Dto5D.reset(nullptr);
    pReduction5Dto4D.reset(nullptr);
    pGradientJointAux.reset(nullptr);
    pGradientWithSmoothTerms.reset(nullptr);
    pPermutedV.reset(nullptr);
    pAuxiliarMask.reset(nullptr);
    pAuxiliarCost.reset(nullptr);
    pCost6DReduction.reset(nullptr);
    pCost6D.reset(nullptr);
    pUpdateTransformation.reset(nullptr);
    pPermuteTAux.reset(nullptr);
    pShiftTAux.reset(nullptr);
    pRegularization.reset(nullptr);
    pInitIT.reset(nullptr);
    pInterpolator.reset(nullptr);
    pSumLambdaMetric.reset(nullptr);
  }


  /**
   * @brief  Assign weights to smoothnes terms
   * @param[in] lambda vector containing weights to smoothness terms
   */
  void GroupwiseRegistration::setTerms(const std::vector<float> lambda){
    this->term.lambda = lambda;
  }

  /**
   * @brief Assign Gradient Descent algorithm parameters
   * @param[in] paramNmax Maximum number of iterations
   * @param[in] paramEt Transformation norm threshold
   * @param[in] paramEh Metric variation threshold
   */
  void GroupwiseRegistration::setParametersGD(const unsigned int paramNmax, const float paramEt, const float paramEh){
    this->param.nmax = paramNmax;
    this->param.et = paramEt;
    this->param.eh = paramEh;
  }

  /**
   * @brief Assign images size
   * @param[in] dataSize image size, vector containing rows, columns and frames
   */
  void GroupwiseRegistration::setDataSize(std::vector<cl_uint> dataSize){
    this->dataSize = dataSize;
  }

  /**
   * @brief Sums all lambda terms (four terms)
   * @return sum of all lambda terms
   */
  float GroupwiseRegistration::sumLambda() {
    return std::accumulate(term.lambda.begin(), term.lambda.end(), 0);
  }


  /**
   * @brief Calculates the differences between inputs (norm and y metric) and recast them to adequeate them to output conditions
   * @param[out] Tdif norm difference of the transformation matrix
   * @param[out] Hdif metric variation
   * @param[out] Wn weight
   * @param[in] Dif transformation matrix (displacement)
   * @param[in] H values of the metric for each iteration
   * @param[in] iter current iteration
   * @param[in] flagW flag for adaptive W algorithm
   */
  void GroupwiseRegistration::evolution(float *Tdif, float *Hdif, float *Wn, realType* Dif, realType* H, int iter, bool flagW){
    // Calculus of the differences
    float sum_squares = 0.0f;
    for(uint i=0; i<dim.Dif[0]*dim.Dif[1]*dim.Dif[2]*dim.Dif[3]; i++){
      sum_squares += Dif[i]*Dif[i];
    }
    Tdif[iter-1] = sqrt(sum_squares);       // Norm
    Hdif[iter-1] = H[iter-1] - H[iter];     // Metric
    if(flagW){                              // Adaptation of Wn, flagW
      if(Hdif[iter-1]>0){(*Wn) = (*Wn)*1.2;}else{(*Wn) = (*Wn)/2;}
    }
  }

  /**
   * @brief Draws a circle of a given radius over the ROI in the center of the image
   * @param[out] X matrix representing the circular mask
   * @param[out] bound_box bounding box of X (ROI)
   * @param[in] radius radius of the circle
   * @return total area of the ROI (Region Of Interest)
   */
  int GroupwiseRegistration::setMask(int *X, int *bound_box, const int radius){
    GROUPWISEREGISTRATION_CERR("\n ---> Setting mask with radius " << std::to_string(radius) << "...\n" << std::endl);
    int radius2 = pow(radius,2); // Precomputing the squared radius
    // Center coordinates
    float center[2] = {float(dataSize[1]+1)/2 , float(dataSize[0]+1)/2};
    GROUPWISEREGISTRATION_CERR("\tCenter: [" << std::to_string(center[0]) << ","<< std::to_string(center[1]) << "]" << std::endl);
    // Mesh of the distances to the center of each pixel
    float C[dataSize[0]][dataSize[1]];
    float D[dataSize[0]][dataSize[1]];
    int areaX = 0;
    uint min_fil1=dataSize[0];
    uint max_fil1=0;
    uint min_col1=dataSize[1];
    uint max_col1=0;
    for(uint i=0; i<dataSize[0]; i++)	{ // Rows (height)
      for(uint j=0; j<dataSize[1]; j++)	{ // Columns (width)
	C[i][j] = pow(1-center[0] + j, 2); // Horizontal axis
	D[i][j] = pow(1-center[1] + i, 2); // Vertical axis
	if((C[i][j]+D[i][j])<=radius2){
	  X[i+j*dataSize[0]] = 1;
	  if(i<min_fil1) // Get minimum row index
	    min_fil1=i;
	  if(j<min_col1) // Get minimum column index
	    min_col1=j;
	  if(i>max_fil1) // Get maximum row index
	    max_fil1=i;
	  if(j>max_col1) // Get maximum column index
	    max_col1=j;
	  areaX++; // Area of the ROI
	}
      }
    }
    // Bound_box: limits of the ROI
    bound_box[0] = min_fil1+1;
    bound_box[1] = min_col1+1;
    bound_box[2] = max_fil1+1;
    bound_box[3] = max_col1+1;
    return areaX;
  }

  /**
   * @brief Check the condition to exit the optimization loop
   * @param[in] p1 Tdif
   * @param[in] p3 Hdif
   * @param[in] p2 threshold to compare Tdif
   * @param[in] p4 threshold to compare Hdif
   * @param[in] p5 length
   * @param[in] p6 r2margin length
   * @return True if the output condition is satisfied, false if not
   */
  bool GroupwiseRegistration::stopCondition(const float p1, const float p3, const float p2, const float p4, const int p5, const int p6){
    bool stop = false;
    if((p5==p6) || ((p1<p2)&&(p3<p4)))
      stop = true;
    return stop;
  }

  /**
   * @brief Assigns matrixes dimensions for later use
   * @param[in] ts struct containing the bsplines
   * @param[in] longr1 r1 length
   * @param[in] longr2 r2 length
   * @param[in] longr1margin r1margin length
   * @param[in] longr2margin r2margin length
   * @param[in] select (1|2) -> (creabspline|optimizator)
   */
  void GroupwiseRegistration::setDimensions(const Bspline *ts, const cl_uint longr1, const cl_uint longr2, const cl_uint longr1margin, const cl_uint longr2margin, const unsigned int select){

    switch (select){

      case 1: // Creabspline

	dim.X = {dataSize[0], dataSize[1]};
	dim.x = {dataSize[0], dataSize[1], 2};
	dim.u = {(uint)(ts->C[0][0]+ts->C[0][1]+1), (uint)(ts->C[1][0]+ts->C[1][1]+1), ts->nt};
	dim.pu = dim.u;
	dim.coef = {dataSize[0], 2, ts->nt};
	dim.coefg = {(uint)(ts->C[1][0]+ts->C[1][1]+1), 2, ts->nt};
	dim.BBAux = {dataSize[0], ts->E+1, 3, ts->nt};
	dim.BBAux1 = {(uint)(ts->C[1][0]+ts->C[1][1]+1), (uint)(2*ceil(ts->rp[0])+1), 3, 2};
	dim.auxBB = {ts->E+1, ts->E+1, dataSize[0], dataSize[1], ts->nt}; // auxBB, auxBB1, auxBB2, auxBB11
	dim.auxBBg = {(uint)(2*ceil(ts->rp[1])+1), (uint)(2*ceil(ts->rp[0])+1), (uint)(ts->C[1][0]+ts->C[1][1]+1), (uint)(ts->C[0][0]+ts->C[0][1]+1), 2}; // auxBB1g, auxBB2g, auxBB11g
	dim.aux1BBg = {(uint)(2*ceil(ts->rp[1])+1), (uint)(2*ceil(ts->rp[0])+1), (uint)(ts->C[1][0]+ts->C[1][1]+1), (uint)(ts->C[0][0]+ts->C[0][1]+1), 2, dataSize[2], 2}; // aux1BB1g, aux1BB2g, aux1BB11g
	dim.BB = {dataSize[0], dataSize[1], dataSize[2], ts->nt, ts->E+1, ts->E+1, 2}; // BB, BB1, BB2, BB11
	dim.BBg = {(uint)(2*ceil(ts->rp[1])+1), (uint)(2*ceil(ts->rp[0])+1), dataSize[2], (uint)(ts->C[1][0]+ts->C[1][1]+1), (uint)(ts->C[0][0]+ts->C[0][1]+1), 2, 2}; // BBg, BB1g, BB2g, BB11g
	break;

      case 2: // Optimizer

	dim.V = {dataSize[0], dataSize[1]};
	dim.mean3D = {dataSize[0], dataSize[1]};
	dim.I = dataSize;
	dim.mean4D = {(uint)(ts->C[0][0]+ts->C[0][1]+1), (uint)(ts->C[1][0]+ts->C[1][1]+1), ts->nt};
	dim.T = {(uint)(ts->C[1][0]+ts->C[1][1]+1), (uint)(ts->C[0][0]+ts->C[0][1]+1), dataSize[2], ts->nt};
	dim.Dif = dim.T;
	dim.d = {dataSize[0], dataSize[1], dataSize[2], 2};
	dim.dH = {(uint)(ts->C[0][0]+ts->C[0][1]+1), (uint)(ts->C[1][0]+ts->C[1][1]+1), dataSize[2], ts->nt};
	dim.xn = {dataSize[0], dataSize[1], dataSize[2], ts->nt};
	dim.Aux = {(uint)(2*ceil(ts->rp[1])+1), (uint)(2*ceil(ts->rp[0])+1), dataSize[2], (uint)(ts->C[1][0]+ts->C[1][1]+1), (uint)(ts->C[0][0]+ts->C[0][1]+1)};
	dim.dtau = {dataSize[0], dataSize[1], dataSize[2], ts->nt, 2};
	dim.dV = {(uint)(2*ceil(ts->rp[1])+1), (uint)(2*ceil(ts->rp[0])+1), dataSize[2], (uint)(ts->C[1][0]+ts->C[1][1]+1), (uint)(ts->C[0][0]+ts->C[0][1]+1), ts->nt};
	dim.TAux = {longr2, longr1, dataSize[2], ts->nt, ts->E+1, ts->E+1};
	dim.dtheta = {(uint)(2*ceil(ts->rp[1])+1), (uint)(2*ceil(ts->rp[0])+1), dataSize[2], (uint)(ts->C[1][0]+ts->C[1][1]+1), uint(ts->C[0][0]+ts->C[0][1]+1), ts->nt, 2};
	break;

      default:
	std::cout << " Error: setDimensions() " << std::endl;
	exit(-1);
	break;
    }
  }

  /**
   * @brief Launch a non rigid groupwise registration based on B-splines with motion estimation
   * @param profilingEnabled flag to enable profiling
   */
  void GroupwiseRegistration::launch(ProfileParameters profileParameters) {
    auto pLP=std::dynamic_pointer_cast<LaunchParameters>(pLaunchParameters);

    infoItems.addInfoItem("Title", "GroupwiseRegistration info");
    startProfiling(profileParameters.profilingEnabled);

    GROUPWISEREGISTRATION_CERR("\n\n|=============================================================================|" << std::endl);
    GROUPWISEREGISTRATION_CERR("|\tGROUPWISE REGISTRATION" << std::endl);
    GROUPWISEREGISTRATION_CERR("|=============================================================================|\n" << std::endl);


    ///////////////////////////////////////////////////////////////////////////////////////////
    // Preparing the data                                                                    //
    // Images are complex. To run the code in a proper way we need to compute the modulus,   //
    // in a range between 0 and a maximum we define                                          //
    ///////////////////////////////////////////////////////////////////////////////////////////
    // 1. We need to know the max value over all images                                      //
    ///////////////////////////////////////////////////////////////////////////////////////////
    std::vector<dimIndexType>* pDims = new std::vector<dimIndexType>(*(getInput()->getNDArray(0)->getDims())); // Image size (rows x cols)
    std::vector<std::vector<dimIndexType>*>* pImagesNormDims = new std::vector<std::vector<dimIndexType>*>();
    std::vector<dimIndexType>* pDynNormDims = new std::vector<dimIndexType>({getInput()->getDynDims()->at(0)});

    std::vector<uint>* inputDims = new std::vector<uint>(*(getInput()->getNDArray(0)->getDims()));
    inputDims->push_back(getInput()->getDynDims()->at(0));

    for (index1DType i = 0; i < pDynNormDims->at(0); i++) {
      pImagesNormDims->push_back(new std::vector<dimIndexType>(*pDims));
    }

    std::shared_ptr<Data> normInputData = std::make_shared<XData>(getApp(), pImagesNormDims, pDynNormDims); // Normalized data (only real part)

    // Necesitamos el m�ximo, en m�dulo, de las im�genes originales
    std::vector<float> maxImages(getInput()->getNumNDArrays());
    dimIndexType width, height, depth;
    for (uint i = 0; i < getInput()->getNumNDArrays(); i++) {
      width = NDARRAYWIDTH(getInput()->getNDArray(i));
      height = NDARRAYHEIGHT(getInput()->getNDArray(i));
      depth = NDARRAYDEPTH(getInput()->getNDArray(i));
      if (depth == 0)
	depth = 1;

      std::vector<float> eachImageAbs(width*height);
      complexType* pComplexArray;
      pComplexArray = (complexType*) getInput()->getHostBuffer(i);
      for (dimIndexType index1D = 0;  index1D < height*width*depth; index1D++) {
	eachImageAbs[index1D] = std::fabs(pComplexArray[index1D]);
      }
      maxImages[i] = *max_element(eachImageAbs.begin(), eachImageAbs.end());
    }
    float oldmax = *max_element(maxImages.begin(), maxImages.end());
    float newmax = 700;

    ///////////////////////////////////////////////////////////////////////////////////////////
    // 2. Now we know which is the maximum, in module, overall images, and have to           //
    //    convert images to that range                                                       //
    ///////////////////////////////////////////////////////////////////////////////////////////
    GROUPWISEREGISTRATION_CERR("\n ---> Normalizing data from 0 to " << std::to_string(newmax) << "... ");
    std::unique_ptr<Process> pDataNormalization(new DataNormalization(getApp()));
    pDataNormalization->setInput(getInput());
    pDataNormalization->setOutput(normInputData);
    pDataNormalization->init();
    auto parDataNormalization = std::make_shared<DataNormalization::LaunchParameters>(oldmax,newmax,*inputDims);
    pDataNormalization->setLaunchParameters(parDataNormalization);
    pDataNormalization->launch(profileParameters.profilingEnabled);
    pDataNormalization.reset(nullptr);


    // Size rows, cols, frames (height, width, temporal dimensions)
    setDataSize(*inputDims);
    // Parameters for gradient descent optimization (Max number of iterations for the optimization loop, 0.01 pixels, 0.5% of initial)
    setParametersGD(pLP->nmax,pLP->et,pLP->eh);
    // Set spatio-temporal weigths (1st spatial, 2nd spatial, 1st temp, 2nd temp)
    setTerms(pLP->lambda);


    // Create mask -----------------------------------------------------------------------------
    int* bound_box = new int[4]();
    int* X = new int[dataSize[0]*dataSize[1]](); // Region of interest - ROI
    int areaX = 0;

    if (pLP->radius!=0){ // ROI circular
      areaX = setMask(X, bound_box, pLP->radius);

    }else{ // ROI wholebox
      int Nh = pLP->Dp[0]*(pLP->E+1);
      int Nv = pLP->Dp[1]*(pLP->E+1);
      for(uint i=Nv; i<dataSize[0]-Nv; i++) { // Rows
	for(uint j=Nh; j<dataSize[1]-Nh; j++) { // Columns
	  X[i+j*dataSize[0]] = 1;
	  areaX++;
	}
      }
      bound_box[0] = Nv+1;
      bound_box[1] = dataSize[0]-Nv;
      bound_box[2] = Nh+1;
      bound_box[3] = dataSize[1]-Nh;
    }


    // Compute x --------------------------------------------------------------------------------

    int* x = new int[dataSize[0]*dataSize[1]*2](); // Mesh
    for(uint i=0; i<dataSize[0]; i++){
      for(uint j=0; j<dataSize[1]; j++){
	x[i+j*dataSize[0]+0*dataSize[0]*dataSize[1]] = j+1;
	x[i+j*dataSize[0]+1*dataSize[0]*dataSize[1]] = i+1;
      }
    }

    // CREABSPLINE execution -------------------------------------------------------------
    if (pLP->currentMotionIter==0){
      creabspline(pLP->ts, x, bound_box, pLP->Dp, pLP->E, profileParameters.profilingEnabled);
    }

    // OPTIMIZATOR execution -------------------------------------------------------------
    float Wn = pLP->W/areaX; // Normalization
    optimizer(normInputData, pLP->TData, x, pLP->ts, X, Wn, bound_box, pLP->flagW, pLP->argsMC, profileParameters.profilingEnabled);

    delete[] X;
    delete[] x;

    GROUPWISEREGISTRATION_CERR("\n|=============================================================================|" << std::endl);
    GROUPWISEREGISTRATION_CERR("|\tDONE" << std::endl);
    GROUPWISEREGISTRATION_CERR("|=============================================================================|\n" << std::endl);
    stopProfiling(profileParameters.profilingEnabled);
  }
}
#undef GROUPWISEREGISTRATION_DEBUG
