/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodrí­guez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos MartÃ­n Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
/*
 * SumReduce.cpp
 *
 *  Created on: 7 de mar. de 2019
 *      Author: fedsim
 */

#include <OpenCLIPER/processes/SumReduce.hpp>
#include <OpenCLIPER/CLapp.hpp>
#include <OpenCLIPER/hostKernelFunctions.hpp>
#include <LPISupport/InfoItems.hpp>
#include <OpenCLIPER/XData.hpp>

namespace OpenCLIPER {

void SumReduce::init() {
    auto pIP=std::dynamic_pointer_cast<InitParameters>(pInitParameters);
    if(!pIP) pIP=std::unique_ptr<InitParameters>(new InitParameters());

    if(!getInput())
	throw(cl::Error(CL_INVALID_MEM_OBJECT,"OpenCLIPER::SumReduce::init(): init() called before setInputData()"));

    if(!getOutput())
	throw(cl::Error(CL_INVALID_MEM_OBJECT,"OpenCLIPER::SumReduce::init(): init() called before setOutputData()"));

    if(!getInput()->getAllSizesEqual())
	throw(std::invalid_argument("OpenCLIPER::Normalize::init(): SumReduce for variable-size data objects is not implemented at this time"));

    kernel=getApp()->getKernel("reduce_sum");

    const std::vector<dimIndexType>& dimsVector=*(getInput()->getDataDimsAndStridesVector());
    dimIndexType nSpatialDims=dimsVector[NumSpatialDimsPos];
    dimIndexType nTotalDims=nSpatialDims + (dimsVector[NumCoilsPos]>=2?1:0) + dimsVector[NumTemporalDimsPos];

    // Calculate batch sizes and distances
    batchSize=1;
    for(unsigned i= nSpatialDims; i < nTotalDims; i++)
        batchSize *= getDimSize(static_cast<const cl_uint*> (getInput()->getHostBuffer(0)),i,0);

    //getDimStride will return batchDistance=0 if there are spatial dimensions only
    batchDistance = getDimStride(static_cast<const cl_uint*> (getInput()->getHostBuffer(0)),nSpatialDims,0);

    queue = getApp()->getCommandQueue();

    globalSize = cl::NDRange(NDARRAYWIDTH(getInput()->getData()->at(0)) * NDARRAYHEIGHT(getInput()->getData()->at(0)) * NDARRAYDEPTH(getInput()->getData()->at(0)));
    localSize = CLapp::calcLocalSize(kernel, getApp()->getDevice(), globalSize);
    nWorkgroups = (globalSize[0] - 1) / localSize[0] + 1;

    partialOutputs=std::make_shared<XData> (getApp(), nWorkgroups, std::type_index(typeid(realType)));
    partialOutputs2=std::make_shared<XData> (getApp(), (nWorkgroups > localSize[0])? (nWorkgroups / localSize[0]):1, std::type_index(typeid(realType)));
}

SumReduce::~SumReduce() {
}

void SumReduce::launch(ProfileParameters profileParameters) {
    cl::Buffer* inBuffer=getInput()->getDeviceBuffer();
    cl::Buffer* outBuffer=getOutput()->getDeviceBuffer();

    cl::Buffer* scratchGlobalBuffer=partialOutputs->getDeviceBuffer();		// scratch buffers in global memory
    cl::Buffer* scratchGlobalBuffer2=partialOutputs2->getDeviceBuffer();	// alternately used as input/output

    // First iteration: reduce the whole input data to <nWorkgroups> elements
    // ----------------------------------------------------------------------
    kernel.setArg(0, *inBuffer);					// input (global memory)
    kernel.setArg(1, *scratchGlobalBuffer);				// output (global memory)
    kernel.setArg(2, cl::Local(sizeof(realType) * localSize[0]));	// scratch buffer in local memory
    kernel.setArg(3, batchSize);
    kernel.setArg(4, batchDistance);

    // Launch reduction
    cl_int err;
    if((err=queue.enqueueNDRangeKernel(kernel, cl::NullRange, globalSize, localSize, NULL, NULL))!=CL_SUCCESS)
	throw(cl::Error(err,getApp()->getOpenCLErrorCodeStr(err)));

    // Successive iterations: continue reducing until there is one element left
    // ------------------------------------------------------------------------
    cl::NDRange currentGlobalSize = cl::NDRange(nWorkgroups);
    cl::NDRange currentLocalSize;
    size_t currentNWorkgroups = nWorkgroups;

    kernel.setArg(3, 1); // there is just one batch for successive iterations

    while(currentGlobalSize[0] > 1) {
	currentLocalSize = CLapp::calcLocalSize(kernel, getApp()->getDevice(), currentGlobalSize);
	currentNWorkgroups = currentGlobalSize[0] / currentLocalSize[0];

	// set last output buffer as kernel input
	kernel.setArg(0, *scratchGlobalBuffer);

	// set final output buffer as kernel output if this is the last iteration, or a temporary buffer otherwise
	if(currentNWorkgroups == 1)
	    kernel.setArg(1, *outBuffer);
	else
	    kernel.setArg(1, *scratchGlobalBuffer2);

	// Launch reduction
        if((err=queue.enqueueNDRangeKernel(kernel, cl::NullRange, currentGlobalSize, currentLocalSize, NULL, NULL))!=CL_SUCCESS)
            throw(cl::Error(err,getApp()->getOpenCLErrorCodeStr(err)));

	// set input size for next iteration
	currentGlobalSize=cl::NDRange(currentNWorkgroups);

	// Swap input/output scratch buffers
        cl::Buffer* aux=scratchGlobalBuffer;
        scratchGlobalBuffer=scratchGlobalBuffer2;
        scratchGlobalBuffer2=aux;
    }
}

} /* namespace OpenCLIPER */
