/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
#include <OpenCLIPER/processes/examples/SimpleMRIRecon.hpp>

namespace OpenCLIPER {
/**
 * @brief Method for kernel initialization.
 *
 * It creates and initializes subkernels (FFT, complex element product of x-images by conjugated sensitivity maps and
 * sum of x-images captured by all the coils at the same time frame)
 *
 */
void SimpleMRIRecon::init() {
    pProcInvFFT.reset(new FFT(getApp()));
    pProcInvFFT->setInput(getInput());
    pProcInvFFT->setOutput(getInput());
    pProcInvFFT->init();

    //pProcSensMapProd.reset(new ComplexElementProd(getApp()));
    //pProcSensMapProd->init();

    //pProcAddXImages.reset(new XImageSum(getApp()));
    //pProcAddXImages->init();
}

/**
 * Destructor. Frees resources (pointers to subkernels).
 */
SimpleMRIRecon::~SimpleMRIRecon() {
    pProcInvFFT.reset(nullptr);
    //pProcSensMapProd.reset(nullptr);
    //pProcAddXImages.reset(nullptr);
}

/**
 * @brief Launches a kernel.
 *
 * It sets kernel execution parameters and requests kernel execution 1 or serveral times (according to field of
 * profileParamters struct). If profiling is enabled (according to field of profileParameters struct),
 * kernel execution times are stored.
 * @param[in] profileParameters profiling configuration
 */
void SimpleMRIRecon::launch(ProfileParameters profileParameters) {
    // Step 0: Inverse FFT of initial KData in place
    auto launchParmsInvFFT = std::make_shared<FFT::LaunchParameters>(FFT::BACKWARD);
    pProcInvFFT->setLaunchParameters(launchParmsInvFFT);
    pProcInvFFT->launch(profileParameters);

	//std::dynamic_pointer_cast<KData>(getInput())->iFFT();

	// Step 1: product of x-images (result of step 0) by sensitivity maps, in place
    //KData* pKInData = dynamic_cast<KData*>(getApp()->getData(getInHandle()));
    //std::shared_ptr<SensitivityMapsData> sensMapsData = pKInData->getSensitivityMapsData();
    //pProcSensMapProd->setInput(getInput());
    //pProcSensMapProd->setOutput(getInput());

    // Set parameters: handle of registered sensitivity maps and conjugate sensitivity maps before multiply
    //auto launchParamsSensMapsProd = std::make_shared<ComplexElementProd::LaunchParameters>(ComplexElementProd::conjugate, sensMapsData);
    //pProcSensMapProd->setLaunchParameters(launchParamsSensMapsProd);
    //pProcSensMapProd->launch(profileParameters);
    std::dynamic_pointer_cast<KData>(getInput())->multSensMaps(ComplexElementProd::conjugate);

    // Step 2: addition of all x-images of same frame (number of x-images per frame is the number of coils used for image adquisition)
    //pProcAddXImages->setInput(getInput());
    //pProcAddXImages->setOutput(getOutput());

    //pProcAddXImages->launch(profileParameters);
    std::dynamic_pointer_cast<KData>(getInput())->xImageSum(std::dynamic_pointer_cast<XData>(getOutput()));
}
}
