/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
/*
 * creabspline.cpp
 *
 *  Created on: Apr 8, 2017
 *      Author: Elena Martin Gonzalez
 */
#include <OpenCLIPER/processes/GroupwiseRegistration.hpp>

// Uncomment to show class-specific debug messages
#define CREABSPLINE_DEBUG

#if !defined NDEBUG && defined CREABSPLINE_DEBUG
    #define CREABSPLINE_CERR(x) CERR(x)
#else
    #define CREABSPLINE_CERR(x)
    #undef CREABSPLINE_DEBUG
#endif

namespace OpenCLIPER {

  /**
   * @brief Stores the neccesary elements for the gradient and bspline transformation calculus,
   * B-spline product matrixes with the correspondent coefficients
   * @param[in] x initial mesh
   * @param[in] bound_box boundaries of ROI
   * @param[in] Dp density of control points
   * @param[in] E B-spline order
   * @param[in] profilingEnabled flag specifying profiling
   * @return Struct where the spline variables are stored
   */
  void GroupwiseRegistration::creabspline(Bspline* ts, int *x, int *bound_box, int Dp[], int E, bool profilingEnabled){

    CREABSPLINE_CERR("\n\n CREATING B-SPLINES..." << std::endl);
    CREABSPLINE_CERR("===========================================\n" << std::endl);


    ts->nt = 2;
    ts->Dp[0] = Dp[0];
    ts->Dp[1] = Dp[1];
    ts->E = E; // Orden de splines
    ts->c[0] = float(bound_box[0]+bound_box[2])/2; // Vertical
    ts->c[1] = float(bound_box[1]+bound_box[3])/2; // Horizontal
    ts->rp[0] = (ts->E+1)*ts->Dp[0]/2;// Radio de influencia
    ts->rp[1] = (ts->E+1)*ts->Dp[1]/2;// Radio de influencia

    for (uint l=0; l<ts->nt; l++){ // dimensiones
      for (uint b=0; b<2; b++){ // min-max
	ts->C[l][b] = floor((abs(ts->c[l] - bound_box[l+b*2]) + ts->rp[l]) / ts->Dp[l]);
      }
    }

    // Dimensiones -----------------------------------------------------------------------------------
    setDimensions(ts, 0, 0, 0, 0, 1);


    ts->coef = new int[dim.coef[0]*dim.coef[1]*dim.coef[2]]();
    ts->coefg = new int[dim.coefg[0]*dim.coefg[1]*dim.coefg[2]]();
    int* coef = new int[dim.coef[0]*dim.coef[1]*dim.coef[2]]();


    // Empty XData
    std::vector<dimIndexType>* pDims;
    std::vector<std::vector<dimIndexType>*>* pNDArrayDims;
    std::vector<dimIndexType>* pNDArrayDynDims;

    pDims = new std::vector<dimIndexType>(dim.pu);
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> puData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType)));
    DataHandle puDataHandle = puData->getHandle();

    // Meter en el mismo Data 4 NDArrays que contengan auxBB, auxBB1, auxBB2, auxBB11
    pDims = new std::vector<dimIndexType>({dim.auxBB[0],dim.auxBB[1],dim.auxBB[2],dim.auxBB[3]});
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pDims = new std::vector<dimIndexType>({dim.auxBB});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pDims = new std::vector<dimIndexType>({dim.auxBB});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pDims = new std::vector<dimIndexType>({dim.auxBB[0],dim.auxBB[1],dim.auxBB[2],dim.auxBB[3]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pNDArrayDynDims = new std::vector <dimIndexType>({4});
    std::shared_ptr<Data> auxBBallData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType)));

    // Meter en el mismo Data 4 NDArrays que contengan BB, BB1, BB2, BB11
    pDims = new std::vector<dimIndexType>({dim.BB[0],dim.BB[1],dim.BB[2],dim.BB[3],dim.BB[4],dim.BB[5]});
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pDims = new std::vector<dimIndexType>({dim.BB});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pDims = new std::vector<dimIndexType>({dim.BB});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pDims = new std::vector<dimIndexType>({dim.BB[0],dim.BB[1],dim.BB[2],dim.BB[3],dim.BB[4],dim.BB[5]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pNDArrayDynDims = new std::vector <dimIndexType>({4});
    std::shared_ptr<Data> BBallData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); 
    ts->BBall = BBallData;

    // Meter en el mismo Data 3 NDArrays que contengan auxBB1g, auxBB2g, auxBB11g
    pDims = new std::vector<dimIndexType>(dim.auxBBg);
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pDims = new std::vector<dimIndexType>({dim.auxBBg});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pDims = new std::vector<dimIndexType>({dim.auxBBg[0],dim.auxBBg[1],dim.auxBBg[2],dim.auxBBg[3]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pNDArrayDynDims = new std::vector <dimIndexType>({3});
    std::shared_ptr<Data> auxBBgallData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType)));

    // Meter en el mismo Data 3 NDArrays que contengan aux1BB1g, aux1BB2g, aux1BB11g
    pDims = new std::vector<dimIndexType>(dim.aux1BBg);
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pDims = new std::vector<dimIndexType>(dim.aux1BBg);
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pDims = new std::vector<dimIndexType>({dim.aux1BBg[0],dim.aux1BBg[1],dim.aux1BBg[2],dim.aux1BBg[3],dim.aux1BBg[5],dim.aux1BBg[6]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pNDArrayDynDims = new std::vector <dimIndexType>({3});
    std::shared_ptr<Data> aux1BBgallData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); 

    // Meter en el mismo Data 4 NDArrays que contengan BBg, BB1g, BB2g, BB11g
    pDims = new std::vector<dimIndexType>({dim.BBg[0],dim.BBg[1],dim.BBg[3],dim.BBg[4]});
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pDims = new std::vector<dimIndexType>(dim.BBg);
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pDims = new std::vector<dimIndexType>({dim.BBg});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pDims = new std::vector<dimIndexType>({dim.BBg[0],dim.BBg[1],dim.BBg[2],dim.BBg[3],dim.BBg[4],dim.BBg[5]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pNDArrayDynDims = new std::vector <dimIndexType>({4});
    std::shared_ptr<Data> BBgallData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory
    ts->BBgall = BBgallData;

    cl::Buffer coefobj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_WRITE, dim.coef[0]*dim.coef[1]*dim.coef[2]*sizeof(int), NULL, NULL);
    cl::Buffer coefauxobj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_ONLY, dim.coef[0]*dim.coef[1]*dim.coef[2]*sizeof(int), NULL, NULL);

    // Buffer con contenido
    cl::Buffer cobj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, 2*sizeof(float), ts->c, NULL);
    cl::Buffer Dpobj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, 2*sizeof(int), ts->Dp, NULL);
    cl::Buffer boundboxobj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, 4*sizeof(int), bound_box, NULL);


    pControlPoints->setOutput(puData);

    pPermuteBB->setInput(auxBBallData);
    pPermuteBB->setOutput(BBallData);

    pRepmatBBg->setInput(auxBBgallData);
    pRepmatBBg->setOutput(aux1BBgallData);

    pPermuteBBg->setInput(aux1BBgallData);
    pPermuteBBg->setOutput(BBgallData);


    auto parControlPoints = std::make_shared<ControlPointsLocation::LaunchParameters>(ts->C[0][0], ts->C[1][0], cobj, Dpobj);
    pControlPoints->setLaunchParameters(parControlPoints);
    auto parPermuteBB = std::make_shared<PermuteBB::LaunchParameters>(Dpobj);
    pPermuteBB->setLaunchParameters(parPermuteBB);
    auto parCreateBBg = std::make_shared<CreateBBg::LaunchParameters>(BBgallData, Dpobj);
    pCreateBBg->setLaunchParameters(parCreateBBg);


    // ControlPointsLocation (pu) ---------------------------------------------------------
    // Localizaci�n de los puntos de control en p�xeles, a partir de los �ndices de
    // esos puntos de control
    pControlPoints->launch();

    // Create coef ------------------------------------------------------------------------
    // Coeficientes y matriz de productos para transformacion
    for(cl_uint b=0; b<dim.coef[1]; b++){ // min-max
      if (ts->E%2 != 0){ // Impar - Orden 1 y 3
	for(cl_uint i=0; i<dim.coef[0]; i++){
	  // Fila i, columna b, primera dimension - coef{1}
	  coef[i+b*dim.coef[0]+0*dim.coef[0]*dim.coef[1]] = floor((x[0 + i*dim.x[0] + 0*dim.x[0]*dim.x[1]] - ts->c[0])/ts->Dp[0]) + ((ts->E+pow(-1,b+1))/2)*pow(-1,b+1);
	  // Fila i, columna b, segunda dimension - coef{2}
	  coef[i+b*dim.coef[0]+1*dim.coef[0]*dim.coef[1]] = floor((x[i + 0*dim.x[0] + 1*dim.x[0]*dim.x[1]] - ts->c[1])/ts->Dp[1]) + ((ts->E+pow(-1,b+1))/2)*pow(-1,b+1);
	}
      } else { // Par - Orden 2
	for(cl_uint i=0; i<dim.coef[0]; i++){
	  // Fila i, columna b, primera dimension - coef{1}
	  coef[i+b*dim.coef[0]+0*dim.coef[0]*dim.coef[1]] = round((x[0 + i*dim.x[0] + 0*dim.x[0]*dim.x[1]] - ts->c[0])/ts->Dp[0]) + (ts->E/2)*pow(-1,b+1);
	  // Fila i, columna b, segunda dimension - coef{2}
	  coef[i+b*dim.coef[0]+1*dim.coef[0]*dim.coef[1]] = round((x[i + 0*dim.x[0] + 1*dim.x[0]*dim.x[1]] - ts->c[1])/ts->Dp[1]) + (ts->E/2)*pow(-1,b+1);
	}
      }
    }


    std::vector<realType>* BBAux = new std::vector<realType>(dim.BBAux[0]*dim.BBAux[1]*dim.BBAux[2]*dim.BBAux[3]);
    // Matriz auxiliar para los coeficientes
    for(uint l=0; l<ts->nt; l++){
      int xbis[dataSize[l]];
      for(uint p=0; p<dataSize[l]; p++){
	xbis[p] = p+1;
      }
      // Margen para la imagen para la transformacion
      for(uint n=0; n<dataSize[l]; n++){ // Filas
	for(int k=coef[n+0*dim.coef[0]+l*dim.coef[0]*dim.coef[1]]; k<=coef[n+1*dim.coef[0]+l*dim.coef[0]*dim.coef[1]]; k++){ // Columnas
	  BBAux->at(((k-1)-coef[n+0*dim.coef[0]+l*dim.coef[0]*dim.coef[1]]+1)+n*dim.BBAux[1]+0*dim.BBAux[1]*dim.BBAux[0]+l*dim.BBAux[1]*dim.BBAux[0]*dim.BBAux[2]) = bsplineN(((xbis[n]-ts->c[l])/ts->Dp[l])-k,ts->E);
	  BBAux->at(((k-1)-coef[n+0*dim.coef[0]+l*dim.coef[0]*dim.coef[1]]+1)+n*dim.BBAux[1]+1*dim.BBAux[1]*dim.BBAux[0]+l*dim.BBAux[1]*dim.BBAux[0]*dim.BBAux[2]) = b1splineN(((xbis[n]-ts->c[l])/ts->Dp[l])-k,ts->E);
	  BBAux->at(((k-1)-coef[n+0*dim.coef[0]+l*dim.coef[0]*dim.coef[1]]+1)+n*dim.BBAux[1]+2*dim.BBAux[1]*dim.BBAux[0]+l*dim.BBAux[1]*dim.BBAux[0]*dim.BBAux[2]) = b2splineN(((xbis[n]-ts->c[l])/ts->Dp[l])-k,ts->E);

	}
      }
    }


    pDims = new std::vector<dimIndexType>(dim.BBAux);
    NDArray* pBBAuxNDArray = new ConcreteNDArray<realType>(pDims, BBAux);
    std::vector<realType>* pPixelSize = new std::vector<realType>({1,1,1,1});
    std::vector<NDArray*>* pObjNDArrays = new std::vector<NDArray*>();
    pObjNDArrays->push_back(pBBAuxNDArray);
    std::vector<dimIndexType>* pDynDims = new std::vector<dimIndexType>({1});
    std::shared_ptr<Data> BBAuxData = std::make_shared<XData>(getApp(), pObjNDArrays, pDynDims, pPixelSize, std::type_index(typeid(realType)));


    pCreateBB->setInput(BBAuxData);
    pCreateBB->setOutput(auxBBallData);
    pCreateBB->launch();

    BBAuxData = nullptr;

    // Offset Coef ------------------------------------------------------------------------
    // Add offset to coeficients, used in transformation
    for(cl_uint i=0; i<dim.coef[0]; i++){
      for(cl_uint b=0; b<dim.coef[1]; b++){ // min-max
	for(cl_uint l=0; l<dim.coef[1]; l++){ // min-max
	  ts->coef[i + b*dim.coef[0] + l*dim.coef[0]*dim.coef[1]] = coef[i + b*dim.coef[0] + l*dim.coef[0]*dim.coef[1]] + 1 - coef[(bound_box[l]) + 0*dim.coef[0] + l*dim.coef[0]*dim.coef[1]];
	}
      }
    }

    // Permute BB ------------------------------------------------------------------------
    pPermuteBB->launch();
    auxBBallData = nullptr;

    // Coeficientes y productos matriciales para el c�lculo del gradiente
    // M�todo de coeficientes separados
    getApp()->device2Host(puDataHandle, SyncSource::BUFFER_ONLY);
    realType* pu = (realType *)(getApp()->getData(puDataHandle)->getHostBuffer(0));

    for(uint i=0; i<dim.coefg[0]; i++){
      for(uint j=0; j<dim.coefg[1]; j++){ // min-max
	ts->coefg[i+j*dim.coefg[0]+0*dim.coefg[0]*dim.coefg[1]] = ceil(pu[i+0*dim.pu[1]+0*dim.pu[1]*dim.pu[0]] - ceil(ts->rp[0])) + j*2*ceil(ts->rp[0]);
	ts->coefg[i+j*dim.coefg[0]+1*dim.coefg[0]*dim.coefg[1]] = ceil(pu[0+i*dim.pu[1]+1*dim.pu[1]*dim.pu[0]] - ceil(ts->rp[1])) + j*2*ceil(ts->rp[1]);
      }
    }

    int limxy[2] = {(ts->C[1][0]+ts->C[1][1]+1),(ts->C[0][0]+ts->C[0][1]+1)};
    float* xbis2 = new float[limxy[0]*2];
    for(uint l=0; l<2; l++){
      for(int i=0; i<limxy[l]; i++){
	if(l==0){
	  xbis2[i + 0*limxy[l]] = pu[i + 0*dim.pu[1] + 0*dim.pu[1]*dim.pu[0]];
	} else {
	  xbis2[i + 1*limxy[l]] = pu[0 + i*dim.pu[1] + 1*dim.pu[1]*dim.pu[0]];
	}
      }
    }

    std::vector<realType>* BBAux1 = new std::vector<realType>(dim.BBAux1[0]*dim.BBAux1[1]*dim.BBAux1[2]*dim.BBAux1[3]);
    for(uint n=0; n<dim.coefg[0]; n++){
      uint idxcoefg = n + 0*dim.coefg[0] + 0*dim.coefg[0]*dim.coefg[1];
      for(int k=ts->coefg[n+0*dim.coefg[0]+0*dim.coefg[0]*dim.coefg[1]]; k<=ts->coefg[n+1*dim.coefg[0]+0*dim.coefg[0]*dim.coefg[1]]; k++){
	BBAux1->at(((k-1)-ts->coefg[idxcoefg]+1) + n*dim.BBAux1[1] + 0*dim.BBAux1[1]*dim.BBAux1[0] + 0*dim.BBAux1[1]*dim.BBAux1[0]*dim.BBAux1[2]) = bsplineN(((k-xbis2[n + 0*limxy[0]])/ts->Dp[0]), ts->E);
	BBAux1->at(((k-1)-ts->coefg[idxcoefg]+1) + n*dim.BBAux1[1] + 0*dim.BBAux1[1]*dim.BBAux1[0] + 1*dim.BBAux1[1]*dim.BBAux1[0]*dim.BBAux1[2]) = bsplineN(((k-xbis2[n + 1*limxy[0]])/ts->Dp[1]), ts->E);
	BBAux1->at(((k-1)-ts->coefg[idxcoefg]+1) + n*dim.BBAux1[1] + 1*dim.BBAux1[1]*dim.BBAux1[0] + 0*dim.BBAux1[1]*dim.BBAux1[0]*dim.BBAux1[2]) = b1splineN(((k-xbis2[n + 0*limxy[0]])/ts->Dp[0]), ts->E);
	BBAux1->at(((k-1)-ts->coefg[idxcoefg]+1) + n*dim.BBAux1[1] + 1*dim.BBAux1[1]*dim.BBAux1[0] + 1*dim.BBAux1[1]*dim.BBAux1[0]*dim.BBAux1[2]) = b1splineN(((k-xbis2[n + 1*limxy[0]])/ts->Dp[1]), ts->E);
	BBAux1->at(((k-1)-ts->coefg[idxcoefg]+1) + n*dim.BBAux1[1] + 2*dim.BBAux1[1]*dim.BBAux1[0] + 0*dim.BBAux1[1]*dim.BBAux1[0]*dim.BBAux1[2]) = b2splineN(((k-xbis2[n + 0*limxy[0]])/ts->Dp[0]), ts->E);
	BBAux1->at(((k-1)-ts->coefg[idxcoefg]+1) + n*dim.BBAux1[1] + 2*dim.BBAux1[1]*dim.BBAux1[0] + 1*dim.BBAux1[1]*dim.BBAux1[0]*dim.BBAux1[2]) = b2splineN(((k-xbis2[n + 1*limxy[0]])/ts->Dp[1]), ts->E);
      }
    }
    delete[] xbis2;

    pDims = new std::vector<dimIndexType>(dim.BBAux1);
    NDArray* pBBAux1NDArray = new ConcreteNDArray<realType>(pDims, BBAux1);
    pPixelSize = new std::vector<realType>({1,1,1,1});
    pObjNDArrays = new std::vector<NDArray*>();
    pObjNDArrays->push_back(pBBAux1NDArray);
    pDynDims = new std::vector<dimIndexType>({1});
    std::shared_ptr<Data> BBAux1Data = std::make_shared<XData>(getApp(), pObjNDArrays, pDynDims, pPixelSize, std::type_index(typeid(realType)));

    pCreateBBg->setInput(BBAux1Data);
    pCreateBBg->setOutput(auxBBgallData);
    pCreateBBg->launch();

    // Repmat BBg ------------------------------------------------------------------------
    pRepmatBBg->launch();

    // Permute BBg ------------------------------------------------------------------------
    pPermuteBBg->launch();
    aux1BBgallData = nullptr;

    // Modificaci�n de los coeficientes para evitar bordes
    for(uint i=0; i<dim.coefg[0]; i++){
      for(uint j=0; j<dim.coefg[1]; j++){ // min-max
	for(uint l=0; l<dim.coefg[2]; l++){
	  if(ts->coefg[i+j*dim.coefg[0]+l*dim.coefg[0]*dim.coefg[1]]<1)
	    ts->coefg[i+j*dim.coefg[0]+l*dim.coefg[0]*dim.coefg[1]] = 1;
	  if(ts->coefg[i+j*dim.coefg[0]+l*dim.coefg[0]*dim.coefg[1]]>(int)dataSize[l])
	    ts->coefg[i+j*dim.coefg[0]+l*dim.coefg[0]*dim.coefg[1]] = dataSize[l];
	}
      }
    }

    delete[] coef;


    CREABSPLINE_CERR(" Done.\n" << std::endl);
  }

}

#undef CREABSPLINE_DEBUG
