/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
/*
 * InitZeroRect.cpp
 *
 *  Created on: Apr 27, 2017
 *      Author: Elena Martin Gonzalez
 */
#include <OpenCLIPER/processes/groupwiseRegistration/initialize/InitZeroRect.hpp>

// Uncomment to show class-specific debug messages
//#define INITZERORECT_DEBUG

#if !defined NDEBUG && defined INITZERORECT_DEBUG
    #define INITZERORECT_CERR(x) CERR(x)
#else
    #define INITZERORECT_CERR(x)
    #undef INITZERORECT_DEBUG
#endif

namespace OpenCLIPER {

  void InitZeroRect::init(){
    kernel=getApp()->getKernel("initZeroRect");

    auto pLP=std::dynamic_pointer_cast<LaunchParameters>(pLaunchParameters);

    cl_uint width = NDARRAYWIDTH(getInput()->getNDArray(0));
    cl_uint height = NDARRAYHEIGHT(getInput()->getNDArray(0));
    cl_uint numFrames = getInput()->getDynDimsTotalSize();

    int margin[3] = {(int)ceil(height/40), (int)ceil(width/40), (int)ceil(numFrames/40)};

    int longr1margin = (pLP->bound_box[3]+margin[0]+1)-(pLP->bound_box[1]-margin[0]-1)+1;
    int longr2margin = (pLP->bound_box[2]+margin[1]+1)-(pLP->bound_box[0]-margin[1]-1)+1;

    int r1margin[longr1margin]; // ndice filas bound_box
    int r2margin[longr2margin]; // ndice columnas bound_box

    for(int i=0; i<longr1margin; i++){
      r1margin[i] = (pLP->bound_box[1]-margin[0]-1)+i-1;
    }

    for(int i=0; i<longr2margin; i++){
      r2margin[i] = (pLP->bound_box[0]-margin[1]-1)+i-1;
    }

    r1marginobj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, longr1margin*sizeof(int), r1margin, NULL);
    r2marginobj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, longr2margin*sizeof(int), r2margin, NULL);
  }

  void InitZeroRect::launch(ProfileParameters profileParameters){
    auto pLP=std::dynamic_pointer_cast<LaunchParameters>(pLaunchParameters);

    cl::Program program;
    cl::Device selected_device;
    cl::CommandQueue queue;
    infoItems.addInfoItem("Title", "InitZeroRect info");

    cl_uint width = NDARRAYWIDTH(getInput()->getNDArray(0));
    cl_uint height = NDARRAYHEIGHT(getInput()->getNDArray(0));
    cl_uint numFrames = getInput()->getDynDimsTotalSize();

    int margin[3] = {(int)ceil(height/40), (int)ceil(width/40), (int)ceil(numFrames/40)};

    int longr1margin = (pLP->bound_box[3]+margin[0]+1)-(pLP->bound_box[1]-margin[0]-1)+1;
    int longr2margin = (pLP->bound_box[2]+margin[1]+1)-(pLP->bound_box[0]-margin[1]-1)+1;

    startProfiling(profileParameters.profilingEnabled);
    try {
      std::vector<cl::Event> kernelsExecEventList;
      selected_device = getApp()->getDevice();
      queue = getApp()->getCommandQueue();
      cl::Event event;

      const cl::Buffer* outputData = getOutput()->getDeviceBuffer();

      cl::NDRange globalWorkSize = cl::NDRange(longr1margin,longr2margin,getInput()->getDynDimsTotalSize());

      kernel.setArg( 0, *outputData);
      kernel.setArg( 1, r1marginobj);
      kernel.setArg( 2, r2marginobj);
      //kernel.setArg( 3, NDARRAYHEIGHT(getInput()->getNDArray(0)));
      //kernel.setArg( 4, NDARRAYWIDTH(getInput()->getNDArray(0)));

      queue.enqueueNDRangeKernel(kernel, cl::NullRange, globalWorkSize, cl::NullRange, NULL, &event);
      kernelsExecEventList.push_back(event);
      stopProfiling(profileParameters.profilingEnabled);
      if (profileParameters.profilingEnabled)
	getKernelGroupExecutionTimes(kernelsExecEventList, "OpenCLIPER::InitZeroRect::launch kernel", "OpenCLIPER::InitZeroRect::launch group of kernels");


    } catch (cl::Error& err) {
      /////////////////////////////////////////////////////////////////
      // Catch OpenCL errors and print log if it is a build error
      /////////////////////////////////////////////////////////////////
      std::cerr << "ERROR: " << err.what() << " (" << err.err() << ", "
      << CLapp::getOpenCLErrorCodeStr(err.err()) << ")"
      << "\tfile: " << __FILE__ << "\tline: " << __LINE__
      << std::endl;
      if ((err.err() == CL_BUILD_PROGRAM_FAILURE) || (err.err() == CL_INVALID_KERNEL)) {
	std::cerr << "Extended info: ";
	std::string str = program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(selected_device);
	std::cerr << "Program Info: " << str << std::endl << std::flush;
      }
      throw;
    } catch (std::string msg) {
      std::cerr << "Exception caught in InitZeroRect(): " << msg << std::endl;
      throw;
    }

  }

}
