/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
/*
 * optimizer.cpp
 *
 *  Created on: Apr 8, 2017
 *      Author: Elena Martin Gonzalez
 */
#include <OpenCLIPER/processes/GroupwiseRegistration.hpp>

// Uncomment to show class-specific debug messages
#define OPTIMIZER_DEBUG

#if !defined NDEBUG && defined OPTIMIZER_DEBUG
    #define OPTIMIZER_CERR(x) CERR(x)
#else
    #define OPTIMIZER_CERR(x)
    #undef OPTIMIZER_DEBUG
#endif

namespace OpenCLIPER{

  /**
   * @brief Performs optimization loop
   * @param[out] H metric
   * @param[in] x initial mesh
   * @param[in] ts struct containing the bsplines
   * @param[in] X matrix representing the circular mask
   * @param[in] Wn initial weight
   * @param[in] bound_box bounding box of X
   * @param[in] flagW flag for adaptive W algorithm
   * @param[in] argsMC struct containing 3 objects p
   */
  void GroupwiseRegistration::optimizer(std::shared_ptr<Data> inputData, std::shared_ptr<Data> TData, int *x, Bspline *ts, int *X,
                                        float Wn, int *bound_box, bool flagW, ArgumentsMotionCompensation *argsMC, bool profilingEnabled) {
    auto pLP=std::dynamic_pointer_cast<LaunchParameters>(pLaunchParameters);

    OPTIMIZER_CERR("\n OPTIMIZER..." << std::endl);
    OPTIMIZER_CERR("===========================================\n" << std::endl);

    int iter = 0;

    int longr1 = bound_box[2]-bound_box[0]+1;
    int longr2 = bound_box[3]-bound_box[1]+1;

    int margin[3] = {(int)ceil(dataSize[0]/40), (int)ceil(dataSize[1]/40), (int)ceil(dataSize[2]/40)};
    int longr1margin = (bound_box[3]+margin[0]+1)-(bound_box[1]-margin[0]-1)+1;
    int longr2margin = (bound_box[2]+margin[1]+1)-(bound_box[0]-margin[1]-1)+1;

    // Dimensions -----------------------------------------------------------------------------------
    if (pLP->currentMotionIter!=0){
      setDimensions(ts, 0, 0, 0, 0, 1);
    }
    setDimensions(ts, longr1, longr2, longr1margin, longr2margin, 2);


    const std::vector<dimIndexType>* pImageDims = inputData->getNDArrays()->at(0)->getDims();
    std::vector<dimIndexType>* pDynTransDims = new std::vector<dimIndexType>(*inputData->getDynDims());
    std::vector<std::vector<dimIndexType>*>* pImagesTransDims = new std::vector<std::vector<dimIndexType>*>;
    for (uint i = 0; i < inputData->getNumNDArrays(); i++) {
      pImagesTransDims->push_back(new std::vector<dimIndexType>(*pImageDims));
    }

    OPTIMIZER_CERR("\tCopying initial data to transformed data, just to initialize the object... ");
    std::shared_ptr<Data> transformedData = std::make_shared<XData>(getApp(), pImagesTransDims, pDynTransDims); // only copies dimensions and allocs memory

    std::unique_ptr<Process> pCopy(new CopyDataGPU(getApp()));
    pCopy->setInput(inputData);
    pCopy->setOutput(transformedData);
    pCopy->init();
    pCopy->launch(profilingEnabled);
    pCopy.reset(nullptr);
    OPTIMIZER_CERR("Done.");


    OPTIMIZER_CERR("\n\tAllocating memory... ");
    float* Hdif = new float[getParametersGD().nmax](); // Vector diferencia de costes
    float* Tdif = new float[getParametersGD().nmax](); // Vector diferencia transformacion
    OPTIMIZER_CERR("Done.");

    OPTIMIZER_CERR("\tCreating buffers... ");
    // Empty XData
    std::vector<dimIndexType>* pDims;
    std::vector<std::vector<dimIndexType>*>* pNDArrayDims;
    std::vector<dimIndexType>* pNDArrayDynDims;

    // XData containing the transformation matrix - T
    if (TData == nullptr){
      pDims = new std::vector<dimIndexType>(dim.T);
      pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
      pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
      pNDArrayDynDims = new std::vector <dimIndexType>({1});
      TData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory
    }

    pDims = new std::vector<dimIndexType>(dim.TAux);
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // TAux
    pDims = new std::vector<dimIndexType>({dim.TAux[1],dim.TAux[0],dim.TAux[2],dim.TAux[3],dim.TAux[4],dim.TAux[5]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // TAuxPermuted
    pDims = new std::vector<dimIndexType>({dim.TAux[1],dim.TAux[0],dim.TAux[2],dim.TAux[3],dim.TAux[4],dim.TAux[5]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // TAuxShift1
    pDims = new std::vector<dimIndexType>({dim.TAux[1],dim.TAux[0],dim.TAux[2],dim.TAux[3],dim.TAux[4],dim.TAux[5]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // TAuxShift2
    pNDArrayDynDims = new std::vector <dimIndexType>({4});
    std::shared_ptr<Data> TAuxData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory

    // XData containig the metric - V
    pDims = new std::vector<dimIndexType>(dim.V);
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> VData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory

    // XData containing regularization terms
    pDims = new std::vector<dimIndexType>(dim.dtau);
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // dtaux
    pDims = new std::vector<dimIndexType>(dim.dtau);
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // dtaux2
    pDims = new std::vector<dimIndexType>({dim.dtau[0],dim.dtau[1],dim.dtau[2],dim.dtau[3]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // dtauxy
    pDims = new std::vector<dimIndexType>({dim.dtau[0],dim.dtau[1],dim.dtau[2],dim.dtau[3]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // dtaut
    pDims = new std::vector<dimIndexType>({dim.dtau[0],dim.dtau[1],dim.dtau[2],dim.dtau[3]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // dtaut2
    pNDArrayDynDims = new std::vector <dimIndexType>({5});
    std::shared_ptr<Data> dtauallData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory

    // XData containing regularization terms
    pDims = new std::vector<dimIndexType>(dim.dtheta);
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // dthetax
    pDims = new std::vector<dimIndexType>(dim.dtheta);
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // dthetax2
    pDims = new std::vector<dimIndexType>({dim.dtheta[0],dim.dtheta[1],dim.dtheta[2],dim.dtheta[3],dim.dtheta[4],dim.dtheta[5]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // dthetaxy
    pDims = new std::vector<dimIndexType>({dim.dtheta[0],dim.dtheta[1],dim.dtheta[2],dim.dtheta[3],dim.dtheta[4],dim.dtheta[5]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // dthetat
    pDims = new std::vector<dimIndexType>({dim.dtheta[0],dim.dtheta[1],dim.dtheta[2],dim.dtheta[3],dim.dtheta[4],dim.dtheta[5]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // dthetat2
    pNDArrayDynDims = new std::vector <dimIndexType>({5});
    std::shared_ptr<Data> dthetaallData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory

    // XData containing metric cost
    pDims = new std::vector<dimIndexType>({getParametersGD().nmax+1});
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> HData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory
    HData->device2Host(SyncSource::BUFFER_ONLY);
    realType* Hbuffer = (realType *)(HData->getHostBuffer(0));

    pDims = new std::vector<dimIndexType>({dim.d[0],dim.d[1],dim.d[2]});
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // d1
    pDims = new std::vector<dimIndexType>({dim.d[0],dim.d[1],dim.d[2]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // d2
    pNDArrayDynDims = new std::vector <dimIndexType>({2});
    std::shared_ptr<Data> dallData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory

    pDims = new std::vector<dimIndexType>(dim.d);
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> dxData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory

    pDims = new std::vector<dimIndexType>({dim.d[0],dim.d[1],dim.d[2]});
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> dyData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory

    pDims = new std::vector<dimIndexType>(dim.xn);
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> xnData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory

    /*pDims = new std::vector<dimIndexType>({dim.dtheta[0],dim.dtheta[1],dim.dtheta[2],dim.dtheta[3],dim.dtheta[4],dim.dtheta[5]});
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // aux
    pDims = new std::vector<dimIndexType>({dim.dtheta[0],dim.dtheta[1],dim.dtheta[2],dim.dtheta[3],dim.dtheta[4]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // aux1
    pDims = new std::vector<dimIndexType>({dim.dtheta[0],dim.dtheta[1],dim.dtheta[3],dim.dtheta[4]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // aux2
    pNDArrayDynDims = new std::vector <dimIndexType>({3});
    std::shared_ptr<Data> auxiliarData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory*/
    
    pDims = new std::vector<dimIndexType>({dim.dtheta[0],dim.dtheta[1],dim.dtheta[2],dim.dtheta[3],dim.dtheta[4],dim.dtheta[5]});
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // aux
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> auxData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory
    
    pDims = new std::vector<dimIndexType>({dim.dtheta[0],dim.dtheta[1],dim.dtheta[2],dim.dtheta[3],dim.dtheta[4]});
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // aux1
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> aux1Data = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory
    
    pDims = new std::vector<dimIndexType>({dim.dtheta[0],dim.dtheta[1],dim.dtheta[3],dim.dtheta[4]});
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // aux2
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> aux2Data = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory

    pDims = new std::vector<dimIndexType>(dim.Aux);
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> AuxData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory

    pDims = new std::vector<dimIndexType>(dim.dV);
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // dV
    pDims = new std::vector<dimIndexType>({dim.dV[0],dim.dV[1],dim.dV[3],dim.dV[4],dim.dV[2],dim.dV[5]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // permutedV
    pNDArrayDynDims = new std::vector <dimIndexType>({2});
    std::shared_ptr<Data> dVData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory

    /*pDims = new std::vector<dimIndexType>({dim.dV[0],dim.dV[1],1,1,dim.dV[2],dim.dV[5]});
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // Xaux
    pDims = new std::vector<dimIndexType>({dim.dV[0],dim.dV[1],1,1,dim.dV[2],dim.dV[5]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // cost1
    pDims = new std::vector<dimIndexType>({dim.dV[0],1,1,1,dim.dV[2],dim.dV[5]});
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // cost2
    pNDArrayDynDims = new std::vector <dimIndexType>({3});
    std::shared_ptr<Data> auxiliarCostData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory*/
    
    pDims = new std::vector<dimIndexType>({dim.dV[0],dim.dV[1],dim.dV[2],dim.dV[5]});
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // Xaux
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> XauxData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory
    
    pDims = new std::vector<dimIndexType>({dim.dV[0],dim.dV[1],dim.dV[2],dim.dV[5]});
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // cost1
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> cost1Data = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory
    
    pDims = new std::vector<dimIndexType>({dim.dV[0],dim.dV[2],dim.dV[5]});
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims)); // cost2
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> cost2Data = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory

    pDims = new std::vector<dimIndexType>(dim.dH);
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> dHData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory

    pDims = new std::vector<dimIndexType>(dim.Dif);
    pNDArrayDims = new std::vector<std::vector<dimIndexType>*>();
    pNDArrayDims->push_back(new std::vector<dimIndexType>(*pDims));
    pNDArrayDynDims = new std::vector <dimIndexType>({1});
    std::shared_ptr<Data> DifData = std::make_shared<XData>(getApp(), pNDArrayDims, pNDArrayDynDims, std::type_index(typeid(realType))); // only copies dimensions and allocs memory
    DifData->device2Host(SyncSource::BUFFER_ONLY);
    realType* Difbuffer = (realType *)(DifData->getHostBuffer(0));

    cl::Buffer Xobj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, dim.X[0]*dim.X[1]*sizeof(int), X, NULL);
    cl::Buffer coefgobj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, dim.coefg[0]*dim.coefg[1]*dim.coefg[2]*sizeof(int), ts->coefg, NULL);
    cl::Buffer coefobj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, dim.coef[0]*dim.coef[1]*dim.coef[2]*sizeof(int), ts->coef, NULL);
    cl::Buffer xobj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, dim.x[0]*dim.x[1]*dim.x[2]*sizeof(int), x, NULL);
    cl::Buffer Wnobj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_WRITE|CL_MEM_COPY_HOST_PTR, sizeof(float), &Wn, NULL);
    OPTIMIZER_CERR("Done." << std::endl);


    OPTIMIZER_CERR("\tSetting input in corresponding processes... ");
    pMetric->setInput(transformedData);
    pMetric->setOutput(VData);

    pCostReduction->setInput(VData);
    pCostReduction->setOutput(HData);

    pGradient->setInput(inputData);
    pGradient->setOutput(dallData);

    pGradientMetric->setInput(transformedData);
    pGradientMetric->setOutput(dyData);

    pTransformationAux->setInput(TData);
    pTransformationAux->setOutput(TAuxData);

    pDeformation->setInput(TAuxData);
    pDeformation->setOutput(xnData);

    pDeformationAdjust->setOutput(xnData);

    pInitdx->setInput(dallData);
    pInitdx->setOutput(dxData);

    pGradientInterpolator->setInput(xnData);
    pGradientInterpolator->setOutput(dxData);

    pGradientRegularization->setInput(dtauallData);
    pGradientRegularization->setOutput(dthetaallData);

    pReduction7Dto6D->setInput(dthetaallData);
    pReduction7Dto6D->setOutput(auxData);

    pReduction6Dto5D->setInput(auxData);
    pReduction6Dto5D->setOutput(aux1Data);

    pReduction5Dto4D->setInput(aux1Data);
    pReduction5Dto4D->setOutput(aux2Data);

    pGradientJointAux->setInput(aux2Data);
    pGradientJointAux->setOutput(AuxData);

    pGradientWithSmoothTerms->setInput(AuxData);
    pGradientWithSmoothTerms->setOutput(dVData);

    pPermutedV->setInput(dVData);
    pPermutedV->setOutput(dVData);

    pAuxiliarMask->setInput(dVData);
    pAuxiliarMask->setOutput(XauxData);

    pAuxiliarCost->setInput(dVData);
    pAuxiliarCost->setOutput(cost1Data);

    pCost6DReduction->setInput(cost1Data);
    pCost6DReduction->setOutput(cost2Data);

    pCost6D->setInput(cost2Data);
    pCost6D->setOutput(dHData);

    pUpdateTransformation->setInput(dHData);
    pUpdateTransformation->setOutput(TData);

    pPermuteTAux->setInput(TAuxData);
    pPermuteTAux->setOutput(TAuxData);

    pShiftTAux->setInput(TAuxData);
    pShiftTAux->setOutput(TAuxData);

    pRegularization->setInput(TAuxData);
    pRegularization->setOutput(dtauallData);

    pInitIT->setInput(inputData);
    pInitIT->setOutput(transformedData);

    pInterpolator->setInput(inputData);
    pInterpolator->setOutput(transformedData);

    pSumLambdaMetric->setInput(dtauallData);
    pSumLambdaMetric->setOutput(VData);
    OPTIMIZER_CERR("Done." << std::endl);


    OPTIMIZER_CERR("\tSetting parameters... ");
    auto parTransformationAux = std::make_shared<TransformationAux::LaunchParameters>(coefobj, bound_box, dim.coef[0]);
    pTransformationAux->setLaunchParameters(parTransformationAux);

    auto parDeformation = std::make_shared<Deformation::LaunchParameters>(ts->BBall, bound_box);
    pDeformation->setLaunchParameters(parDeformation);

    auto parDeformationAdjust = std::make_shared<DeformationAdjust::LaunchParameters>(xobj);
    pDeformationAdjust->setLaunchParameters(parDeformationAdjust);

    auto parGradientInterpolator = std::make_shared<GradientInterpolator::LaunchParameters>(dallData, xobj, bound_box);
    pGradientInterpolator->setLaunchParameters(parGradientInterpolator);

    auto parGradientRegularization = std::make_shared<GradientRegularization::LaunchParameters>(coefgobj, ts->BBgall);
    pGradientRegularization->setLaunchParameters(parGradientRegularization);

    auto parReduction7Dto6D = std::make_shared<Reduction7Dto6D::LaunchParameters>(term.lambda[0], term.lambda[1], term.lambda[2], term.lambda[3]);
    pReduction7Dto6D->setLaunchParameters(parReduction7Dto6D);

    auto parReduction6Dto5D = std::make_shared<ReductionSum::LaunchParameters>(5);
    pReduction6Dto5D->setLaunchParameters(parReduction6Dto5D);
    
    auto parReduction5Dto4D = std::make_shared<ReductionSum::LaunchParameters>(2);
    pReduction5Dto4D->setLaunchParameters(parReduction5Dto4D);
    
    auto parGradientWithSmoothTerms = std::make_shared<GradientWithSmoothTerms::LaunchParameters>(dyData, dxData, ts->BBgall, coefgobj);
    pGradientWithSmoothTerms->setLaunchParameters(parGradientWithSmoothTerms);

    auto parCost6DReduction = std::make_shared<Cost6DReduction::LaunchParameters>(dim.dV);
    pCost6DReduction->setLaunchParameters(parCost6DReduction);

    auto parUpdateTransformation = std::make_shared<UpdateTransformation::LaunchParameters>(DifData, Wnobj);
    pUpdateTransformation->setLaunchParameters(parUpdateTransformation);

    auto parRegularization = std::make_shared<Regularization::LaunchParameters>(ts->BBall, bound_box);
    pRegularization->setLaunchParameters(parRegularization);

    auto parSumLambdaMetric = std::make_shared<SumLambdaMetric::LaunchParameters>(term.lambda[0], term.lambda[1], term.lambda[2], term.lambda[3], bound_box);
    pSumLambdaMetric->setLaunchParameters(parSumLambdaMetric);

    auto parInterpolator = std::make_shared<Interpolator::LaunchParameters>(xnData, xobj, bound_box);
    pInterpolator->setLaunchParameters(parInterpolator);
    OPTIMIZER_CERR("Done." << std::endl);


    OPTIMIZER_CERR("\tInitializing processes... ");
    pTransformationAux->init();
    pDeformation->init();
    pGradientInterpolator->init();
    pRegularization->init();
    pInterpolator->init();
    pSumLambdaMetric->init();
    OPTIMIZER_CERR("Done." << std::endl);


    ////////////////////////////////////////////////////////////////////////////////////////////////
    // Start the optimization  ---------------------------------------------------------------------
    ////////////////////////////////////////////////////////////////////////////////////////////////
    
    //pInitZero->setOutput(TData);
    //pInitZero->launch();

    pInitZero->setOutput(dtauallData);
    pInitZero->launch();

    pMetric->launch();

    pCostReduction->setLaunchParameters(std::make_shared<CostReduction::LaunchParameters>(Xobj,iter));
    pCostReduction->launch();
    getApp()->device2Host(HData, SyncSource::BUFFER_ONLY);

    OPTIMIZER_CERR("\n --> Initial cost = " << std::scientific << Hbuffer[0]);
    OPTIMIZER_CERR("\n --> Optimizing...\n" << std::endl);

    pGradient->launch();

    OPTIMIZER_CERR("  =========================================================================== ");
    OPTIMIZER_CERR("\n |  n\t| Cost\t\tWn\tTdif\tHdif\t\t|" << std::endl);

    // Parameters needed for stop condition
    float p2 = getParametersGD().et*dataSize[2]*ts->nt;
    float p4 = Hbuffer[0]*getParametersGD().eh;

    bool stop = false;
    iter = 1;

    std::shared_ptr<AuxiliarMask::LaunchParameters> parAuxiliarMask;
    std::shared_ptr<AuxiliarCost::LaunchParameters> parAuxiliarCost;
    std::shared_ptr<Cost6D::LaunchParameters> parCost6D;

    while (stop==false){

      pInitZero->setOutput(TAuxData); // Initialize TAux
      pInitZero->launch();

      pTransformationAux->launch(); // Output TAux

      pInitZero->setOutput(xnData); // Initialize xn
      pInitZero->launch();

      pDeformation->launch();
      pDeformationAdjust->launch(); // Output xn

      pInitdx->launch();
      pGradientInterpolator->launch(); // Output dx
      pGradientMetric->launch(); // Output dy
      pGradientRegularization->launch(); // Output dtheta

      pReduction7Dto6D->launch(); // Output aux
      pReduction6Dto5D->launch(); // Output aux1
      pReduction5Dto4D->launch(); // Output aux2

      pGradientJointAux->launch(); // Output Aux
      pGradientWithSmoothTerms->launch(); // Output dV

      pPermutedV->launch();

      for(uint k=0; k<dim.dV[3];k++){
	for(uint l=0; l<dim.dV[4];l++){

	  parAuxiliarMask = std::make_shared<AuxiliarMask::LaunchParameters>(Xobj, coefgobj, l, k, dim.coefg[0], dim.coefg[1], dim.X[0]);
	  pAuxiliarMask->setLaunchParameters(parAuxiliarMask);
	  pAuxiliarMask->launch();    // Output XAux

	  parAuxiliarCost = std::make_shared<AuxiliarCost::LaunchParameters>(XauxData, l, k);
	  pAuxiliarCost->setLaunchParameters(parAuxiliarCost);
	  pAuxiliarCost->launch();    // Output cost1

	  pCost6DReduction->launch(); // Output cost2

	  parCost6D = std::make_shared<Cost6D::LaunchParameters>(l, k);
	  pCost6D->setLaunchParameters(parCost6D);
	  pCost6D->launch();          // Output dH
	}
      }

      // Projection of gradient
      pUpdateTransformation->launch();
      DifData->device2Host(SyncSource::BUFFER_ONLY);
      pInitZero->setOutput(TAuxData); // Initialize TAux
      pInitZero->launch();

      pTransformationAux->launch(); // Output TAux

      pInitZero->setOutput(xnData); // Initialize xn
      pInitZero->launch();

      pDeformation->launch();
      pDeformationAdjust->launch(); // Output xn

      pPermuteTAux->launch();
      pShiftTAux->launch();

      pInitZero->setOutput(dtauallData); // Initialize every dtau
      pInitZero->launch();

      pRegularization->launch();

      pInitIT->launch();
      pInterpolator->launch();

      pMetric->launch();
      if(sumLambda() != 0){
	pSumLambdaMetric->launch();
      }

      pCostReduction->setLaunchParameters(std::make_shared<CostReduction::LaunchParameters>(Xobj,iter));
      pCostReduction->launch();
      getApp()->device2Host(HData, SyncSource::BUFFER_ONLY);

      getApp()->getCommandQueue().enqueueReadBuffer(Wnobj, CL_TRUE, 0, sizeof(float), &Wn, NULL, NULL);
      evolution(Tdif, Hdif, &Wn, Difbuffer, Hbuffer, iter, flagW);
      getApp()->getCommandQueue().enqueueWriteBuffer(Wnobj, CL_TRUE, 0, sizeof(float), &Wn, NULL, NULL);

      OPTIMIZER_CERR("  ---------------------------------------------------------------------------- " << std::endl);
      //printf(" |  %d\t| %e\t%.6f\t%.6f\t%e\t|\n", iter, Hbuffer[iter], Wn, Tdif[iter-1], Hdif[iter-1]);
      OPTIMIZER_CERR("|  " << iter << "\t| " << Hbuffer[iter] << "\t" << Wn << "\t" << Tdif[iter-1] << "\t" << Hdif[iter-1] << "\t|\n");
      stop = stopCondition(Tdif[iter-1], Hdif[iter-1], p2, p4, getParametersGD().nmax, iter);
      iter = iter+1;
    }
    OPTIMIZER_CERR("  =========================================================================== " << std::endl);
       
    
    // Struct with necessary arguments to launch the motion compensation
    argsMC->xnData = xnData;
    argsMC->TData = TData;
    argsMC->xobj = xobj;
    argsMC->bound_box = bound_box;

    OPTIMIZER_CERR("\n\tDeleting data...");
    transformedData = nullptr;
    TAuxData = nullptr;
    VData = nullptr;
    dtauallData = nullptr;
    dthetaallData = nullptr;
    HData = nullptr;
    dallData = nullptr;
    dxData = nullptr;
    dyData = nullptr;
    //auxiliarData = nullptr;
    auxData = nullptr;
    aux1Data = nullptr;
    aux2Data = nullptr;
    AuxData = nullptr;
    dVData = nullptr;
    //auxiliarCostData = nullptr;
    XauxData = nullptr;
    cost1Data = nullptr;
    cost2Data = nullptr;
    dHData = nullptr;
    DifData = nullptr;
    OPTIMIZER_CERR(" Done." << std::endl);

    OPTIMIZER_CERR("\tReleasing memory... ");
    delete[] Hdif;
    delete[] Tdif;
    OPTIMIZER_CERR("Done.\n" << std::endl);

  }
}

#undef OPTIMIZER_DEBUG
