/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
/*
 * GradientInterpolator.cpp
 *
 *  Created on: Apr 27, 2017
 *      Author: Elena Martin Gonzalez
 */
#include <OpenCLIPER/processes/groupwiseRegistration/optimizer/GradientInterpolator.hpp>

// Uncomment to show class-specific debug messages
//#define GRADIENTINTERPOLATOR_DEBUG

#if !defined NDEBUG && defined GRADIENTINTERPOLATOR_DEBUG
    #define GRADIENTINTERPOLATOR_CERR(x) CERR(x)
#else
    #define GRADIENTINTERPOLATOR_CERR(x)
    #undef GRADIENTINTERPOLATOR_DEBUG
#endif

namespace OpenCLIPER {

  void GradientInterpolator::init(){
    kernel=getApp()->getKernel("gradientInterpolator");

    auto pLP=std::dynamic_pointer_cast<LaunchParameters>(pLaunchParameters);

    const std::vector<cl_uint> dataSize = *(getOutput()->getNDArray(0)->getDims());
    int margin[3] = {(int)ceil(dataSize[0]/40), (int)ceil(dataSize[1]/40), (int)ceil(dataSize[2]/40)};

    int longr1margin = (pLP->bound_box[3]+margin[0]+1)-(pLP->bound_box[1]-margin[0]-1)+1;
    int longr2margin = (pLP->bound_box[2]+margin[1]+1)-(pLP->bound_box[0]-margin[1]-1)+1;

    int r1margin[longr1margin]; // ndice filas bound_box
    int r2margin[longr2margin]; // ndice columnas bound_box

    for(int i=0; i<longr1margin; i++){
      r1margin[i] = (pLP->bound_box[1]-margin[0]-1)+i-1;
    }

    for(int i=0; i<longr2margin; i++){
      r2margin[i] = (pLP->bound_box[0]-margin[1]-1)+i-1;
    }

    r1marginobj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, longr1margin*sizeof(int), r1margin, NULL);
    r2marginobj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, longr2margin*sizeof(int), r2margin, NULL);
  }

  void GradientInterpolator::launch(ProfileParameters profileParameters){
    auto pLP=std::dynamic_pointer_cast<LaunchParameters>(pLaunchParameters);

    cl::Program program;
    cl::Device selected_device;
    cl::CommandQueue queue;

    startProfiling(profileParameters.profilingEnabled);
    try {
      std::vector<cl::Event> kernelsExecEventList;
      selected_device = getApp()->getDevice();
      queue = getApp()->getCommandQueue();
      cl::Event event;

      const cl::Buffer* d1Data = (pLP->dallData)->getDeviceBuffer(d::d1);
      const cl::Buffer* d2Data = (pLP->dallData)->getDeviceBuffer(d::d2);
      const cl::Buffer* xnData = getInput()->getDeviceBuffer();
      const cl::Buffer* dxData = getOutput()->getDeviceBuffer();

      const std::vector<cl_uint> dataSize = *(getOutput()->getNDArray(0)->getDims());

      
      // Habra que realizar una comprobacin sobre el margen, no sea que se salga de los bordes de la imagen
      int margin[3] = {(int)ceil(dataSize[0]/40), (int)ceil(dataSize[1]/40), (int)ceil(dataSize[2]/40)};

      int longr1margin = (pLP->bound_box[3]+margin[0]+1)-(pLP->bound_box[1]-margin[0]-1)+1;
      int longr2margin = (pLP->bound_box[2]+margin[1]+1)-(pLP->bound_box[0]-margin[1]-1)+1;

      cl::NDRange globalWorkSize = cl::NDRange(longr1margin,longr2margin,dataSize[2]);

      kernel.setArg( 0, *dxData);
      kernel.setArg( 1, *d1Data);
      kernel.setArg( 2, *d2Data);
      kernel.setArg( 3, *xnData);
      kernel.setArg( 4, pLP->xobj);
      kernel.setArg( 5, r1marginobj);
      kernel.setArg( 6, r2marginobj);
      //kernel.setArg( 7, dataSize[0]);
      //kernel.setArg( 8, dataSize[1]);

      queue.enqueueNDRangeKernel(kernel, cl::NullRange, globalWorkSize, cl::NullRange, NULL, &event);
      kernelsExecEventList.push_back(event);
      stopProfiling(profileParameters.profilingEnabled);
      if (profileParameters.profilingEnabled)
	getKernelGroupExecutionTimes(kernelsExecEventList, "OpenCLIPER::GradientInterpolator::launch kernel", "OpenCLIPER::GradientInterpolator::launch group of kernels");

    } catch (cl::Error& err) {
      /////////////////////////////////////////////////////////////////
      // Catch OpenCL errors and print log if it is a build error
      /////////////////////////////////////////////////////////////////
      std::cerr << "ERROR: " << err.what() << " (" << err.err() << ", "
      << CLapp::getOpenCLErrorCodeStr(err.err()) << ")"
      << "\tfile: " << __FILE__ << "\tline: " << __LINE__
      << std::endl;
      if ((err.err() == CL_BUILD_PROGRAM_FAILURE) || (err.err() == CL_INVALID_KERNEL)) {
	std::cerr << "Extended info: ";
	std::string str = program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(selected_device);
	std::cerr << "Program Info: " << str << std::endl << std::flush;
      }
      throw;
    } catch (std::string msg) {
      std::cerr << "Exception caught in GradientInterpolator(): " << msg << std::endl;
      throw;
    }

  }

}

#undef GRADIENTINTERPOLATOR_DEBUG
