/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
/*
 * SumLambdaMetric.cpp
 *
 *  Created on: Apr 27, 2017
 *      Author: Elena Martin Gonzalez
 */
#include <OpenCLIPER/processes/groupwiseRegistration/optimizer/SumLambdaMetric.hpp>

// Uncomment to show class-specific debug messages
//#define SUMLAMBDAMETRIC_DEBUG

#if !defined NDEBUG && defined SUMLAMBDAMETRIC_DEBUG
    #define SUMLAMBDAMETRIC_CERR(x) CERR(x)
#else
    #define SUMLAMBDAMETRIC_CERR(x)
    #undef SUMLAMBDAMETRIC_DEBUG
#endif

namespace OpenCLIPER {

  void SumLambdaMetric::init(){
    kernel=getApp()->getKernel("sumLambdaMetric");

    auto pLP=std::dynamic_pointer_cast<LaunchParameters>(pLaunchParameters);

    int longr1 = pLP->bound_box[2]-pLP->bound_box[0]+1;
    int longr2 = pLP->bound_box[3]-pLP->bound_box[1]+1;

    int r1[longr1]; // ndice filas bound_box
    int r2[longr2]; // ndice columnas bound_box

    for(int i=0; i<longr1; i++){
      r1[i] = (pLP->bound_box[0]-1)+i;
    }

    for(int i=0; i<longr2; i++){
      r2[i] = (pLP->bound_box[1]-1)+i;
    }

    r1obj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, longr1*sizeof(int), r1, NULL);
    r2obj = cl::Buffer(getApp()->getContext(), CL_MEM_READ_ONLY|CL_MEM_COPY_HOST_PTR, longr2*sizeof(int), r2, NULL);

  }

  void SumLambdaMetric::launch(ProfileParameters profileParameters){
    auto pLP=std::dynamic_pointer_cast<LaunchParameters>(pLaunchParameters);

    cl::Program program;
    cl::Device selected_device;
    cl::CommandQueue queue;

    startProfiling(profileParameters.profilingEnabled);
    try {
      std::vector<cl::Event> kernelsExecEventList;
      selected_device = getApp()->getDevice();
      queue = getApp()->getCommandQueue();
      cl::Event event;

      int longr1 = pLP->bound_box[2]-pLP->bound_box[0]+1;
      int longr2 = pLP->bound_box[3]-pLP->bound_box[1]+1;

      const cl::Buffer* VData = getOutput()->getDeviceBuffer();

      const cl::Buffer* dtauxData = getInput()->getDeviceBuffer(dtau::dtaux);
      const cl::Buffer* dtaux2Data = getInput()->getDeviceBuffer(dtau::dtaux2);
      const cl::Buffer* dtauxyData = getInput()->getDeviceBuffer(dtau::dtauxy);
      const cl::Buffer* dtautData = getInput()->getDeviceBuffer(dtau::dtaut);
      const cl::Buffer* dtaut2Data = getInput()->getDeviceBuffer(dtau::dtaut2);

      const std::vector<cl_uint> sizeV = *(getOutput()->getNDArray(0)->getDims());
      const std::vector<cl_uint> sizedtau = *(getInput()->getNDArray(dtau::dtaux)->getDims());

      cl::NDRange globalWorkSize = cl::NDRange(longr1,longr2);

      kernel.setArg( 0,  *VData);
      kernel.setArg( 1,  sizedtau[2]);
      kernel.setArg( 2,  sizedtau[3]);
      kernel.setArg( 3,  *dtauxData);
      kernel.setArg( 4,  *dtaux2Data);
      kernel.setArg( 5,  *dtauxyData);
      kernel.setArg( 6,  *dtautData);
      kernel.setArg( 7,  *dtaut2Data);
      kernel.setArg( 8,  r1obj);
      kernel.setArg( 9,  r2obj);
      kernel.setArg( 10, pLP->term0);
      kernel.setArg( 11, pLP->term1);
      kernel.setArg( 12, pLP->term2);
      kernel.setArg( 13, pLP->term3);
      //kernel.setArg( 14, sizeV[0]);
      //kernel.setArg( 15, sizeV[1]);

      queue.enqueueNDRangeKernel(kernel, cl::NullRange, globalWorkSize, cl::NullRange, NULL, &event);
      kernelsExecEventList.push_back(event);
      stopProfiling(profileParameters.profilingEnabled);
      if (profileParameters.profilingEnabled)
	getKernelGroupExecutionTimes(kernelsExecEventList, "OpenCLIPER::SumLambdaMetric::launch kernel", "OpenCLIPER::SumLambdaMetric::launch group of kernels");

    } catch (cl::Error& err) {
      /////////////////////////////////////////////////////////////////
      // Catch OpenCL errors and print log if it is a build error
      /////////////////////////////////////////////////////////////////
      std::cerr << "ERROR: " << err.what() << " (" << err.err() << ", "
      << CLapp::getOpenCLErrorCodeStr(err.err()) << ")"
      << "\tfile: " << __FILE__ << "\tline: " << __LINE__
      << std::endl;
      if ((err.err() == CL_BUILD_PROGRAM_FAILURE) || (err.err() == CL_INVALID_KERNEL)) {
	std::cerr << "Extended info: ";
	std::string str = program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(selected_device);
	std::cerr << "Program Info: " << str << std::endl << std::flush;
      }
      throw;
    } catch (std::string msg) {
      std::cerr << "Exception caught in SumLambdaMetric(): " << msg << std::endl;
      throw;
    }

  }

}

#undef SUMLAMBDAMETRIC_DEBUG
