/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodríguez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
/*
 * NestaUp.cpp
 *
 *  Created on: 20 de nov. de 2017
 *      Author: Elisa Moya-Saez
 */

#include <OpenCLIPER/processes/nesta/NestaUp.hpp>

// Uncomment to show class-specific debug messages
#define NESTAUP_DEBUG

#if !defined NDEBUG && defined NESTAUP_DEBUG
    #define NESTAUP_CERR(x) CERR(x)
#else
    #define NESTAUP_CERR(x)
    #undef NESTAUP_DEBUG
#endif

namespace OpenCLIPER {

  void NestaUp::init() {

    std::shared_ptr<Data> pAuxFFT = std::make_shared<KData>(getApp(), std::dynamic_pointer_cast<KData>(getInput()), false, false);

    pFFTOutOfPlace.reset(new FFT(getApp()));
    pFFTOutOfPlace->setInput(getInput());
    pFFTOutOfPlace->setOutput(pAuxFFT);
    pFFTOutOfPlace->init();

    pDataAndSensitivityMapsProduct.reset(new ComplexElementProd(getApp()));
    pDataAndSensitivityMapsProduct->init();

    pXImagesAllCoilSameFrameAddition.reset(new XImageSum(getApp()));
    pXImagesAllCoilSameFrameAddition->init();

    pDataAndSamplingMasksProduct.reset(new BlankDataRows(getApp()));
    pDataAndSamplingMasksProduct->init();

    pTemporalTV.reset(new TemporalTV(getApp()));
    pTemporalTV->setInitParameters(std::make_shared<TemporalTV::InitParameters>(TemporalTV::FORWARD));
    pTemporalTV->init();

    pTemporalTVt.reset(new TemporalTV(getApp()));
    pTemporalTVt->setInitParameters(std::make_shared<TemporalTV::InitParameters>(TemporalTV::ADJOINT));
    pTemporalTVt->init();

    pVectorNormalization.reset(new VectorNormalization(getApp()));
    pVectorNormalization->init();

    pMotionCompensation.reset(new MotionCompensation(getApp()));
    pAdjointMotionCompensation.reset(new AdjointMotionCompensation(getApp()));

    pCopy.reset(new CopyDataGPU(getApp()));
    pCopy->init();

  }


  NestaUp::~NestaUp() {
    pVectorNormalization.reset(nullptr);
    pFFTOutOfPlace.reset(nullptr);
    pDataAndSensitivityMapsProduct.reset(nullptr);
    pXImagesAllCoilSameFrameAddition.reset(nullptr);
    pDataAndSamplingMasksProduct.reset(nullptr);
    pTemporalTV.reset(nullptr);
    pTemporalTVt.reset(nullptr);
    pCopy.reset(nullptr);
    pMotionCompensation.reset(nullptr);
    pAdjointMotionCompensation.reset(nullptr);
  }


  /**
   * Adjoint Encoding operator.
   * @param[in] inputDataHandle KData [Nx*Ny*numFrames*numCoils]
   * @param[in] sensitivityMapsDataHandle Sensitivity maps
   * @param[out] outputDataHandle XData [Nx*Ny*numFrames]
   * @param[in] auxDataHandleFFT Axiliar DataHandle to the output of inverse FFT
   * @param[in] profilingEnabled
   */
  void NestaUp::operatorAt(std::shared_ptr<Data> inputData, std::shared_ptr<SensitivityMapsData> sensitivityMapsData,
                           std::shared_ptr<Data> outputData, std::shared_ptr<Data> auxDataFFT, bool profilingEnabled){

    try {
      // IFFT out of place (backward FFT)
      pFFTOutOfPlace->setInput(inputData);
      pFFTOutOfPlace->setOutput(auxDataFFT);
      pFFTOutOfPlace->setLaunchParameters(std::make_shared<FFT::LaunchParameters>(FFT::BACKWARD));
      pFFTOutOfPlace->launch();

      // XImage*Conj(SensMap) in place
      pDataAndSensitivityMapsProduct->setInput(auxDataFFT);
      pDataAndSensitivityMapsProduct->setOutput(auxDataFFT);
      pDataAndSensitivityMapsProduct->setLaunchParameters(std::make_shared<ComplexElementProd::LaunchParameters>(ComplexElementProd::conjugate,
                                                                                                            sensitivityMapsData));
      pDataAndSensitivityMapsProduct->launch();

      if ((std::dynamic_pointer_cast<KData>(inputData))->getNCoils() != 1) {
	pXImagesAllCoilSameFrameAddition->setInput(auxDataFFT);
	pXImagesAllCoilSameFrameAddition->setOutput(outputData);
	pXImagesAllCoilSameFrameAddition->launch();
      } else {
	pCopy->setInput(auxDataFFT);
	pCopy->setOutput(outputData);
	pCopy->launch();
      }

    } catch (cl::Error& err) {
      /////////////////////////////////////////////////////////////////
      // Catch OpenCL errors and print log if it is a build error
      /////////////////////////////////////////////////////////////////
      std::cerr << "ERROR: " << err.what() << " (" << err.err() << ", "
      << CLapp::getOpenCLErrorCodeStr(err.err()) << ")"
      << "\tfile: " << __FILE__ << "\tline: " << __LINE__
      << std::endl;
      throw;
    } catch (std::exception& e) {
      std::cout << "Error: " << e.what() << std::endl;
    }
  }

  /**
   * Encoding operator.
   * @param[in] inputDataHandle XData [Nx*Ny*numFrames]
   * @param[in] sensitivityMapsDataHandle Sensitivity maps
   * @param[in] samplingMasksDataHandle Sampling mask
   * @param[out] outputDataHandle KData [Nx*Ny*numFrames*numCoils]
   *
   * @param[in] profilingEnabled
   */
  void NestaUp::operatorA(std::shared_ptr<Data> inputData, std::shared_ptr<SensitivityMapsData> sensitivityMapsData,
                          std::shared_ptr<SamplingMasksData> samplingMasksData, std::shared_ptr<Data> outputData,
                          std::shared_ptr<Data> auxDataFFT, bool profilingEnabled){

    try {
      // XImage*SensMap Out of place
      pDataAndSensitivityMapsProduct->setInput(inputData);
      pDataAndSensitivityMapsProduct->setOutput(auxDataFFT);
      pDataAndSensitivityMapsProduct->setLaunchParameters(std::make_shared<ComplexElementProd::LaunchParameters>(ComplexElementProd::notConjugate, sensitivityMapsData));
      pDataAndSensitivityMapsProduct->launch();

      // FFT out of place (forward FFT)
      pFFTOutOfPlace->setInput(auxDataFFT);
      pFFTOutOfPlace->setOutput(outputData);
      pFFTOutOfPlace->setLaunchParameters(std::make_shared<FFT::LaunchParameters>(FFT::FORWARD));
      pFFTOutOfPlace->launch();

      pDataAndSamplingMasksProduct->setInput(outputData);
      pDataAndSamplingMasksProduct->setOutput(outputData);
      pDataAndSamplingMasksProduct->setLaunchParameters(std::make_shared<BlankDataRows::LaunchParameters>(samplingMasksData));
      pDataAndSamplingMasksProduct->launch();

    } catch (cl::Error& err) {
      /////////////////////////////////////////////////////////////////
      // Catch OpenCL errors and print log if it is a build error
      /////////////////////////////////////////////////////////////////
      std::cerr << "ERROR: " << err.what() << " (" << err.err() << ", "
      << CLapp::getOpenCLErrorCodeStr(err.err()) << ")"
      << "\tfile: " << __FILE__ << "\tline: " << __LINE__
      << std::endl;
      throw;
    } catch (std::exception& e) {
      std::cout << "Error: " << e.what() << std::endl;
    }
  }


  /**
   * Sparsity operator. Motion Conpensation operator (if the recontruction is with motion compensation) and Temporal Total Variation (tTV) operator.
   * @param[in] inputDataHandle XData [Nx*Ny*numFrames]
   * @param[out] outputDataHandle XData [Nx*Ny*numFrames]
   * @param[in] auxiliarDataHandleMC Auxiliar XData [Nx*Ny*numFrames] for MotionCompensation
   * @param[in] argsMC  Structure with arguments obtained in Motion Estimation, needed in Motion Compensation
   * @param[in] profilingEnabled
   */
  void NestaUp::operatorU(std::shared_ptr<Data> inputData, std::shared_ptr<Data> outputData,
                          std::shared_ptr<Data> auxiliarDataMC, ArgumentsMotionCompensation* argsMC, bool profilingEnabled){

    // In the case we perform Nesta reconstruction w/ motion compensation
    if (argsMC != nullptr){
      try {
	pMotionCompensation->setInput(inputData);
	pMotionCompensation->setOutput(auxiliarDataMC);
	pMotionCompensation->setLaunchParameters(std::make_shared<MotionCompensation::LaunchParameters>(argsMC));
	pMotionCompensation->launch();
      } catch (cl::Error& err) {
	/////////////////////////////////////////////////////////////////
	// Catch OpenCL errors and print log if it is a build error
	/////////////////////////////////////////////////////////////////
	std::cerr << "ERROR: " << err.what() << " (" << err.err() << ", "
	<< CLapp::getOpenCLErrorCodeStr(err.err()) << ")"
	<< "\tfile: " << __FILE__ << "\tline: " << __LINE__
	<< std::endl;
	throw;
      } catch (std::exception& e) {
	std::cout << "Error: " << e.what() << std::endl;
      }
    } else{
      auxiliarDataMC = inputData;
    }

    try {
      pTemporalTV->setInput(auxiliarDataMC);
      pTemporalTV->setOutput(outputData);
      pTemporalTV->launch();

    } catch (cl::Error& err) {
      /////////////////////////////////////////////////////////////////
      // Catch OpenCL errors and print log if it is a build error
      /////////////////////////////////////////////////////////////////
      std::cerr << "ERROR: " << err.what() << " (" << err.err() << ", "
      << CLapp::getOpenCLErrorCodeStr(err.err()) << ")"
      << "\tfile: " << __FILE__ << "\tline: " << __LINE__
      << std::endl;
      throw;
    } catch (std::exception& e) {
      std::cout << "Error: " << e.what() << std::endl;
    }
  }


  /**
   * Adjoint Sparsity operator. Adjoint Motion Conpensation operator (if the recontruction is with motion compensation) and adjoint Temporal Total Variation operator.
   * @param[in] inputDataHandle XData [Nx*Ny*numFrames]
   * @param[out] outputDataHandle XData [Nx*Ny*numFrames]
   * @param[in] auxiliarDataHandleMC Auxiliar XData [Nx*Ny*numFrames] for MotionCompensation
   * @param[in] argsMC  Structure with arguments obtained in Motion Estimation, needed in Motion Compensation
   * @param[in] profilingEnabled
   */
  void NestaUp::operatorUt(std::shared_ptr<Data> inputData, std::shared_ptr<Data> outputData,
                           std::shared_ptr<Data> auxiliarDataMC, ArgumentsMotionCompensation* argsMC, bool profilingEnabled){

    try {
      pTemporalTVt->setInput(inputData);
      pTemporalTVt->setOutput(auxiliarDataMC);
      pTemporalTVt->launch();

    } catch (cl::Error& err) {
      /////////////////////////////////////////////////////////////////
      // Catch OpenCL errors and print log if it is a build error
      /////////////////////////////////////////////////////////////////
      std::cerr << "ERROR: " << err.what() << " (" << err.err() << ", "
      << CLapp::getOpenCLErrorCodeStr(err.err()) << ")"
      << "\tfile: " << __FILE__ << "\tline: " << __LINE__
      << std::endl;
      throw;
    } catch (std::exception& e) {
      std::cout << "Error: " << e.what() << std::endl;
    }

    // In the case we perform Nesta reconstruction w/ motion compensation
    if (argsMC != nullptr){
      try {
	pAdjointMotionCompensation->setInput(auxiliarDataMC);
	pAdjointMotionCompensation->setOutput(outputData);
	pAdjointMotionCompensation->setLaunchParameters(std::make_shared<AdjointMotionCompensation::LaunchParameters>(argsMC));
	pAdjointMotionCompensation->launch();
      } catch (cl::Error& err) {
	/////////////////////////////////////////////////////////////////
	// Catch OpenCL errors and print log if it is a build error
	/////////////////////////////////////////////////////////////////
	std::cerr << "ERROR: " << err.what() << " (" << err.err() << ", "
	<< CLapp::getOpenCLErrorCodeStr(err.err()) << ")"
	<< "\tfile: " << __FILE__ << "\tline: " << __LINE__
	<< std::endl;
	throw;
      } catch (std::exception& e) {
	std::cout << "Error: " << e.what() << std::endl;
      }
    } else{
      pCopy->setInput(auxiliarDataMC);
      pCopy->setOutput(outputData);
      pCopy->launch();
    }
  }


  /**
   * Estimates the spectral norm of a operator using the power method. Used for estimate the norm of Sparsity operator.
   * @param[in] rows Rows of the data
   * @param[in] cols Cols of the data
   * @param[in] NumFrames Number of frames of the data
   * @param[in] argsMC  Structure with arguments obtained in Motion Estimation, needed in Motion Compensation
   * @param[in] profilingEnabled
   * @return the norm of the operator
   */
  float NestaUp::myNormest(uint rows, uint cols, uint NumFrames,  ArgumentsMotionCompensation* argsMC, bool profilingEnabled){

    float n = rows*cols*NumFrames;
    float tol = 1E-3;
    unsigned int maxiter = 30;
    unsigned int cnt = 0;
    float e;
    float e0;
    cl_float2 aux;

    cl_int status;
    cl_command_queue queue;
    queue = (getApp()->getCommandQueue(0))();
    cl_event event;

    float * norm = new float[1]();
    float * nnzSx = new float[1];

    cl_mem normObj = clCreateBuffer((getApp()->getContext())(), CL_MEM_READ_WRITE, sizeof(cl_float2), NULL, NULL);
    cl_mem nnzSxObj = clCreateBuffer((getApp()->getContext())(), CL_MEM_READ_WRITE, sizeof(cl_float2), NULL, NULL);

    complexType* pElement = new complexType(1,0);
    std::vector <complexType>* pData = new std::vector<complexType>(rows*cols, *pElement);
    std::vector <dimIndexType > *pDims = new std::vector<dimIndexType>( { rows, cols });
    NDArray* pObjNDArray = new ConcreteNDArray<complexType>(pDims, pData);
    std::vector<NDArray*>* pObjNDArrays = new std::vector<NDArray*>();
    for(uint i=0; i<NumFrames-1; i++){
        pObjNDArrays->push_back(pObjNDArray);
        pObjNDArray = new ConcreteNDArray<complexType>(pObjNDArray); // Create a copy of NDArray object
    }
    pObjNDArrays->push_back(pObjNDArray); // Add last element copied
    std::vector<realType>* pPixelSize = new std::vector<realType>( { 1, 1 });
    std::vector<dimIndexType>* pDynDims = new std::vector<dimIndexType>({NumFrames});


    std::shared_ptr<Data> pXData = std::make_shared<XData>(getApp(), pObjNDArrays, pDynDims, pPixelSize);
    std::shared_ptr<Data> pSxData = std::make_shared<XData>(getApp(), std::dynamic_pointer_cast<XData>(pXData), false);
    std::shared_ptr<Data> pMCAux = std::make_shared<XData>(getApp(), std::dynamic_pointer_cast<XData>(pXData), false);


    std::vector<complexType>* Sx;
    cl_mem pXBuffer = (*(pXData->getDeviceBuffer()))();
    cl_mem pSxBuffer = (*(pSxData->getDeviceBuffer()))();

    status = CLBlastScnrm2(n, normObj, 0, pXBuffer, 0, 1, &queue, &event);
    if (status != CL_SUCCESS) {
      printf("CLBlastScnrm2() failed with %d\n", status);
    } else {
      // Wait for calculations to be finished.
      status = clWaitForEvents(1, &event);
      // Fetch results of calculations from GPU memory.
      status = clEnqueueReadBuffer(queue, normObj, CL_TRUE, 0, sizeof(cl_float),norm, 0, NULL, NULL);
    }

    e=norm[0];
    aux.x=1/e;
    aux.y=0;

    status = CLBlastCscal( n, aux, pXBuffer, 0, 1, &queue, &event);
    if (status != CL_SUCCESS) {
      printf("CLBlastCscal() failed with %d\n", status);
    }

    e0 = 0;

    while((abs(e-e0)>(tol*e)) && (cnt <maxiter)){

      e0=e;

      operatorU(pXData, pSxData, pMCAux, argsMC, profilingEnabled);

      status = CLBlastScasum ( n, nnzSxObj, 0, pSxBuffer, 0, 1, &queue, &event);
      if (status != CL_SUCCESS) {
	printf("CLBlastSasum() failed with %d\n", status);
      } else {
	// Wait for calculations to be finished.
	status = clWaitForEvents(1, &event);
	// Fetch results of calculations from GPU memory.
	status = clEnqueueReadBuffer(queue, nnzSxObj, CL_TRUE, 0, sizeof(cl_float), nnzSx, 0, NULL, NULL);
      }

      if (nnzSx[0] == 0){
	pObjNDArrays = new std::vector<NDArray*>();

	for(uint i=0; i<NumFrames; i++){
	  Sx = new std::vector<complexType>(rows*cols);
	  for(uint j=0; j<rows*cols; j++){
	    Sx->at(j)=complexType(drand48(), 0);
	  }
	  pDims = new std::vector<dimIndexType>( { rows, cols });
	  pObjNDArray = new ConcreteNDArray<complexType>(pDims, Sx);
	  pObjNDArrays->push_back(pObjNDArray);
	}
	pPixelSize = new std::vector<realType>( { 1, 1 });
	pDynDims = new std::vector<dimIndexType>({NumFrames});
        pSxData = std::make_shared<XData>(getApp(), pObjNDArrays, pDynDims, pPixelSize);
        pSxBuffer =(*(pSxData->getDeviceBuffer()))();
      }

      status = CLBlastScnrm2(n, normObj, 0, pSxBuffer, 0, 1, &queue, &event);
      if (status != CL_SUCCESS) {
	printf("CLBlastScnrm2() failed with %d\n", status);
      } else {
	// Wait for calculations to be finished.
	status = clWaitForEvents(1, &event);
	// Fetch results of calculations from GPU memory.
	status = clEnqueueReadBuffer(queue, normObj, CL_TRUE, 0, sizeof(cl_float),norm, 0, NULL, NULL);
      }

      e = norm[0];

      operatorUt(pSxData, pXData, pMCAux, argsMC, profilingEnabled);

      pXBuffer =(*(pXData->getDeviceBuffer()))();
      status = CLBlastScnrm2(n, normObj, 0, pXBuffer, 0, 1, &queue, &event);
      if (status != CL_SUCCESS) {
	printf("CLBlastScnrm2() failed with %d\n", status);
      } else {
	// Wait for calculations to be finished.
	status = clWaitForEvents(1, &event);
	// Fetch results of calculations from GPU memory.
	status = clEnqueueReadBuffer(queue, normObj, CL_TRUE, 0, sizeof(cl_float), norm, 0, NULL, NULL);
      }

      aux.x=1/norm[0];
      aux.y=0;

      status = CLBlastCscal( n, aux, pXBuffer, 0, 1, &queue, &event);
      if (status != CL_SUCCESS) {
	printf("CLBlastCscal() failed with %d\n", status);
      }
      cnt=cnt+1;
    }


    pXData = nullptr;
    pSxData = nullptr;

    status = clReleaseMemObject(normObj);
    status = clReleaseMemObject(nnzSxObj);

    delete(norm);
    delete(nnzSx);

    return e;
  }


  void NestaUp::launch(ProfileParameters profileParameters){
    auto pLP=std::dynamic_pointer_cast<LaunchParameters>(pLaunchParameters);

    infoItems.addInfoItem("Title", "NestaUp info");
    startProfiling(profileParameters.profilingEnabled);

    std::shared_ptr<SensitivityMapsData> sensitivityMapsData = std::dynamic_pointer_cast<KData>(getInput())->getSensitivityMapsData();
    std::shared_ptr<SamplingMasksData> samplingMasksData = std::dynamic_pointer_cast<KData>(getInput())->getSamplingMasksData();


    std::shared_ptr<Data> pAuxFFT = std::make_shared<KData>(getApp(), std::dynamic_pointer_cast<KData>(getInput()), false, true);

    // Soluci�n del adjunto para inicializar Nesta en la primera reconstrucci�n (sin MC)
    if (pLP->argsMC == nullptr){
        operatorAt(getInput(), sensitivityMapsData, getOutput(), pAuxFFT, profileParameters.profilingEnabled);
    }

    //getOutput()->device2Host(SyncSource::BUFFER_ONLY); // (automatically called by matlabSave)
    //std::dynamic_pointer_cast<XData>(getOutput())->matlabSave("NESTA_At_recon.mat", SyncSource::BUFFER_ONLY);


    std::shared_ptr<Data> pAuxMC = std::make_shared<XData>(getApp(), std::dynamic_pointer_cast<XData>(getOutput()), false);

    std::shared_ptr<Data> pUx_RefImage = std::make_shared<XData>(getApp(), std::dynamic_pointer_cast<XData>(getOutput()), false);

    operatorU(getOutput(), pUx_RefImage, pAuxMC, pLP->argsMC, profileParameters.profilingEnabled);

    uint cols = NDARRAYWIDTH(getInput()->getNDArray(0));
    uint rows = NDARRAYHEIGHT(getInput()->getNDArray(0));
    uint numFrames = getInput()->getDynDimsTotalSize();
    uint numCoils = (std::dynamic_pointer_cast<KData>(getInput()))->getNCoils();



    std::shared_ptr<Data> pWkXData = std::make_shared<XData>(getApp(), std::dynamic_pointer_cast<XData>(getOutput()), false);
    std::shared_ptr<Data> pXkXData = std::make_shared<XData>(getApp(), std::dynamic_pointer_cast<XData>(getOutput()), false);
    std::shared_ptr<Data> pYkXData = std::make_shared<XData>(getApp(), std::dynamic_pointer_cast<XData>(getOutput()), false);
    std::shared_ptr<Data> pZkXData = std::make_shared<XData>(getApp(), std::dynamic_pointer_cast<XData>(getOutput()), false);
    std::shared_ptr<Data> pUkXData = std::make_shared<XData>(getApp(), std::dynamic_pointer_cast<XData>(getOutput()), false);
    std::shared_ptr<Data> pAuxFxXData = std::make_shared<XData>(getApp(), std::dynamic_pointer_cast<XData>(getOutput()), false);
    std::shared_ptr<Data> pDfXData = std::make_shared<XData>(getApp(), std::dynamic_pointer_cast<XData>(getOutput()), false);
    std::shared_ptr<Data> pResKData = std::make_shared<KData>(getApp(), std::dynamic_pointer_cast<KData>(getInput()), false, true);
    std::shared_ptr<Data> pAResXData = std::make_shared<XData>(getApp(), std::dynamic_pointer_cast<XData>(getOutput()), false);


    cl_mem fxObj = clCreateBuffer((getApp()->getContext())(), CL_MEM_READ_WRITE, sizeof(cl_float2), NULL, NULL);
    cl_mem normResObj = clCreateBuffer((getApp()->getContext())(), CL_MEM_READ_WRITE, sizeof(cl_float2), NULL, NULL);
    cl_mem normUkObj = clCreateBuffer((getApp()->getContext())(), CL_MEM_READ_WRITE, sizeof(cl_float2), NULL, NULL);

    cl_mem pXrefBuffer;
    cl_mem pXkBuffer;
    cl_mem pYkBuffer;
    cl_mem pZkBuffer;
    cl_mem pAuxFxBuffer;
    cl_mem pUkBuffer;
    cl_mem pResBuffer;
    cl_mem pInitialKImageBuffer;
    cl_mem pDfBuffer;
    cl_mem pAResBuffer;
    cl_mem pWkBuffer;


    // Image maximum calculation (host operation)
    getOutput()->device2Host(SyncSource::BUFFER_ONLY);
    std::vector<float> maxImages(getOutput()->getNumNDArrays());
    dimIndexType width, height, depth;
    for (uint i = 0; i < getOutput()->getNumNDArrays(); i++) {
      width = NDARRAYWIDTH(getOutput()->getData()->at(i));
      height = NDARRAYHEIGHT(getOutput()->getData()->at(i));
      depth = NDARRAYDEPTH(getOutput()->getData()->at(i));
      if (depth == 0)
	depth = 1;

      std::vector<float> eachImageAbs(width*height);
      complexType* pComplexArray;
      pComplexArray = (complexType*) getOutput()->getHostBuffer(i);

      for (dimIndexType index1D = 0;  index1D < height*width*depth; index1D++) {
	eachImageAbs[index1D] = std::fabs(pComplexArray[index1D]);
      }
      maxImages[i] = *max_element(eachImageAbs.begin(), eachImageAbs.end());
    }
    float maxXref = *max_element(maxImages.begin(), maxImages.end());


    // Image maximum calculation (host operation)
    pUx_RefImage->device2Host(SyncSource::BUFFER_ONLY);
    for (uint i = 0; i < pUx_RefImage->getNumNDArrays(); i++) {
      width = NDARRAYWIDTH(pUx_RefImage->getData()->at(i));
      height = NDARRAYHEIGHT(pUx_RefImage->getData()->at(i));
      depth = NDARRAYDEPTH(pUx_RefImage->getData()->at(i));
      if (depth == 0)
	depth = 1;

      std::vector<float> eachImageAbs(width*height);
      complexType* pComplexArray;
      pComplexArray = (complexType*) pUx_RefImage->getHostBuffer(i);

      for (dimIndexType index1D = 0;  index1D < height*width*depth; index1D++) {
	eachImageAbs[index1D] = std::fabs(pComplexArray[index1D]);
      }
      maxImages[i] = *max_element(eachImageAbs.begin(), eachImageAbs.end());
    }
    float maxUXref = *max_element(maxImages.begin(), maxImages.end());

    //Initizalize variables
    float lambda = (pLP->lambda_i)*maxXref;
    float muf = (pLP->mu_f)*maxXref;
    float mu0 = 0.9*maxUXref;
    float normU = myNormest(rows, cols, numFrames,  pLP->argsMC, profileParameters.profilingEnabled);
    float muL = lambda/(pLP->La);
    if (muL>mu0){
      mu0 = muL;
    }
    float gamma = pow((muf/mu0), (1.0/(pLP->maxIntIter)));
    float mu = mu0;
    float gammat = pow(((pLP->tolVar)/0.1), (1.0/(pLP->maxIntIter)));
    pLP->tolVar = 0.1;

    float * fx = new float[1];
    float * normRes = new float[1];
    float * normUk = new float[1];

    float Ak;
    float apk;
    float tauk;

    float *fmean=new float[10];
    float mean_fmean;

    bool OK;
    float Lmu;
    float Lmu1;
    float qp; //Stopping criterion

    cl_float2 f;
    cl_command_queue queue = (getApp()->getCommandQueue(0))();
    cl_event event;
    cl_int status;

    for(uint nl = 1; nl <= (pLP->maxIntIter); nl++){
      mu = mu*gamma;
      pLP->tolVar = (pLP->tolVar)*gammat;
      NESTAUP_CERR("\n Beginning L1 Minimization; mu = " << mu << std::endl);

      /////////////////////////////
      ////   CORE NESTEROV UP  ////
      /////////////////////////////

      f.x=0.0f;
      f.y=0.0f;

      pWkBuffer = (*(pWkXData->getDeviceBuffer()))();

      //Scales a complex-float pWkBuffer by a complex-float constant (initialize to zero).
      status = CLBlastCscal( rows*cols*numFrames, f, pWkBuffer, 0, 1, &queue, &event);
      if (status != CL_SUCCESS) {
	std::cout << "\n CLBlastCsscal() failed with " << status << std::endl;
      }


      //Copies complex-float elements from pXrefBuffer to pXkBuffer.
      pXrefBuffer =(*(getOutput()->getDeviceBuffer()))();
      pXkBuffer = (*(pXkXData->getDeviceBuffer()))();
      status = CLBlastCcopy( rows*cols*numFrames, pXrefBuffer, 0, 1, pXkBuffer, 0, 1, &queue, &event);
      if (status != CL_SUCCESS) {
	std::cout << "\n CLBlastCcopy() failed with " << status << std::endl;
      }

      //Copies complex-float elements from pXrefBuffer to pYkBuffer.
      pYkBuffer = (*(pYkXData->getDeviceBuffer()))();
      status = CLBlastCcopy( rows*cols*numFrames, pXrefBuffer, 0, 1, pYkBuffer, 0, 1, &queue, &event);
      if (status != CL_SUCCESS) {
	std::cout << "\n CLBlastCcopy() failed with " << status << std::endl;
      }

      //Copies complex-float elements from pXrefBuffer to pZkBuffer.
      pZkBuffer = (*(pZkXData->getDeviceBuffer()))();
      status = CLBlastCcopy( rows*cols*numFrames, pXrefBuffer, 0, 1, pZkBuffer, 0, 1, &queue, &event);
      if (status != CL_SUCCESS) {
	std::cout << "\n CLBlastCcopy() failed with " << status << std::endl;
      }

      Ak = 0.0f;
      Lmu = normU/mu;
      fmean[0] = (2.2251E-308)/10;
      OK = false;

      Lmu = lambda*Lmu + (pLP->La);

      Lmu1 = 1/Lmu;


      for(uint k = 0; k < (pLP->maxIter); k++){

	////----START PERFORM L1 CONSTRAINT----////

	//Apply sparse operator
	operatorU(pXkXData, pUkXData, pAuxMC, pLP->argsMC, profileParameters.profilingEnabled);

	pAuxFxBuffer = (*(pAuxFxXData->getDeviceBuffer()))();
	pUkBuffer = (*(pUkXData->getDeviceBuffer()))();

        //Copies complex-float elements from pUkBuffer to pAuxFxBuffer.
	status = CLBlastCcopy( rows*cols*numFrames, pUkBuffer, 0, 1, pAuxFxBuffer, 0, 1,&queue, &event);
	if (status != CL_SUCCESS) {
	  std::cout << "\n CLBlastCcopy() failed with " << status << std::endl;
	}

	//process vectorNormalization to normalize pUkXData
	pVectorNormalization->setInput(pUkXData);
	pVectorNormalization->setOutput(pUkXData);
	pVectorNormalization->setLaunchParameters(std::make_shared<VectorNormalization::LaunchParameters>(mu));
	pVectorNormalization->launch();

        //computes the euclidean norm of pUkBuffer containing float-complex elements
	status = CLBlastScnrm2(rows*cols*numFrames, normUkObj, 0, pUkBuffer, 0, 1, &queue, &event);
	if (status != CL_SUCCESS) {
	  std::cout << "\n CLBlastScnrm2() failed with " << status << std::endl;
	} else {
	  /* Wait for calculations to be finished. */
	  status = clWaitForEvents(1, &event);
	  /* Fetch results of calculations from GPU memory. */
	  status = clEnqueueReadBuffer(queue, normUkObj, CL_TRUE, 0, sizeof(cl_float), normUk, 0, NULL, NULL);
	}

	//dot product of two vectors containing float-complex elements (pUkBuffer and pAuxFxBuffer) conjugating pUkBuffer
	status = CLBlastCdotc(rows*cols*numFrames, fxObj, 0, pUkBuffer, 0, 1, pAuxFxBuffer, 0, 1, &queue, &event);
	if (status != CL_SUCCESS) {
	  std::cout << "\n CLBlastCdotu() failed with " << status << std::endl;
	} else {
	  /* Wait for calculations to be finished. */
	  status = clWaitForEvents(1, &event);
	  /* Fetch results of calculations from GPU memory. */
	  status = clEnqueueReadBuffer(queue, fxObj, CL_TRUE, 0, sizeof(cl_float), fx, 0, NULL, NULL);
	}

	fx[0]=fx[0]-(mu/2.0f)*pow(normUk[0],2);

        //Apply sparse operator (adjoint)
	operatorUt(pUkXData, pDfXData, pAuxMC, pLP->argsMC, profileParameters.profilingEnabled);

	////----END PERFORM L1 CONSTRAINT----////

        //Apply encoding operator
	operatorA(pXkXData, sensitivityMapsData, samplingMasksData, pResKData, pAuxFFT, profileParameters.profilingEnabled);

	pResBuffer = (*(pResKData->getDeviceBuffer()))();
	pInitialKImageBuffer = (*(getInput()->getDeviceBuffer()))();

        f.x=-1.0f;
        f.y=0.0f;

        /*
        CERR("tama�o pInitialKImageBuffer: "
            << getInput()->getDeviceBuffer()->getInfo<CL_MEM_SIZE>() << "bytes \n");
        CERR("tama�o pResBuffer: "
            << pResKData->getDeviceBuffer()->getInfo<CL_MEM_SIZE>() << "bytes \n");
        CERR("pInitialKImageBuffer->getDeviceBuffer(): " << getInput()->getDeviceBuffer() << "\n");
        CERR("pResBuffer->getDeviceBuffer(): " << pResKData->getDeviceBuffer() << "\n");
        CERR("pInitialKImageBuffer->getDeviceBuffer() cl_mem: " << (*(getInput()->getDeviceBuffer()))() << "\n");
        CERR("pResBuffer->getDeviceBuffer() cl_mem: " << (*(pResKData->getDeviceBuffer()))() << "\n");
        */
        //Scale vector pInitialKImageBuffer of complex-float elements and add to pResBuffer.
	status = CLBlastCaxpy( rows*cols*numFrames*numCoils,  f, pInitialKImageBuffer, 0, 1, pResBuffer, 0, 1,&queue, &event);
	if (status != CL_SUCCESS) {
	  std::cout << "\n CLBlastCaxpy() failed with " << status << std::endl;
	}

	//Apply encoding operator (adjoint)
	operatorAt(pResKData, sensitivityMapsData, pAResXData, pAuxFFT, profileParameters.profilingEnabled);

        f.x=lambda;
        f.y=0;

	pDfBuffer = (*(pDfXData->getDeviceBuffer()))();

        //Scales a complex-float pDfBuffer by a complex-float constant.
	status = CLBlastCscal( rows*cols*numFrames, f, pDfBuffer, 0, 1, &queue, &event);
	if (status != CL_SUCCESS) {
	  std::cout << "\n CLBlastCsscal() failed with " << status << std::endl;
	}

	pAResBuffer = (*(pAResXData->getDeviceBuffer()))();

        f.x=1.0f;
        f.y=0;

        //Scale vector pAResBuffer of complex-float elements and add to pDfBuffer.
	status = CLBlastCaxpy( rows*cols*numFrames, f , pAResBuffer, 0, 1, pDfBuffer , 0, 1, &queue, &event);
	if (status != CL_SUCCESS) {
	  std::cout << "\n CLBlastCaxpy() failed with " << status << std::endl;
	}

	//computes the euclidean norm of pResBuffer containing float-complex elements
	status = CLBlastScnrm2(rows*cols*numFrames*numCoils, normResObj, 0, pResBuffer, 0, 1, &queue, &event);
	if (status != CL_SUCCESS) {
	  std::cout << "\n CLBlastScnrm2() failed with " << status << std::endl;
	} else {
	  /* Wait for calculations to be finished. */
	  status = clWaitForEvents(1, &event);
	  /* Fetch results of calculations from GPU memory. */
	  status = clEnqueueReadBuffer(queue, normResObj, CL_TRUE, 0, sizeof(cl_float), normRes, 0, NULL, NULL);
	}

	fx[0] = lambda*fx[0] + (0.5f)*pow(normRes[0],2);

	//--- Updating yk ---//

        //Copies complex-float elements from pXkBuffer to pYkBuffer.
	status = CLBlastCcopy(rows*cols*numFrames, pXkBuffer, 0, 1, pYkBuffer, 0, 1, &queue, &event);
	if (status != CL_SUCCESS) {
	  std::cout << "\n CLBlastCcopy() failed with " << status << std::endl;
	}

	f.x=-Lmu1;
        f.y=0;

        //Scale vector pDfBuffer of complex-float elements and add to pYkBuffer.
	status = CLBlastCaxpy(rows*cols*numFrames, f, pDfBuffer, 0, 1, pYkBuffer, 0, 1, &queue, &event);
	if (status != CL_SUCCESS) {
	  std::cout << "\n CLBlastCaxpy() failed with " << status << std::endl;
	}

        //-------------------//

	//---Stopping criterion ---//
	mean_fmean=0.0f;

	if(k<=8){
	  for(uint i=0;i<(k+1);i++){
	    mean_fmean+=fmean[i];
	  }
	  mean_fmean=mean_fmean/(k+1);
	} else{
	  for(uint i=0;i<10;i++){
	    mean_fmean+=fmean[i];
	  }
	  mean_fmean=mean_fmean/(10.0f);
	}
	qp = fabs((fx[0] - mean_fmean))/mean_fmean;

	switch (pLP->stoptest){

	  case 1:

	    if(qp <= (pLP->tolVar) && OK==true){
	      NESTAUP_CERR("\n Stop criterion 2 OK!!!"<<std::endl);
	      NESTAUP_CERR("\n Iter: "<<k+1<<std::endl);
	      NESTAUP_CERR("\n fx: "<<fx[0]<<std::endl);
	      NESTAUP_CERR("\n qp: "<<qp<<std::endl);
	      NESTAUP_CERR("\n TolVar: "<<(pLP->tolVar)<<std::endl);

	      k=(pLP->maxIter);
	    }
	    if(qp <= (pLP->tolVar) && OK==false){
	      NESTAUP_CERR("\n Stop criterion 1 OK!!!"<<std::endl);
	      NESTAUP_CERR("\n Iter: "<<k+1<<std::endl);
	      NESTAUP_CERR("\n fx: "<<fx[0]<<std::endl);
	      NESTAUP_CERR("\n qp: "<<qp<<std::endl);
	      NESTAUP_CERR("\n TolVar: "<<(pLP->tolVar)<<std::endl);

	      OK=true;
	    }
	    break;

	  default:
	    break;
	}

	//-------------------//

	if (k<(pLP->maxIter)){

	  if(k<=8){
	    fmean[k+1]=fx[0];
	  } else{
	    for(uint i=0;i<9;i++){
	      fmean[i]=fmean[i+1];
	    }
	    fmean[9]=fx[0];
	  }

	  //--- Updating zk ---//

	  apk =0.5f*(k+1.0f);
	  Ak = Ak + apk;
	  tauk = 2.0f/(k+3.0f);

          f.x=apk;
          f.y=0;

          //Scale vector pDfBuffer of complex-float elements and add to pWkBuffer.
	  status = CLBlastCaxpy( rows*cols*numFrames, f , pDfBuffer, 0, 1, pWkBuffer, 0, 1, &queue, &event);
	  if (status != CL_SUCCESS) {
	    std::cout << "\n CLBlastCaxpy() failed with " << status << std::endl;
	  }

	  //Copies complex-float elements from pXrefBuffer to pZkBuffer.
	  status = CLBlastCcopy( rows*cols*numFrames, pXrefBuffer, 0, 1, pZkBuffer, 0, 1, &queue, &event);
	  if (status != CL_SUCCESS) {
	    std::cout << "\n CLBlastCcopy() failed with " << status << std::endl;
	  }

          f.x=-Lmu1;
          f.y=0;

          //Scale vector pWkBuffer of complex-float elements and add to pZkBuffer.
	  status = CLBlastCaxpy( rows*cols*numFrames, f , pWkBuffer, 0, 1, pZkBuffer, 0, 1, &queue, &event);
	  if (status != CL_SUCCESS) {
	    std::cout << "\n CLBlastCaxpy() failed with " << status << std::endl;
	  }


          //-------------------//

	  //--- Updating xk ---//

	  //Copies complex-float elements from pZkBuffer to pXkBuffer.
	  status = CLBlastCcopy( rows*cols*numFrames, pZkBuffer, 0, 1, pXkBuffer, 0, 1, &queue, &event);
	  if (status != CL_SUCCESS) {
	    std::cout << "\n CLBlastCcopy() failed with " << status << std::endl;
	  }

	  f.x=tauk;
          f.y=0;

          //Scales a complex-float pXkBuffer by a complex-float constant.
	  status = CLBlastCscal( rows*cols*numFrames, f, pXkBuffer, 0, 1, &queue, &event);
	  if (status != CL_SUCCESS) {
	    std::cout << "\n CLBlastCsscal() failed with " << status << std::endl;
	  }

          f.x=(1.0f-tauk);
          f.y=0;

          //Scale vector pYkBuffer of complex-float elements and add to pXkBuffer.
	  status = CLBlastCaxpy( rows*cols*numFrames, f, pYkBuffer, 0, 1, pXkBuffer, 0, 1, &queue, &event);
	  if (status != CL_SUCCESS) {
	    std::cout << "\n CLBlastCaxpy() failed with " << status << std::endl;
	  }

	  if ((k+1)%(pLP->verbose)==0){
	    NESTAUP_CERR("Iter: "<<(k+1)<<" ~ fmu: "<< fx[0]<<" ~ Rel. Variation of fmu: "<<qp<<std::endl);
	  }
	}
      }


      /////////////////////////////////
      ////   END CORE NESTEROV UP  ////
      /////////////////////////////////

      NESTAUP_CERR("----------------------------------------------------------------------------"<<std::endl);

      //Copies complex-float elements from pXkBuffer to pXrefBuffer.
      status = CLBlastCcopy( rows*cols*numFrames, pXkBuffer, 0, 1, pXrefBuffer, 0, 1, &queue, &event);
      if (status != CL_SUCCESS) {
	std::cout << "\n CLBlastCcopy() failed with " << status << std::endl;
      } else {
	/* Wait for calculations to be finished. */
	status = clWaitForEvents(1, &event);
      }
    }


    if(pLP->argsMC != nullptr){
      pAuxMC = nullptr;
    }

    pAuxFFT = nullptr;
    pUx_RefImage = nullptr;
    pWkXData = nullptr;
    pXkXData = nullptr;
    pYkXData = nullptr;
    pZkXData = nullptr;
    pUkXData = nullptr;
    pAuxFxXData = nullptr;
    pDfXData = nullptr;
    pResKData = nullptr;
    pAResXData = nullptr;

    clReleaseMemObject(fxObj);
    clReleaseMemObject(normResObj);
    clReleaseMemObject(normUkObj);

    delete[](fx);
    delete[](normRes);
    delete[](normUk);
    delete[](fmean);

    stopProfiling(profileParameters.profilingEnabled);

  }
}
#undef NESTAUP_DEBUG
