/* Copyright (C) 2018 Federico Simmross Wattenberg,
 *                    Manuel Rodrí­guez Cayetano,
 *                    Javier Royuela del Val,
 *                    Elena Martín González,
 *                    Elisa Moya Sáez,
 *                    Marcos Martín Fernández and
 *                    Carlos Alberola López
 *
 * This file is part of OpenCLIPER.
 *
 * OpenCLIPER is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * OpenCLIPER is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenCLIPER; If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Contact:
 *
 *  Federico Simmross Wattenberg
 *  E.T.S.I. Telecomunicación
 *  Universidad de Valladolid
 *  Paseo de Belén 15
 *  47011 Valladolid, Spain.
 *  fedsim@tel.uva.es
 */
#include <OpenCLIPER/OpenCLIPERDataModel.hpp>
#include <OpenCLIPER/ProgramConfig.hpp>

// Uncomment to show specific debug messages
#define RECON_DEBUG

#if !defined NDEBUG && defined RECON_DEBUG
    #define RECON_CERR(x) CERR(x)
#else
    #define RECON_CERR(x)
    #undef RECON_DEBUG
#endif


using namespace OpenCLIPER;
int main(int argc, char *argv[]) {
  try{
    CLapp::PlatformTraits platformTraits;
    CLapp::DeviceTraits deviceTraits;
    ProgramConfig* pProgramConfig = new ProgramConfig(argc, argv);
    auto pConfigTraits = std::dynamic_pointer_cast<ProgramConfig::ConfigTraits>(
        pProgramConfig->getConfigTraits());

    deviceTraits = pConfigTraits->deviceTraits;
    platformTraits = pConfigTraits->platformTraits;
    // CL_QUEUE_PROFILING_ENABLE slows down the queue. Beware!
    deviceTraits.queueProperties = cl::QueueProperties();
    auto pCLapp=CLapp::create(platformTraits,deviceTraits);

    // Load OpenCL kernel(s)
    std::vector<std::string> kernelFileNames = {"complexElementProd.cl", "xImageSum.cl", "blankDataRows_kernel.cl",
						"creabspline.cl", "initialize.cl", "optimizer.cl", "reductions.cl",
						"temporalTV.cl", "vectorNormalization_kernel.cl"};
    pCLapp->loadKernels(kernelFileNames);

    // Load input data from Matlab file
    std::string filename = "../../tests/matlabKImages/data_160x160_coils23_frames20_single.mat";
    std::shared_ptr<Data> pInputKData = std::make_shared<KData>(pCLapp, filename);

    // Create output with suitable size
    std::shared_ptr<XData> pOutputXData = std::make_shared<XData>(pCLapp, std::dynamic_pointer_cast<KData>(pInputKData));

    // Parameters for NESTA execution
    // -------------------------------------------------------------------------------------------------
    float lambda_i = 3E-2;	//3E-2
    float mu_f = 1E-4;		//1E-10
    float La = 1;
    uint maxIntIter = 4; 	// 6
    float tolVar = 1E-3;	//1E-4
    int verbose = 10;
    int maxIter = 3000;		// 3000 // 300
    uint stoptest = 1;

    // Parameters for GROUPWISE REGISTRATION execution
    // -------------------------------------------------------------------------------------------------
    uint MOTION_ITERS = 3; 	// Number of iterations to perform GroupwiseRegistration + Nesta
    uint nmax = 100; 		// Maximum number of iterations
    float et = 0.001f; 		// Transformation norm threshold
    float eh = 0.001f; 		// Metric variation threshold
    std::vector<float> lambda = {1.f,6.f,1.f,5.f}; // Smoothnes/Regularization weights (1st spatial, 2nd spatial, 1st temp, 2nd temp)
    int radius = 30; 		// Radius of the circular ROI (0 if 'wholebox')
    int E = 3; 			// E: spline order
    int Dp[2] = {4,4}; 		// Dp: point density
    float W = 1.f; 		// Weight for each point displacement
    bool flagW = true; 		// Flag for adaptative set of W



    // Create new process, set its input/output data sets and bind it to our CL app
    // Reuse process NestaUp
    std::unique_ptr<Process> nestaUpProcess(new NestaUp(pCLapp));
    nestaUpProcess->setInput(pInputKData);
    nestaUpProcess->setOutput(pOutputXData);
    nestaUpProcess->init();

    // Reuse process GroupwiseRegistration
    std::shared_ptr<Data> TData;
    Bspline* ts = new Bspline();

    std::unique_ptr<Process> GWRegistrationProcess(new GroupwiseRegistration(pCLapp));
    GWRegistrationProcess->setInput(pOutputXData); // Only needs input
    GWRegistrationProcess->init();
    std::shared_ptr<GroupwiseRegistration::LaunchParameters> paramsGW;

    //Start measuring execution time now
    std::chrono::high_resolution_clock::time_point beginReconExecTime = std::chrono::high_resolution_clock::now();

    ///////////////////////////////////////////////////////////////////////////////////////////
    // NESTA - First reconstruction                                                          //
    ///////////////////////////////////////////////////////////////////////////////////////////

    auto paramsNestaUp = std::make_shared<NestaUp::LaunchParameters>(lambda_i, mu_f, La, maxIntIter, tolVar, verbose, maxIter, stoptest, nullptr);
    nestaUpProcess->setLaunchParameters(paramsNestaUp);
    nestaUpProcess->launch();

    pOutputXData->device2Host(SyncSource::BUFFER_ONLY);
#if !defined NDEBUG && defined RECON_DEBUG
      (std::dynamic_pointer_cast<XData>(pOutputXData))->matlabSave("../../output/NESTA_nomotion_recon.mat");
#endif

    for(unsigned int i=0; i<MOTION_ITERS; i++){
      ///////////////////////////////////////////////////////////////////////////////////////////
      // Groupwise Registration to estimate cardiac motion                                     //
      ///////////////////////////////////////////////////////////////////////////////////////////
      ArgumentsMotionCompensation *argsMC = new ArgumentsMotionCompensation();
      paramsGW = std::make_shared<GroupwiseRegistration::LaunchParameters>(argsMC, nmax, et, eh, lambda, radius, E, Dp, W, flagW, TData, ts, i);
      GWRegistrationProcess->setLaunchParameters(paramsGW);
      GWRegistrationProcess->launch();
      TData = argsMC->TData; // Reuse T data obtained in current iteration as initialization for next iteration

#if !defined NDEBUG && defined RECON_DEBUG
	std::shared_ptr<XData> pOutputRegData = std::make_shared<XData>(pCLapp, pOutputXData);
        std::unique_ptr<Process> MCProcess(new MotionCompensation(pCLapp));
        MCProcess->setLaunchParameters(std::make_shared<MotionCompensation::LaunchParameters>(argsMC));
        MCProcess->setInput(pOutputXData);
        MCProcess->setOutput(pOutputRegData);
        MCProcess->launch();
        MCProcess.reset(nullptr);
        pOutputRegData->device2Host(SyncSource::BUFFER_ONLY);
        pOutputRegData->matlabSave("../../output/registration_"+ std::to_string(i+1) + "recon.mat");
        pOutputRegData = nullptr;
#endif

      ///////////////////////////////////////////////////////////////////////////////////////////
      // NESTA - Reconstruction with motion compensation                                       //
      ///////////////////////////////////////////////////////////////////////////////////////////
      paramsNestaUp = std::make_shared<NestaUp::LaunchParameters>(lambda_i, mu_f, La, maxIntIter, tolVar, verbose, maxIter, stoptest, argsMC);
      nestaUpProcess->setLaunchParameters(paramsNestaUp);
      nestaUpProcess->launch();
      delete argsMC;

#if !defined NDEBUG && defined RECON_DEBUG
        pOutputXData->device2Host(SyncSource::BUFFER_ONLY);
        pOutputXData->matlabSave("../../output/NESTA_motion"+ std::to_string(i+1) + "recon.mat");
#endif
    }

    delete ts;

    // Execution time measurement ends now
    std::chrono::high_resolution_clock::time_point endReconExecTime = std::chrono::high_resolution_clock::now();
    TIME_DIFF_TYPE elapsedTimeRecon = (std::chrono::duration_cast<std::chrono::nanoseconds>(endReconExecTime - beginReconExecTime).count()) / 1e9;

    std::ostringstream ostream2;
    ostream2 << std::fixed << std::setprecision(PROFILINGTIMESPRECISION) << elapsedTimeRecon;
    std::cout << " --> Time elapsed in reconstruction end to end (s): " << ostream2.str() << "\n" << std::endl;

    // Show reconstructed data
    pOutputXData->show();

    // Save reconstructed data
    pOutputXData->matlabSave("MRIReconEnd2EndResult.mat");

    GWRegistrationProcess.reset(nullptr);
    nestaUpProcess.reset(nullptr);

    pInputKData = nullptr;
    pOutputXData = nullptr;
    TData = nullptr;

  }
  catch(cl::BuildError& e) {
      CLapp::dumpBuildError(e);
  }
  catch(cl::Error& e) {
  	std::cerr << "CL Exception "<<e.err()<<" ("<<CLapp::getOpenCLErrorCodeStr(e.err())<<"): " << e.what() << std::endl;
  }
  catch (std::exception& e) {
    std::cerr << "Error: " << e.what() << "\tfile: " << __FILE__ << "\tline: " << __LINE__  << std::endl;
  }
}

#undef RECON_DEBUG
